/**
	$Id: PlainTextTemplate.java 7398 2017-07-03 07:37:01Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.plugins;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Collection;
import java.util.List;

import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.xssf.streaming.SXSSFWorkbook;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.apache.poi.xssf.usermodel.XSSFWorkbookType;

import net.sf.basedb.core.Item;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.excel.CellStyleCreator;
import net.sf.basedb.util.excel.ExcelValue;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.util.formatter.PrefixSuffixFormatter;

/**
 	Template implementation that exports data to Excel format (xlsx).
 	 
	@author nicklas
	@since 3.15
*/
public class XlsxTemplate
	implements ExportTemplate
{

	private final OutputStream out;
	private final List<ExportedProperty> properties;
	private int colNum;
	private int rowNum;
	private String collectionSeparator;
	
	private final Workbook workbook;
	private final Sheet sheet;
	private final CellStyleCreator styleCreator;
	private Row currentRow;

	
	/**
		Create a new template.
		@param out The stream to write to
		@param properties All exported properties
		@param itemType The type of items we are exporting
	*/
	public XlsxTemplate(OutputStream out, List<ExportedProperty> properties, Item itemType, long numItems, String collectionSeparator)
	{
		this.out = out;
		this.properties = properties;
		this.collectionSeparator = collectionSeparator == null ? "," : collectionSeparator;
		
		// Use streaming API if more than 1000 items -- do not forget to call workbook.dispose() at the end
		this.workbook = numItems > 1000 ? new SXSSFWorkbook(1000) : new XSSFWorkbook(XSSFWorkbookType.XLSX);
		this.sheet = workbook.createSheet(itemType.name());
		this.styleCreator = new CellStyleCreator(workbook);
	}

	/**
		@return application/vnd.openxmlformats-officedocument.spreadsheetml.sheet
	*/
	@Override
	public String getMimeType()
	{
		return "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet";
	}

	/**
		@return Always "xlsx"
	*/
	@Override
	public String getDefaultFileExtension()
	{
		return "xlsx";
	}
	
	/**
		Reset row and column to 0.
	*/
	@Override
	public void beginDocument() 
		throws IOException
	{
		rowNum = 0;
		colNum = 0;
	}

	/**
		Write the headers on a single row. 
	*/
	@Override
	public void writeHeaders() 
		throws IOException
	{
		currentRow = sheet.createRow(rowNum++);
		colNum = 0;
		for (ExportedProperty ep : properties)
		{
			Cell cell = currentRow.createCell(colNum);
			cell.setCellValue(ep.title);
			// Set column width based on title length + margin --  but not less than 10 or wider than 50
			int colWidth = 2 + Math.min(Math.max(8, ep.title.length()), 48);
			sheet.setColumnWidth(colNum, colWidth*256);
			colNum++;
		}
	}

	/**
		Reset column number and create a new row.
	*/
	@Override
	public void beginItem(int id) 
		throws IOException
	{
		currentRow = sheet.createRow(rowNum++);
		colNum = 0;
	}

	/**
		Join the annotation values into a single string with a comma.
	*/
	@Override
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public void writeAnnotations(ExportedProperty ep, List<?> values, String unit)
		throws IOException
	{
		if (values == null || values.size() == 0)
		{
			writeProperty(ep, null);
		}
		else if (values.size() == 1 && unit == null)
		{
			writeProperty(ep, values.get(0));
		}
		else
		{
			Cell cell = currentRow.createCell(colNum++);
			Formatter f = ep.formatter;
			if (unit != null) f = new PrefixSuffixFormatter("", f, PrefixSuffixFormatter.NBSP + unit);
			ExcelValue ev = ExcelValue.asString(Values.getString(values, collectionSeparator, true, f));
			ev.writeTo(cell, styleCreator);
		}
	}

	/**
		Write the data object.
	*/
	@Override
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public void writeProperty(ExportedProperty ep, Object data)
		throws IOException
	{
		Cell cell = currentRow.createCell(colNum++);		
		ExcelValue ev = ep.excelFormatter.toExcelValue(data);
		if (ev == null) ev = ExcelValue.emptyCell();
		ev.writeTo(cell, styleCreator);
	}
	
	/**
	 	Writes a collection of values.
	 */
	@Override
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public void writeCollection(ExportedProperty ep, Collection<?> values)
		throws IOException
	{
		if (values == null || values.size() == 0)
		{
			writeProperty(ep, null);
		}
		else if (values.size() == 1)
		{
			writeProperty(ep, values.iterator().next());
		}
		else
		{
			Cell cell = currentRow.createCell(colNum++);
			ExcelValue ev = ExcelValue.asString(Values.getString(values, collectionSeparator, true, ep.formatter));
			ev.writeTo(cell, styleCreator);
		}
	}

	/**
		Do nothing.
	*/
	@Override
	public void endItem() 
		throws IOException
	{}

	/**
		Write the Excel workbook to the output stream and close.
	*/
	@Override
	public void endDocument(ProgressReporter progress) 
		throws IOException
	{
		if (progress != null) progress.display(99, "Generating XLSX file. Please wait...");
		workbook.write(out);
	}

	/**
		Dispose and close workbook.
	*/
	@Override
	public void close() 
		throws IOException 
	{
		if (workbook instanceof SXSSFWorkbook)
		{
			((SXSSFWorkbook)workbook).dispose();
		}
		workbook.close();
	}



}
