/*
	$Id: ExceptionCatcher.java 7703 2019-04-11 07:31:26Z nicklas $

	Copyright (C) 2008, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;
import javax.servlet.jsp.tagext.BodyTagSupport;
import javax.servlet.jsp.tagext.TryCatchFinally;

/**
	This tag is a wrapper that catches any errors in it's
	body and outputs the error message.
	<p>
	<b>Example:</b>
	<pre>
   &lt;%@ taglib prefix="base" uri="/WEB-INF/base.tld" %&gt;
   &lt;base:catch"&gt;
   Here is some code that may throw an exception...
   &lt;/base:catch&gt;
</pre>

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;base:catch
      clazz=...
   &gt;
</pre>

	<table class="as-grid">
	<caption>Tag attributes</caption>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>inlineerror</td>
		<td>no</td>
		<td>
			If an exception is thrown the error message will be displayed
			inside a &lt;div&gt; with class set to this value.
		</td>
	</tr>
	</table>

	@author Nicklas
	@version 2.9.1
*/
public class ExceptionCatcher
	extends BodyTagSupport
	implements TryCatchFinally
{
	private static final long serialVersionUID = -5694802936612085039L;
	
	private String clazz = "inlineerror";
	
	/*
		Taglib initialization methods
	*/
	public void setClazz(String clazz)
	{
		this.clazz = clazz;
	}
	public String getClazz()
	{
		return clazz;
	}

	@Override
	public int doEndTag() 
		throws JspException
	{
		try
		{
			getBodyContent().writeOut(getBodyContent().getEnclosingWriter());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex);
		}
		return EVAL_PAGE;
	}
	/*
		From the TryCatchFinally interface
		----------------------------------
	 */
	@Override
	public void doCatch(Throwable t) 
		throws Throwable
	{
		t.printStackTrace();
		StringBuilder sb = new StringBuilder();
		sb.append("<div");
		if (getClazz() != null)
		{
			sb.append(" class=\"").append(getClazz()).append("\"");
		}
		sb.append(" title=\"See log file for more information about this error\"");
		sb.append(">");
		String msg = t.getMessage();
		if (msg == null) msg = t.getClass().getName();
		sb.append(msg);
		sb.append("</div>");
		
		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex);
		}
	}

	@Override
	public void doFinally()
	{}
	// ------------------------------------
}




