/*
	$Id: Menuseparator.java 7703 2019-04-11 07:31:26Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib.menu;

import java.util.Iterator;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;

import net.sf.basedb.clients.web.extensions.DynamicActionAttribute;
import net.sf.basedb.clients.web.taglib.StylableTag;

/**
	This tag defines a horizontal separator line inside a drop-down menu. This tag can 
	only be used inside a {@link Menu} tag.

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;m:menuseparator
      id=...
      clazz=...
      subclass=...
      style=...
      visible=...
   &gt;
</pre>

	<table class="as-grid">
	<caption>Tag attributes</caption>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>menuseparator1, menuseparator2, and so on...</td>
		<td>no</td>
		<td>The ID of the menuseparator. The value of this attribute
		goes directly into the standard HTML <code>id</code> attribute.
		This means that if you need a reference to the menuseparator object in your own
		JavaScript code, you may for example use the following code:
		<p>
		<code>
		menuseparator = document.getElementById('&lt;id&gt;');
		</code>
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>visible</td>
		<td>true</td>
		<td>no</td>
		<td>
			Use <code>false</code> to hide this separator entirely. 
		</td>
	</tr>
	</table>

	@author Nicklas
	@version 2.0
*/
public class Menuseparator 
	extends StylableTag
{

	/**
		
	*/
	private static final long serialVersionUID = -1758168104130104042L;

	private Menu menu;
	
	/**
		Optional <code>style</code> attribute
	*/
	private String style = null;

	/**
		If the separator is visible.
	*/
	private boolean visible = true;

	/*
		Taglib initialization methods
	*/
	public Menuseparator()
	{
		super("menuseparator");
	}
	
	public void setVisible(boolean visible)
	{
		this.visible = visible;
	}
	public boolean isVisible()
	{
		return visible;
	}	
	
	public void setDynamic(Iterator<DynamicActionAttribute> dynamicAttributes)
	{
		setMoreDynamicAttributes(dynamicAttributes);
	}


	@Override
	public int doStartTag() 
		throws JspException
	{
		if (!(getParent() instanceof Menu)) throw new JspTagException("Tag <m:menuseparator> must be inside a <m:menu> tag");
		menu = (Menu)getParent();

		if (!isVisible()) return SKIP_BODY;

		StringBuilder sb = new StringBuilder();
		sb.append("<div");
		addIdAndStyles(sb);
		addDynamicAttributes(sb);
		sb.append("></div>");
		try
		{
			menu.addMenuitem(sb.toString());
    	}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return SKIP_BODY;
	}
}


