/*
  $Id: AnnotationTypeCategory.java 7308 2017-03-13 08:15:46Z nicklas $

  	Copyright (C) 2006 Nicklas Nordborg, Dominic Oyeniran, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.core;

import java.util.Date;

import net.sf.basedb.core.data.AnnotationTypeCategoryData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;


/**
	Objects of this class defines an annotation type category. 
	The categories are used to group related {@link AnnotationType}:s.
	<p>
 	An annotation type can belong to more than one category. 
 	
 	@author Dominic Oyeniran
	@version 2.0
	@see AnnotationType
	@base.modified $Date: 2006-05-11 (Thu, 11 May 2006) $
*/

public class AnnotationTypeCategory 
	extends CommonItem
	implements Registered

{
	/**
	 	The type of item represented by this class.
	 	@see Item#ANNOTATIONTYPECATEGORY
	 	@see #getType()
    */
	public static final Item TYPE = Item.ANNOTATIONTYPECATEGORY;
	
		
	/**
	 	Creates a new <code>AnnotationTypeCategory</code> item.
	 	@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@return The new <code>AnnotationType</code> item
		@throws BaseException If there is an error
	*/
	public static AnnotationTypeCategory getNew(DbControl dc)
		throws BaseException
	{
		AnnotationTypeCategory atcg = dc.newItem(AnnotationTypeCategory.class);
		atcg.setName("New category");
		atcg.getData().setEntryDate(new Date());
		return atcg;
	}

	/**
	 	Get an <code>AnnotationTypeCategory</code> item when you know the id.
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param id The id of the item to load
		@return The <code>AnnotationTypeCategory</code> item
		@throws ItemNotFoundException If an item with the specified id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static AnnotationTypeCategory getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		AnnotationTypeCategory atcg = dc.loadItem(AnnotationTypeCategory.class, id);
		if (atcg == null) throw new ItemNotFoundException("AnnotationTypeCategory[id="+id+"]");
		return atcg;
	}

	/**
		Get a {@link ItemQuery} object configured to retrieve <code>AnnotationTypeCategory</code> 
		items.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<AnnotationTypeCategory> getQuery()
	{
		return new ItemQuery<AnnotationTypeCategory>(AnnotationTypeCategory.class);
	}
	

	AnnotationTypeCategory(AnnotationTypeCategoryData annotationTypeCategoryData) 
	{
		super(annotationTypeCategoryData);
	}
	
	@Override
	AnnotationTypeCategoryData getData()
	{
		return (AnnotationTypeCategoryData)getBasicData();
	}
	
	/*
	 	From the Identifiable interface
	 	-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	@Override
	public void setEntryDate(Date entryDate)
	{
		if (isInDatabase()) 
		{
			throw new PermissionDeniedException("Not allowed to change entryDate of " + this);
		}
		getData().setEntryDate(DateUtil.setDateOrToday(entryDate));
	}
	// -------------------------------------------

	/** 
		Gets a query that returns the annotation types that are members of this category
		@param itemType Optional, limit the annotation types to those
			that are enabled for the specified item
		@return An <code>ItemQuery</code> object
	*/
	public ItemQuery<AnnotationType> getAnnotationTypes(Item itemType)
	{
		ItemQuery<AnnotationType> query= AnnotationType.getQuery(itemType);
		query.joinPermanent(Hql.innerJoin("categories", Item.ANNOTATIONTYPECATEGORY.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.ANNOTATIONTYPECATEGORY.getAlias()), 
				Hql.entity(this)
			)
		);
		return query;
	}
}
