/*
	$Id: ArrayDesignBlock.java 6874 2015-04-17 12:02:22Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Set;

import net.sf.basedb.core.data.ArrayDesignBlockData;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;

/**
 
	@base.modified $Date: 2015-04-17 14:02:22 +0200 (Fri, 17 Apr 2015) $
	@author Enell
	@version 2.0
*/
public class ArrayDesignBlock 
	extends ChildItem
{

	/**
		The type of item represented by this class.
		@see Item#ARRAYDESIGNBLOCK
		@see #getType()
	*/
	public static final Item TYPE = Item.ARRAYDESIGNBLOCK;

	
	static ArrayDesignBlock getNew(DbControl dc, ArrayDesign arrayDesign, BlockInfo blockInfo) 
		throws BaseException
	{
		assert blockInfo != null : "blockInfo == null";
		ArrayDesignBlock adb = dc.newItem(ArrayDesignBlock.class);
		adb.setArrayDesign(arrayDesign);
		adb.setBlockNumber(blockInfo.getBlock());
		adb.setMetaGrid(blockInfo.getMetaGridX(), blockInfo.getMetaGridY());
		adb.setBlockSizeX(1);
		adb.setBlockSizeY(1);
		// Save in stateless session to get the ID immediately (even on Postgres)
		HibernateUtil.saveData(dc.getStatelessSession(), adb.getData());
		// Reattach the item to the "real" session so we can update properties on it
		HibernateUtil.lockData(dc.getHibernateSession(), adb.getData(), org.hibernate.LockOptions.NONE);
		return adb;
	}

	/**
		Get a <code>ArrayDesignBlock</code> object when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>ArrayDesignBlock</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have read
			permission to the items
		@throws BaseException If there is another error
	*/
	public static ArrayDesignBlock getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		ArrayDesignBlock adb = dc.loadItem(ArrayDesignBlock.class, id);
		if (adb == null) throw new ItemNotFoundException("ArrayDesignBlock[id=" + id + "]");
		return adb;
	}
	
	/**
		Get a query configured to retrieve array design blocks for 
		the specified array design. 
		
		@param arrayDesign The array design to retreive blocks for, null
			is not allowed
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<ArrayDesignBlock> getQuery(ArrayDesign arrayDesign)
	{
		if (arrayDesign == null) throw new InvalidUseOfNullException("arrayDesign");
		ItemQuery<ArrayDesignBlock> query = new ItemQuery<ArrayDesignBlock>(ArrayDesignBlock.class, null);
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("arrayDesign"),
				Hql.entity(arrayDesign)
			)
		);
		return query;
	}
	
	ArrayDesignBlock(ArrayDesignBlockData data)
	{
		super(data);
	}
	
	@Override
	ArrayDesignBlockData getData()
	{
		return (ArrayDesignBlockData)getBasicData();
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Array design blocks are always used by an array design.
	*/
	@Override
	public boolean isUsed() 
		throws BaseException
	{
		return true;
	}
	/**
		Get the array design this block is a part of.
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
		throws BaseException
	{
		Set<ItemProxy> using = super.getUsingItems();
		ArrayDesign item = getArrayDesign();
		using.add(new ItemProxy(item.getId(), item.getType()));
		return using;
	}
	/**
		WRITE and DELETE permissions are always denied.
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		denied |= Permission.deny(Permission.WRITE, Permission.DELETE);
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	@Override
	SharedData getSharedParent()
	{
		return getData().getArrayDesign();
	}
	// -------------------------------------------

	/**
		Get the <code>ArrayDesign</code> that is linked with this block.
		@return An <code>ArrayDesign</code> item
		@throws PermissionDeniedException If the logged in user doesn't have read
			permission to the item
		@throws BaseException This exception is thrown if there is another
			error
	*/
	public ArrayDesign getArrayDesign() 
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(ArrayDesign.class, getData().getArrayDesign());
	}
	
	/**
		Set the <code>ArrayDesign</code> for this block.
	*/
	private void setArrayDesign(ArrayDesign arrayDesign) 
		throws InvalidDataException
	{
		if (arrayDesign == null) throw new InvalidUseOfNullException("ArrayDesignBlock must have an arrayDesign.");
		getData().setArrayDesign(arrayDesign.getData());
	}

	/**
		Get the column coordinate of this block.
		@return the column coordinate
	*/
	public int getMetaGridX()
	{
		return getData().getMetaGridX();
	}
	
	/**
		Get the row coordinate of this block.
		@return the row coordinate
	*/
	public int getMetaGridY()
	{
		return getData().getMetaGridY();
	}

	/**
		Set the metagrid coordinates. If the coordinates are unknown the both should 
		be set to 0 otherwise the coordinates must be greater then 0.
		@param metaGridX the column coordinate
	*/
	private void setMetaGrid(int metaGridX, int metaGridY) 
		throws InvalidDataException
	{
		if (metaGridX < 0) throw new NumberOutOfRangeException("metaGridX", metaGridX, 0, false);
		if (metaGridY < 0) throw new NumberOutOfRangeException("metaGridY", metaGridY, 0, false);
		if (metaGridX == 0 ^ metaGridY == 0) throw new InvalidDataException("Both or neither of metarow and metacolumn must be 0 (unknown)");
		getData().setMetaGridX(metaGridX);
		getData().setMetaGridY(metaGridY);
	}
	
	/**
		Get the block number of this block.
		@return The number of this block.
	*/
	public int getBlockNumber()
	{
		return getData().getBlockNumber();
	}
	
	/**
		Set the block number of this block.
		@param blockNumber the block number of this block
		@throws InvalidDataException if the blocknumber is less then 1
	*/
	private void setBlockNumber(int blockNumber) 
		throws InvalidDataException
	{
		getData().setBlockNumber(blockNumber);
	}
	
	/**
		Get the number of columns in this block.
		@return the number of columns in this block.
	*/
	public int getBlockSizeX()
	{
		return getData().getBlockSizeX();
	}
	
	/**
		Set the number of columns in this block. Must be greater then 0.
		@param blockSizeX the number of rows in this block.
		@throws InvalidDataException if blockSizeX is less then 1
	*/
	public void setBlockSizeX(int blockSizeX) 
		throws InvalidDataException
	{
		if (blockSizeX < 1) throw new NumberOutOfRangeException("blockSizeX", blockSizeX, 1, false);
		getData().setBlockSizeX(blockSizeX);
	}
	
	/**
		Get the number of rows in this block.
		@return the number of rows in this block
	*/	
	public int getBlockSizeY()
	{
		return getData().getBlockSizeY();
	}
	
	/**
		Set the number of rows in this block. Must be greater then 0.
		@param blockSizeY the number of columns in this block.
		@throws InvalidDataException if blockSizeY is less then 1
	*/
	public void setBlockSizeY(int blockSizeY) 
		throws InvalidDataException
	{
		if (blockSizeY < 1) throw new NumberOutOfRangeException("blockSizeY", blockSizeY, 1, false);
		getData().setBlockSizeY(blockSizeY);
	}

	/**
		Get the distance from the left edge of the slide to the top left spot in the block.
		@return the distance from the edge
	*/
	public int getOriginX()
	{
		return getData().getOriginX();
	}
	
	/**
		Set the distance from the left edge of the slide to the top left spot in this block.
		@param originX the distance from the edge
		@throws InvalidDataException if originX is less then 0
	*/
	public void setOriginX(int originX) 
		throws InvalidDataException
	{
		if (originX < 0) throw new NumberOutOfRangeException("originX", originX, 0, false);
		getData().setOriginX(originX);
	}

	/**
		Get the distance from the top edge of the slide to the top left spot in the block.
		@return the distance from the edge
	*/
	public int getOriginY()
	{
		return getData().getOriginY();
	}
	
	/**
		Set the distance from the top edge of the slide to the top left spot in this block.
		@param originY the distance from the edge
		@throws InvalidDataException if originY is less then 0
	*/
	public void setOriginY(int originY) 
		throws InvalidDataException
	{
		if (originY < 0) throw new NumberOutOfRangeException("originY", originY, 0, false);
		getData().setOriginY(originY);
	}

	/**
		Get the x spacing between the spots in the slide.
		@return the spacing between the spots
	*/
	public int getSpacingX()
	{
		return getData().getSpacingX();
	}
	
	/**
		Set the x spacing between the spots in the slide.
		@param spacingX the spacing between the spots
		@throws InvalidDataException if spacingX is less then 0
	*/
	public void setSpacingX(int spacingX) 
		throws InvalidDataException
	{
		if (spacingX < 0) throw new NumberOutOfRangeException("spacingX", spacingX, 0, false);
		getData().setSpacingX(spacingX);
	}

	/**
		Get the y spacing between the spots in the slide.
		@return the spacing between the spots
	*/
	public int getSpacingY()
	{
		return getData().getSpacingY();
	}
	
	/**
		Set the y spacing between the spots in the slide.
		@param spacingY the spacing between the spots
		@throws InvalidDataException if spacingY is less then 0
	*/
	public void setSpacingY(int spacingY) 
		throws InvalidDataException
	{
		if (spacingY < 0) throw new NumberOutOfRangeException("spacingY", spacingY, 0, false);
		getData().setSpacingY(spacingY);
	}

	/**
		Get the expected diameter of spots on the slide.
		@return The expected diameter.
		@since 2.7
	*/
	public int getFeatureDiameter()
	{
		return getData().getFeatureDiameter();
	}
	
	/**
		Set the expected diameter of spots on the slide.
		@param diameter The expected diameter
		@throws InvalidDataException If the diameter is less then 0
	*/
	public void setFeatureDiameter(int diameter) 
		throws InvalidDataException
	{
		if (diameter < 0) throw new NumberOutOfRangeException("diameter", diameter, 0, false);
		getData().setFeatureDiameter(diameter);
	}
	
	/**
		Checks if this <code>ArrayDesignBlock</code> has any features.
		@return TRUE if the AraryDesignBlock has any fetures
	*/
	public boolean hasFeatures()
	{
		return getData().getArrayDesign().getHasFeatures();
	}
}
