/**
	$Id: Migration.java 7714 2019-05-21 11:37:09Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.StringWriter;
import java.io.Writer;
import java.nio.charset.StandardCharsets;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import net.sf.basedb.core.data.VirtualDbData;
import net.sf.basedb.core.dbengine.DbEngine;
import net.sf.basedb.core.dbengine.PostgresDbEngine;
import net.sf.basedb.core.dbengine.TableInfo;
import net.sf.basedb.core.dbengine.TableInfo.ColumnInfo;
import net.sf.basedb.core.dbengine.TableInfo.ForeignKeyInfo;
import net.sf.basedb.core.dbengine.TableInfo.IndexInfo;
import net.sf.basedb.core.dbengine.TableInfo.PrimaryKeyInfo;
import net.sf.basedb.core.migrate.ResultWriter;
import net.sf.basedb.core.migrate.ResultWriterFactory;
import net.sf.basedb.core.migrate.postgres.PostgresResultWriterFactory;
import net.sf.basedb.util.FileCopyRunnable;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.MD5;
import net.sf.basedb.util.RegexpFileFilter;
import net.sf.basedb.util.Values;

import org.hibernate.query.Query;
import org.hibernate.Session;
import org.hibernate.Transaction;
import org.hibernate.jdbc.Work;
import org.hibernate.mapping.Table;
import org.postgresql.copy.CopyManager;
import org.postgresql.core.BaseConnection;

/**
	Class for migrating a database to PostgreSQL. 
	The {@link #exportAll(String, ProgressReporter)} method should work on 
	both MySQL and PostgreSQL (but is only tested regularly with MySQL). 
	The export generates data that is suitable for importing into
	PostgreSQL using the COPY command. The {@link #importAll(String, ProgressReporter)} 
	only works on a PostgreSQL database.

	@author Nicklas
	@since 3.1
	@base.modified $Date: 2019-05-21 13:37:09 +0200 (Tue, 21 May 2019) $
 */
public final class Migration
	implements Work
{

	
	/**
		Export the current BASE database to data files in the
		directory on the given path.
		
		@param path A path to an existing directory
		@param progress A progress reporter 
	*/
	public static void exportAll(String path, ProgressReporter progress)
		throws SQLException, IOException
	{
		java.io.File destination = new java.io.File(path);
		if (!destination.exists())
		{
			throw new FileNotFoundException("Destination directory doesn't exists: " + path);
		}
		if (!destination.isDirectory())
		{
			throw new FileNotFoundException("Destination path is not a directory: " + path);
		}
		Application.start(false, true, false);
		Session session = null;
		Transaction tx = null;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			Migration exporter = new Migration(true, destination, session, progress);
			exporter.setExportCompressed(Values.getBoolean(Config.getString("migrate.export.compress"), false));
			exporter.setFetchSize(Config.getInt("migrate.export.fetch-size", 20000));
			HibernateUtil.doWork(session, exporter);
			HibernateUtil.commit(tx);
			tx = null;
		}
		finally
		{
			if (tx != null) HibernateUtil.rollback(tx);
			if (session != null) HibernateUtil.close(session);
			Application.stop();
		}
	}
	
	/**
		Import data to a BASE database. The database must exists, but
		not contain any tables.
		@param path The path to the directory with data files
		@param progress A progress reporter
	*/
	public static void importAll(String path, ProgressReporter progress)
		throws SQLException, IOException
	{
		java.io.File destination = new java.io.File(path);
		if (!destination.exists())
		{
			throw new FileNotFoundException("Import directory doesn't exists: " + path);
		}
		if (!destination.isDirectory())
		{
			throw new FileNotFoundException("Import path is not a directory: " + path);
		}
		Application.start(true, false, false);
		Session session = null;
		Transaction tx = null;
		Connection c = null;
		try
		{
			c = HibernateUtil.getPlainConnection();
			session = HibernateUtil.newSession(c);
			tx = HibernateUtil.newTransaction(session);
			Migration importer = new Migration(false, destination, session, progress);
			importer.setAnalyzeAfterImport(Values.getBoolean(Config.getString("migrate.import.analyze"), 
				HibernateUtil.getDbEngine().analyzeAfterBatchOperation()));
			importer.setDropPrimaryKeyBeforeImport(Values.getBoolean(Config.getString("migrate.import.drop-primary-key"), true));
			importer.setDropConstraintsBeforeImport(Values.getBoolean(Config.getString("migrate.import.drop-constraints"), true));
			HibernateUtil.doWork(session, importer);
			HibernateUtil.commit(tx);
			tx = null;
		}
		finally
		{
			if (tx != null) HibernateUtil.rollback(tx);
			if (session != null) HibernateUtil.close(session);
			if (c != null) c.close();
			Application.stop();
		}
	}
	
	private final boolean export;
	private final File directory;
	private final Session session;
	private final ProgressReporter progress;
	private final DecimalFormat numRowsFormat;
	
	private boolean exportCompressed;
	private int fetchSize;
	private boolean dropPrimaryKeyBeforeImport;
	private boolean dropConstraintsBeforeImport;
	private boolean analyzeAfterImport;
	
	private Migration(boolean export, File directory, Session session, ProgressReporter progress)
	{
		this.export = export;
		this.directory = directory;
		this.session = session;
		this.progress = progress;
		// This formatter is used for progress reporting
		DecimalFormatSymbols symbols = new DecimalFormatSymbols();
		symbols.setGroupingSeparator(',');
		this.numRowsFormat = new DecimalFormat("#,###",symbols);
		this.exportCompressed = false;
		this.fetchSize = 20000;
		this.dropPrimaryKeyBeforeImport = true;
		this.dropConstraintsBeforeImport = true;
		this.analyzeAfterImport = HibernateUtil.getDbEngine().analyzeAfterBatchOperation();
	}
	
	/*
		From Work interface
		-------------------
	*/
	@Override
	public void execute(Connection connection)
		throws SQLException
	{
		try
		{
			if (export)
			{
				doExport(connection);
			}
			else
			{
				doImport(connection);
			}
		}
		catch (IOException ex)
		{
			throw new BaseException(ex);
		}
	}
	// -------------------
	
	/**
		Should we drop the primary key before importing data? This can
		speed up the import. The primary key will be re-created after
		the import is complete.
		@param drop TRUE to drop (default), FALSE to leave it
	*/
	public void setDropPrimaryKeyBeforeImport(boolean drop)
	{
		this.dropPrimaryKeyBeforeImport = drop;
	}
	
	/**
		Should we drop the unique constraints and indexes before importing 
		data? This can speed up the import. The constraints will be re-created
		after the import is complete.
		@param drop TRUE to drop (default), FALSE to leave it
	*/
	public void setDropConstraintsBeforeImport(boolean drop)
	{
		this.dropConstraintsBeforeImport = drop;
	}

	/**
		Should we execute an analyis command after the import has
		been completed. This may result in improved performance when
		re-creating indexes and foreign keys. The default is
		whatever the current {@link DbEngine#analyzeAfterBatchOperation()}
		reports.
		@param analyze TRUE to analyze, FALSE to not
	*/
	public void setAnalyzeAfterImport(boolean analyze)
	{
		this.analyzeAfterImport = analyze;
	}

	/**
		Should the exporter compress the data files or leave them
		as is? Compressing can considerably reduce the disk space
		needed for migration. The import will automatically detect
		if the files are compressed or not.
		@param compress TRUE to compress the files, FALSE to not compress (default)
	*/
	public void setExportCompressed(boolean compress)
	{
		this.exportCompressed = compress;
	}

	/**
		Set the JDBC fetch size to use when fetching rows from the database.
		A higher value may improve performance but uses more memory. The
		default value is 20000.
		@param fetchSize The fetch size to use
		@see Statement#setFetchSize(int)
	*/
	public void setFetchSize(int fetchSize)
	{
		this.fetchSize = fetchSize;
	}
	
	/**
		Perform the export. In principle we support all databases that BASE supports,
		but it may depend on the actual underlying SQL column types. The export
		generates files that can be imported into PostgreSQL only.
	*/
	private void doExport(Connection connection)
		throws SQLException, IOException
	{
		
		DatabaseMetaData metaData = connection.getMetaData();
		DbEngine engine = HibernateUtil.getDbEngine();
		ResultWriterFactory rwFactory = new PostgresResultWriterFactory();

		String staticCatalog = connection.getCatalog();
		String staticSchema = connection.getSchema();
		String dynamicCatalog = Application.getDynamicCatalog();
		String dynamicSchema = Application.getDynamicSchema();

		// Load information about all tables in the static and dynamic databases
		progress.display(0, "Counting tables...\n");
		final int TICK_INTERVAL = 2000;
		long nextTick = System.currentTimeMillis() + TICK_INTERVAL;
		List<Table> staticTables = TableInfo.listTables(metaData, staticCatalog, staticSchema);
		progress.display(0, staticTables.size() + " tables in static database\n");

		List<Table> dynamicTables = TableInfo.listTables(metaData, dynamicCatalog, dynamicSchema);
		progress.display(0, dynamicTables.size() + " tables in dynamic database\n");
		progress.display(0, "Counting rows in all tables...");
		int totalTables = staticTables.size() + dynamicTables.size();
		
		// Get metadata (eg. column names+types) 
		// and count the number of data rows for each table
		List<TableInfo> tables = new ArrayList<TableInfo>();
		long approxTotalValues = 0;
		long approxTotalRows = 0;
		for (Table table : staticTables)
		{
			TableInfo info = new TableInfo(table, metaData, staticCatalog, staticSchema);
			long approxRows = info.getRowCount(engine, connection, true);
			approxTotalRows += approxRows;
			approxTotalValues += approxRows * info.getColumns().size();
			tables.add(info);
			if (System.currentTimeMillis() > nextTick)
			{
				progress.display(0, "Counting rows in all tables (" + tables.size() + "/" + totalTables + ")...");
				nextTick = System.currentTimeMillis()+TICK_INTERVAL;
			}
		}
		for (Table table : dynamicTables)
		{
			TableInfo info = new TableInfo(table, metaData, dynamicCatalog, dynamicSchema);
			long approxRows = info.getRowCount(engine, connection, true);
			approxTotalRows += approxRows;
			approxTotalValues += approxRows * info.getColumns().size();
			tables.add(info);
			if (System.currentTimeMillis() > nextTick || tables.size() == totalTables)
			{
				progress.display(0, "Counting rows in all tables (" + tables.size() + "/" + totalTables + ")...");
				nextTick = System.currentTimeMillis()+TICK_INTERVAL;
			}
		}
		
		progress.display(0, "Found approximate " + Values.formatNumber(approxTotalRows / 1000000f, 2) + " million data rows\n");
		
		// Export data for each table...
		long actualExportedValues = 0;
		long actualExportedRows = 0;
		Set<Integer> usedSqlTypes = new HashSet<Integer>();
		Thread compressThread = null;
		for (TableInfo table : tables)
		{
			String tableName = table.getTable().getName();
			long approxRows = table.getRowCount(engine, connection, true);
			int numColumns = table.getColumns().size();

			File exportFile = getDataFile(table, exportCompressed);
			if (exportFile.exists())
			{
				// Skip this table and move on to the next
				actualExportedValues += approxRows * numColumns;
				progress.display((int)((actualExportedValues * 100) / approxTotalValues), 
						tableName + ": skipped (file already exists)\n");
				continue; // with next table
			}
			
			progress.display((int)((actualExportedValues * 100) / approxTotalValues), 
					tableName + ": loading (approx. " + numRows(approxRows) + " rows)...");
			nextTick = System.currentTimeMillis()+TICK_INTERVAL;

			// Get columns in sorted order
			SortedSet<ColumnInfo> tmp =  new TreeSet<ColumnInfo>(table.getColumns());
			ColumnInfo[] columns = tmp.toArray(new ColumnInfo[tmp.size()]);
			
			// File to store all column names
			File columnsFile = getColumnsFile(table);
			// Create SELECT <columns> FROM <table>
			Writer writer = new OutputStreamWriter(FileUtil.getOutputStream(columnsFile), StandardCharsets.UTF_8);
			Statement st = connection.createStatement();
			st.setFetchSize(columns.length < 10 ? fetchSize * 5 : fetchSize);
			StringBuilder select = new StringBuilder("SELECT ");
			String separator = "";
			for (ColumnInfo c : columns)
			{
				select.append(separator);
				select.append(engine.getQuotedName(c.getName()));
				writer.append(separator);
				writer.append(c.getName());
				separator = ",";
				usedSqlTypes.add(c.getSqlType());
			}
			select.append(" FROM ").append(table.getFullQuotedTableName(engine));
			writer.flush();
			writer.close();

			// Execute query
			ResultSet rows = st.executeQuery(select.toString());
			
			// Wait for previous output file to be completely writting
			if (compressThread != null && compressThread.isAlive())
			{
				try
				{
					compressThread.join();
				}
				catch (InterruptedException ex)
				{}
			}
			
			// Create output file
			File tmpFile = new File(exportFile.getParentFile(), exportFile.getName() + ".tmp");
			OutputStream out = getOutputStream(tmpFile);
			
			// Setup streams and threading if using compression 
			if (exportCompressed)
			{
				PipedInputStream pipeIn = new PipedInputStream(1024*1024);
				compressThread = new Thread(new FileCopyRunnable(pipeIn, out));
				out = new PipedOutputStream(pipeIn);
				compressThread.start();
			}
			writer = new BufferedWriter(new OutputStreamWriter(out, StandardCharsets.UTF_8));
			
			// Setup writer factory and factories
			ResultWriter[] writers = new ResultWriter[columns.length];
			for (int i = 0; i < columns.length; ++i)
			{
				writers[i] = rwFactory.create(rows, i+1, writer, columns[i]);
			}

			long actualRows = 0;
			while (rows.next())
			{
				// Progress reporting...
				if (System.currentTimeMillis() > nextTick)
				{
					// Correct for actual vs. approximate mismatch
					if (actualRows > approxRows)
					{
						approxTotalValues += (actualRows - approxRows) * numColumns;
						approxRows = actualRows;
					}
					progress.display((int)(((actualExportedValues+actualRows*numColumns) * 100) / approxTotalValues), 
							tableName + ": exporting... (" + numRows(actualRows) + " of " + numRows(approxRows) + " rows done)");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
				
				// Write the current row
				int index = 0;
				while (index < columns.length)
				{
					if (index > 0) writer.write("\t");
					writers[index++].write();
				}
				writer.write("\n");
				
				// Keep track of exported rows
				actualRows++;
			}
			
			actualExportedValues += actualRows * numColumns;
			actualExportedRows += actualRows;
			// Correct for actual vs. approximate mismatch
			if (actualRows != approxRows)
			{
				// Decrease the approximate total row count
				approxTotalValues += (actualRows - approxRows) * numColumns;
			}

			// Done with this table...
			writer.flush();
			writer.close();
			tmpFile.renameTo(exportFile);

			progress.display((int)((actualExportedValues * 100) / approxTotalValues), 
					tableName + ": complete (" + numRows(actualRows) + " rows)                        \n");
		}
		
		// Wait for compress thread to finish
		if (compressThread != null && compressThread.isAlive())
		{
			try
			{
				compressThread.join();
			}
			catch (InterruptedException ex)
			{}			
		}
		progress.display(100, "Export complete, total row count: " + numRows(actualExportedRows) + "\n");
	}
	
	/**
		Perform the import. Importing is supported on PostgreSQL only.
	*/
	private void doImport(Connection connection)
		throws SQLException, IOException
	{
		DatabaseMetaData metaData = connection.getMetaData();
		DbEngine engine = HibernateUtil.getDbEngine();

		if (!(engine instanceof PostgresDbEngine))
		{
			throw new UnsupportedOperationException("Importing is supported for PostgreSQL only");
		}
		
		// Load information about all tables in the static and dynamic databases
		progress.display(0, "Loading tables...");
		final int TICK_INTERVAL = 1000;
		long nextTick = System.currentTimeMillis() + TICK_INTERVAL;
		
		String staticCatalog = connection.getCatalog();
		String staticSchema = connection.getSchema();
		List<Table> staticTables = TableInfo.listTables(metaData, staticCatalog, staticSchema);
		progress.display(0, staticTables.size() + " tables in static database");
		
		// Get metadata (eg. column names+types) 
		// and count the number of data rows for each table
		List<TableInfo> tables = new ArrayList<TableInfo>();
		long totalFileSize = 0;
		for (Table table : staticTables)
		{
			if (System.currentTimeMillis() > nextTick)
			{
				progress.append(".");
				nextTick = System.currentTimeMillis()+TICK_INTERVAL;
			}
			TableInfo info = new TableInfo(table, metaData, staticCatalog, staticSchema);
			// Check that data files exists
			File importFile = getDataFile(info);
			if (!importFile.exists())
			{
				throw new FileNotFoundException(importFile.getAbsolutePath());
			}
			File columnsFile = getColumnsFile(info);
			if (!columnsFile.exists())
			{
				throw new FileNotFoundException(columnsFile.getAbsolutePath());
			}
			
			totalFileSize += importFile.length();
			tables.add(info);
		}
		
		// And then the dynamic tables
		// Since they don't exists in the database yet, we need to scan the files
		Pattern dynamicFilePattern = Pattern.compile("d(\\d+)(.+)\\.data(\\.gz)?");
		List<File> dynamicDataFiles = FileUtil.findFiles(directory, 
				new RegexpFileFilter(dynamicFilePattern, null));
		progress.display(0, dynamicDataFiles.size() + " files with dynamic data");
		for (File f : dynamicDataFiles)
		{
			totalFileSize += f.length();
		}
		
		progress.display(0, "Found " + (tables.size()+dynamicDataFiles.size()) + " files with " + 
				Values.formatBytes(totalFileSize, 2) + " data"
				);
		

		// Sequence generators for special tables, all other tables use 'hibernate_sequence'
		Map<String, String> sequenceGenerators = new HashMap<String, String>();
		sequenceGenerators.put("Reporters", "reporter_sequence");
		sequenceGenerators.put("Features", "feature_sequence");
		for (RawDataType rdt : RawDataTypes.getRawDataTypes())
		{
			if (rdt.isStoredInDb())
			{
				sequenceGenerators.put(rdt.getTableName(), rdt.getEntityName()+"_sequence");
			}
		}
		// Keep track of the max id used in all tables (so that we can set the 'hibernate_sequence')
		int maxGlobalId = 0;
		
		// Import data for each table in the static database
		Statement st = connection.createStatement();
		// We are using ISO date format in the exported files
		st.executeUpdate("set datestyle TO 'iso'");
		
		
		// Drop foreign keys from all tables so that we can drop primary keys later
		progress.display(0, "Dropping foreign keys...");
		for (TableInfo table : tables)
		{
			String catalog = table.getTable().getCatalog();
			String schema = table.getTable().getSchema();
			String tableName = table.getTable().getName();
			for (ForeignKeyInfo fk : table.getForeignKeys())
			{
				String dropFkSql = engine.getDropForeignKeySql(catalog, schema, tableName, fk.getName());
				st.executeUpdate(dropFkSql);
			}
		}
		
		long totalImported = 0;
		for (TableInfo table : tables)
		{
			File importFile = getDataFile(table);

			if (importFile.length() > 0)
			{
				String tableName = table.getTable().getName();
				String fullTableName = table.getFullQuotedTableName(engine);
				progress.display((int)((totalImported * 80) / totalFileSize), 
						tableName + ": importing " + Values.formatBytes(importFile.length(), 2) + "...");
				
				// Load columns from file
				File columnsFile = getColumnsFile(table);

				// Import the file
				Integer maxId = importFromFile(table, importFile, columnsFile, engine, st, progress);
				
				if (maxId != null) // empty tables have no max id
				{
					if (sequenceGenerators.containsKey(tableName))
					{
						// Set the value for the special sequence generator
						String setSequenceSql = "SELECT setval('" + sequenceGenerators.get(tableName) + "', " + maxId + ")";
						st.execute(setSequenceSql);
					}
					else
					{
						// Global 'hibernate_sequence'
						if (maxGlobalId < maxId) maxGlobalId = maxId;
					}
				}
				
				// Done with this table...
				totalImported += importFile.length();
				progress.append("done");
			}
		}
		
		// Set the value for the 'hibernate_sequence' generator
		String sql = "SELECT setval('hibernate_sequence', " + maxGlobalId + ")";
		st.execute(sql);
		
		// Re-create foreign keys
		int progressOffset = (int)((totalImported * 80) / totalFileSize);
		progress.display(progressOffset, "Creating foreign keys...");
		List<String> fkSql = new ArrayList<String>();
		for (TableInfo table : tables)
		{
			String catalog = table.getTable().getCatalog();
			String schema = table.getTable().getSchema();
			String tableName = table.getTable().getName();
			for (ForeignKeyInfo fk : table.getForeignKeys())
			{
				fkSql.add(engine.getCreateForeignKeySql(catalog, schema, tableName, 
					fk.getName(), fk.getFkColumns(), fk.getRefName(), fk.getRefColumns()));
			}
		}
		int numStatements = fkSql.size();
		int numDone = 0;
		int lastPercent = 0;
		for (String createFkSql : fkSql)
		{
			int percent = progressOffset + (20*numDone) / numStatements;
			if (percent != lastPercent)
			{
				progress.display(percent, "--" + StringUtil.trimString(createFkSql, 55));
				lastPercent = percent;
			}
			else
			{
				progress.append(".");
			}
			st.executeUpdate(createFkSql);
			numDone++;
		}
		
		// We need information about file-only platforms when importing dynamic data
		RawDataTypes.initPlatforms(session); 

		// Import data for each table in the dynamic database
		String dynamicCatalog = Application.getDynamicCatalog();
		String dynamicSchema = Application.getDynamicSchema();
		for (File dynamicDataFile : dynamicDataFiles)
		{
			Matcher m = dynamicFilePattern.matcher(dynamicDataFile.getName());
			if (!m.matches())
			{
				// Should never happen since we use the same filter to match files at the beginning
				throw new AssertionError("File is not a dynamic data file: " + dynamicDataFile.getAbsolutePath());
			}
			
			int virtualDbId = Values.getInt(m.group(1));
			String tableType = m.group(2).toLowerCase();
			
			// Load the VirtualDb object that is managing the dynamic data table
			VirtualDbData vdbData = HibernateUtil.loadData(session, VirtualDbData.class, virtualDbId);
			VirtualDb vdb = new VirtualDb(vdbData);

			// Find the VirtualTable for this file
			VirtualTable vTable = null;
			String tableName = null;
			for (VirtualTable vt : VirtualTable.values())
			{
				tableName = vt.getTableName(vdb); 
				if (tableName.toLowerCase().endsWith(tableType))
				{
					vTable = vt;
					break;
				}
			}

			if (vTable == null)
			{
				throw new ItemNotFoundException("No virtual table found for file: " + dynamicDataFile.getAbsolutePath());
			}
			
			progress.display(20+(int)((totalImported * 80) / totalFileSize), 
					tableName + ": importing " + Values.formatBytes(dynamicDataFile.length(), 2) + "...");
			
			// Create the virtual table
			HibernateUtil.createVirtualTable(vdb, vTable);
			
			Table dynamicTable = new Table(tableName);
			dynamicTable.setCatalog(dynamicCatalog);
			dynamicTable.setSchema(dynamicSchema);
			TableInfo table = new TableInfo(dynamicTable, metaData, dynamicCatalog, dynamicSchema);

			File columnsFile = new File(directory, dynamicDataFile.getName().replaceFirst("\\.data(\\.gz)?", ".columns"));
			
			importFromFile(table, dynamicDataFile, columnsFile, engine, st, progress);
			
			// Done with this table...
			totalImported += dynamicDataFile.length();
			progress.append("done");
		}
		
	}
	
	private String numRows(long numRows)
	{
		return numRowsFormat.format(numRows);
	}
	
	/**
		Get the data file for the given table. Automatically detect if
		a compressed file is present or not.
	*/
	private File getDataFile(TableInfo table)
	{
		String name = table.getTable().getName().toLowerCase()+".data";
		File f = new File(directory, name + ".gz");
		if (!f.exists()) f = new File(directory, name);
		return f;
	}

	/**
		Get the data file for the given table.
	*/
	private File getDataFile(TableInfo table, boolean compress)
	{
		return new File(directory, table.getTable().getName().toLowerCase()+".data" + (compress ? ".gz" : ""));
	}

	/**
		Get the columns file for the given table.
	*/
	private File getColumnsFile(TableInfo table)
	{
		return new File(directory, table.getTable().getName().toLowerCase()+".columns");
	}
	
	/**
		Get a stream for writing to the given file. If the file name ends
		with .gz it is automatically wrapped with a GZIPOutputStream.
	*/
	private OutputStream getOutputStream(File file)
		throws IOException
	{
		OutputStream out = FileUtil.getOutputStream(file);
		if (file.getName().contains(".gz"))
		{
			out = new GZIPOutputStream(out, 8192);
		}
		return out;
	}
	
	/**
		Get a stream for reading from the given file. If the file name ends
		with .gz it is automatically wrapped with a GZIPInputStream.
	*/
	private InputStream getInputStream(File file)
		throws IOException
	{
		InputStream in = FileUtil.getInputStream(file);
		if (file.getName().endsWith(".gz"))
		{
			in = new GZIPInputStream(in, 8192);
		}
		return in;
	}


	/**
		Import data to the given table. 
		@return The max id if 'id' column exists, null otherwise
	*/
	private Integer importFromFile(TableInfo table, File dataFile, File columnsFile, DbEngine engine, 
		Statement st, ProgressReporter progress)
		throws SQLException, IOException
	{
		String catalog = table.getTable().getCatalog();
		String schema = table.getTable().getSchema();
		String tableName = table.getTable().getName();
		
		// Drop PRIMARY KEY if there is one
		PrimaryKeyInfo pk = table.getPrimaryKey();
		String pkName = null;
		if (dropPrimaryKeyBeforeImport && pk != null)
		{
			pkName = pk.getName();
			String dropPkSql = engine.getDropPrimaryKey(catalog, schema, tableName, pkName);
			//System.out.println("pk-drop:"+dropPkSql);
			st.executeUpdate(dropPkSql);
		}
		
		// Drop indexes and unique constraints
		if (dropConstraintsBeforeImport)
		{
			for (IndexInfo index : table.getIndexes())
			{
				// PostgreSQL also report primary key as an index
				boolean isPkIndex = pkName != null && pkName.equals(index.getName());
				if (!isPkIndex)
				{
					String dropIndexSql = engine.getDropIndexSql(catalog, schema, tableName, index.getName(), index.isUnique());
					//System.out.println("index-drop:"+dropIndexSql);
					st.executeUpdate(dropIndexSql);
				}
			}
		}
		
		StringWriter sin = new StringWriter((int)columnsFile.length());
		FileUtil.copy(new InputStreamReader(FileUtil.getInputStream(columnsFile), StandardCharsets.UTF_8), sin);
		String[] columns = sin.toString().split(",");
		
		// Create COPY (<columns>) FROM '<filename>'
		StringBuilder copy = new StringBuilder("COPY ");
		copy.append(table.getFullQuotedTableName(engine));
		copy.append(" (");
		String separator = "";
		boolean hasIdColumn = false;
		for (String col : columns)
		{
			col = col.trim();
			if (table.findColumn(col) == null)
			{
				throw new IOException("Column '" + col + "' is not found in table '" + tableName + 
					"' but was exported to file: " + dataFile.getAbsolutePath());
			}
			copy.append(separator);
			copy.append(engine.getQuotedName(col));
			separator = ",";
			hasIdColumn |= "id".equals(col);
		}
		//copy.append(") FROM '").append(dataFile.getAbsolutePath() + "'");
		copy.append(") FROM STDIN");
		copy.append(" WITH ENCODING 'UTF8'");

		// Setup streams
		InputStream in = getInputStream(dataFile);
		
		// Use separate thread for decompression
		if (dataFile.getName().endsWith(".gz"))
		{
			PipedOutputStream out = new PipedOutputStream();
			Thread decompressThread = new Thread(new FileCopyRunnable(in, out));
			in = new PipedInputStream(out, 1024*1024);
			decompressThread.start();
		}
		
		// Import the data
		ProgressInputStream pin = new ProgressInputStream(in, progress);
		try
		{
			CopyManager copyManager = new CopyManager((BaseConnection)st.getConnection());
			long rows = copyManager.copyIn(copy.toString(), pin);
			pin.append("; "+numRows(rows) + " rows imported");
		}
		finally
		{
			FileUtil.close(pin);
		}
		//st.executeUpdate(copy.toString());

		// Analyze the table data (will increase performance)
		if (analyzeAfterImport)
		{
			progress.append("; analyzing...");
			String analyzeSql = engine.getAnalyzeTableSql(catalog, schema, tableName);
			//System.out.println("analyze-sql:" + analyzeSql);
			if (analyzeSql != null) st.executeUpdate(analyzeSql);
		}

		// Re-create primary key
		if (dropPrimaryKeyBeforeImport && pk != null)
		{
			String createPkSql = engine.getCreatePrimaryKeySql(catalog, schema, tableName, pk.getName(), pk.getColumns());
			progress.append("; creating constraints...");
			//System.out.println("pk-sql:" + createPkSql);
			st.executeUpdate(createPkSql);
		}

		// Re-create indexes and unique constraints
		if (dropConstraintsBeforeImport)
		{
			for (IndexInfo index : table.getIndexes())
			{
				boolean isPkIndex = pkName != null && pkName.equals(index.getName());
				if (!isPkIndex)
				{
					String createIndexSql = engine.getCreateIndexSql(catalog, schema, tableName, index.getName(), index.getColumns(), index.isUnique());
					progress.append("..");
					//System.out.println("index-sql:" + createIndexSql);
					st.execute(createIndexSql);
				}
			}
		}

		// Get max id (if 'id' column exists)
		Integer maxId = null;
		if (hasIdColumn)
		{
			progress.append("..");
			String sql = "SELECT max(id) FROM " + table.getFullQuotedTableName(engine);
			Query<Integer> query = HibernateUtil.createSqlQuery(session, sql, Integer.class);
			maxId = HibernateUtil.loadData(query);
		}
		return maxId;
	}

	
	class ProgressInputStream
		extends FilterInputStream
	{
		private final ProgressReporter progress;
		private long numLines;
		private long nextTick;
		private long TICK_INTERVAL = 2000;
		private String lastMsg;
		
		public ProgressInputStream(InputStream parent, ProgressReporter progress) 
		{
			super(parent);
			this.progress = progress;
		}

		/*
			From the InputStream class
			--------------------------
		*/
		@Override
		public int read(byte[] buffer)
			throws IOException
		{
			return read(buffer, 0, buffer.length);
		}

		@Override
		public int read(byte[] buffer, int offset, int length)
			throws IOException 
		{
			int numRead = super.read(buffer, offset, length);
			if (numRead > 0)
			{
				int end = offset+numRead;
				for (int i = offset; i < end; ++i)
				{
					if (buffer[i] == '\n') numLines++;
				}
				if (System.currentTimeMillis() > nextTick)
				{
					append("; " + numRows(numLines) + " rows...");
					nextTick = System.currentTimeMillis() + TICK_INTERVAL;
				}
			}
			return numRead;
		}
		// ------------------------------
		
		public void append(String msg)
		{
			if (lastMsg != null)
			{
				progress.append(MD5.leftPad("", '\b', lastMsg.length()));
			}
			lastMsg = msg;
			progress.append(msg);
		}
	}

}


