/*
	$Id: Project.java 7605 2019-02-26 10:10:15Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.Transactional.Action;
import net.sf.basedb.core.data.AnnotationSetData;
import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.data.ItemParameterValueData;
import net.sf.basedb.core.data.ItemSubtypeData;
import net.sf.basedb.core.data.NameableData;
import net.sf.basedb.core.data.ProjectData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.data.SubtypableData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.EntityQuery;
import net.sf.basedb.core.query.ResultList;
import net.sf.basedb.util.EqualsHelper;

import java.util.ArrayList;
import java.util.Date;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Collections;

import org.hibernate.Hibernate;

/**
	This class is used to represent projects. A project is used to
	collect items of particular interest in one place. Users and
	groups can also be granted access to projects, allowing everyone easy
	access to shared items within the project.
	<p>
	While working with BASE it is possible to set one project as the
	active project. This enables some useful functionality in the core:
	<ul>
	<li>Every {@link Shareable} item the user creates is automatically
		shared to the project. 
	<li>All queries defaults to only return items that are shared
		to the project.
	</ul>
	<p>
	It is also possible to add an item to other
	projects, or remove it completely from all projects. Every member of
	a project has a special <code>project permission</code> which gives the highest
	permission for accessing items in the project. Normally this permission
	is {@link Permission#USE}. One person might be assigned
	administrator of the project and given higher permissions.
	<p>
	Items may also be given a permission within the project. The normal
	is to give full access, but it may also be set to for example
	{@link Permission#READ}. Then, not even the project administrator
	will get more than read permissions to the item. This feature can
	for example be used when a project needs to access data from
	a user that is not a member of the project.

	@author Nicklas
	@version 2.0
	@see Role
	@see Project
	@see SessionControl#setActiveProject(Project)
	@base.modified $Date: 2019-02-26 11:10:15 +0100 (Tue, 26 Feb 2019) $
*/
public class Project
	extends OwnedItem
	implements Nameable, Removable, Registered, Annotatable
{
	/**
		The type of item represented by this class.
		@see Item#PROJECT
		@see #getType()
	*/
	public static final Item TYPE = Item.PROJECT;

	/**
		This filter will limit a query to only return projects
		where the logged in user is a member or owner unless the logged in user
		has generic read permission.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = new QueryRuntimeFilterImpl();
	
	/**
		Create a new <code>Project</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@return The new <code>Project</code> item
		@throws BaseException If there is an error
	*/
	public static Project getNew(DbControl dc)
		throws BaseException
	{
		Project p = dc.newItem(Project.class);
		p.setName("New project");
		p.getData().setEntryDate(new Date());
		p.setAutoPermission(EnumSet.of(Permission.READ, Permission.USE, Permission.WRITE, Permission.DELETE));
		return p;
	}

	/**
		Get a <code>Project</code> item when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>Project</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Project getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Project p = dc.loadItem(Project.class, id);
		if (p == null) throw new ItemNotFoundException("Project[id="+id+"]");
		return p;
	}

	/**
		Get a query configured to retrieve projects. If the logged in user
		doesn't have generic permission to all projects, only projects
		the user has access to are returned.

		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Project> getQuery()
	{
		return new ItemQuery<Project>(Project.class, RUNTIME_FILTER);
	}

	Project(ProjectData projectData)
	{
		super(projectData);
	}

	@Override
	ProjectData getData()
	{
		return (ProjectData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	@Override
	public String getName()
	{
		return getData().getName();
	}
	@Override
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	@Override
	public String getDescription()
	{
		return getData().getDescription();
	}
	@Override
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	@Override
	public boolean isRemoved()
	{
		return getData().getRemovedBy() != null;
	}
	@Override
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemovedBy(removed ? getSessionControl().getLoggedInUserId() : null);
	}
	@Override
	public User getRemovedBy()
		throws PermissionDeniedException, ItemNotFoundException 
	{
		Integer userId = getData().getRemovedBy();
		return userId == null ? null : User.getById(getDbControl(), userId);
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	@Override
	public void setEntryDate(Date entryDate)
	{
		if (isInDatabase()) 
		{
			throw new PermissionDeniedException("Not allowed to change entryDate of " + this);
		}
		getData().setEntryDate(DateUtil.setDateOrToday(entryDate));
	}
	// -------------------------------------------

	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	@Override
	public AnnotationSet getAnnotationSet()
		throws PermissionDeniedException, BaseException
	{
		checkBatchAnnotatableUsage(true, false);
		DbControl dc = getDbControl();
		AnnotationSet as = dc.getItem(AnnotationSet.class, getData().getAnnotationSet(), this);
		if (as == null)
		{
			as = AnnotationSet.getNew(dc, this);
			getData().setAnnotationSet(as.getData());
			dc.saveItemIf(this, as, false);
		}
		return as;
	}
	@Override
	public boolean isAnnotated()
	{
		return getData().getAnnotationSet() != null;
	}
	@Override
	public void removeAnnotations()
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		AnnotationSetData as = getData().getAnnotationSet();
		if (as != null)
		{
			HibernateUtil.deleteData(getDbControl().getHibernateSession(), as);
			getData().setAnnotationSet(null);
		}
	}
	/**
		Default implementation returns null. Should be overriden by subclasses
		that has protocols.
		@return Always null
	*/
	@Override
	public Protocol getProtocol()
	{
		return null;
	}
	
	/**
		@return Always null
	 */
	@Override
	public Set<Annotatable> getAnnotatableParents()
		throws BaseException
	{
		return null;
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	@Override
	void onBeforeCommit(Action action) 
		throws NotLoggedInException, BaseException 
	{
		super.onBeforeCommit(action);
		if (action == Action.DELETE)
		{
			SessionControl sc = getSessionControl();
			if (sc.getActiveProjectId() == getId())
			{
				sc.setActiveProject(null);
			}
		}
	}
	/**
		If the logged in user is a member of this project, read or use 
		permission is granted.
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		int projectPermission = getSessionControl().getProjectPermission(this);
		granted |= projectPermission;
		super.initPermissions(granted, denied);
	}
	/**
		Make sure plug-ins always has at least read access to the active project.
		@since 2.9
	*/
	@Override
	PluginPermission getPluginPermissions()
	{
		PluginPermission pp = super.getPluginPermissions();
		if (getId() == getSessionControl().getActiveProjectId())
		{
			pp = new PluginPermission(pp.granted | Permission.grant(Permission.READ), pp.denied);
		}
		return pp;
	}
	// -------------------------------------------

	/**
		The permissions to use when automatically sharing new items to
		the project when it is the active project. The default is READ, 
		USE, WRITE and DELETE. Note that this setting is ignored when
		a permission template is used ({@link #getPermissionTemplate()}).
		@since 2.10
	*/
	public Set<Permission> getAutoPermission()
	{
		return Permission.fromInt(getData().getAutoPermission());
	}
	
	/**
		Set the default permissions to use when automatically sharing
		new item the project when it is the active project. Note that
		this setting is ignored when a permission template is used 
		({@link #getPermissionTemplate()}).
		@param permissions The permissions, must not be null or empty
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@throws InvalidDataException If the permissions are null or empty
		@since 2.10
	*/
	public void setAutoPermission(Set<Permission> permissions)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (permissions == null || permissions.isEmpty())
		{
			throw new InvalidUseOfNullException("permissions");
		}
		getData().setAutoPermission(Permission.grant(permissions));
	}

	/**
		Get the permission template that is used to share new items when this project
		is active. Note that if a permission template is used the new items are only
		shared to the active project if the project is part of the template.
		@return A permission template or null if no template has been specified
		@since 2.16
	*/
	public PermissionTemplate getPermissionTemplate()
	{
		return getDbControl().getItem(PermissionTemplate.class, getData().getPermissionTemplate());
	}
	
	/**
		Set the permission template that should be used to share new items when
		this project is active.
		@param template A permission template, or null to not use any template
		@since 2.16
	*/
	public void setPermissionTemplate(PermissionTemplate template)
	{
		checkPermission(Permission.WRITE);
		getData().setPermissionTemplate(template == null ? null : template.getData());
	}
	
	/**
		Grant a user permissions to this project. Use an empty <code>Set</code> or null
		to remove all permissions for the user.

		@param user The <code>User</code>
		@param permissions The permissions to grant, or null to revoke all permissions
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#WRITE} permission for the project
		@throws InvalidDataException If the user is null
		@see Permission
	*/
	public void setPermissions(User user, Set<Permission> permissions)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (user == null) throw new InvalidUseOfNullException("user");
		if (permissions == null || permissions.isEmpty())
		{
			getData().getUsers().remove(user.getData());
		}
		else
		{
			getData().getUsers().put(user.getData(), Permission.grant(permissions));
		}
	}
	
	/**
		Get the permissions for a user in this project.
		@param user The <code>User</code> for which we want to get the permission
		@return A <code>Set</code> containing the granted permissions, or an
			empty set if no permissions have been granted
		@throws InvalidDataException If the user is null
		@see Permission
	*/
	public Set<Permission> getPermissions(User user)
		throws InvalidDataException
	{
		if (user == null) throw new InvalidUseOfNullException("user");
		return Permission.fromInt(getData().getUsers().get(user.getData()));
	}

	/**
		Get a query that returns the users that
		are members of this project. This query excludes users that the logged 
		in user doesn't have permission to read.
	 	@return A restricted item query. 
	 	@throws BaseException If the query couldn't be created or configured. 
		@see User#getQuery()
	*/
	public ItemQuery<User> getUsers()
		throws BaseException
	{
		ItemQuery<User> query = User.getQuery();
		query.joinPermanent(Hql.innerJoin("projects", Item.PROJECT.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.PROJECT.getAlias()), 
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
		Grant a group permissions to this project. Use an empty <code>Set</code> or null
		to remove all permissions for the group. To share to the {@link Group#EVERYONE}
		group {@link Permission#SHARE_TO_EVERYONE} is required.

		@param group The <code>Group</code>
		@param permissions The permissions to grant, or null to revoke all permissions
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#WRITE} permission for the project
		@throws InvalidDataException If the group is null
		@throws BaseException If there is another error
		@see Permission
	*/
	public void setPermissions(Group group, Set<Permission> permissions)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (group == null) throw new InvalidUseOfNullException("group");
		if (permissions == null || permissions.isEmpty())
		{
			getData().getGroups().remove(group.getData());
		}
		else
		{
			if ((group.getId() == SystemItems.getId(Group.EVERYONE)) && 
				!getSessionControl().hasSystemPermission(Permission.SHARE_TO_EVERYONE))
			{
				throw new PermissionDeniedException("Not allowed to share to the EVERYONE group.");
			}
			getData().getGroups().put(group.getData(), Permission.grant(permissions));
		}
	}
	
	/**
		Get the permissions for a group in this project.
		@param group The <code>Group</code> for which we want to get the permission
		@return An <code>EnumSet</code> containing the granted permissions, or an
			empty set if no permissions have been granted
		@throws InvalidDataException If the group is null
		@see Permission
	*/
	public Set<Permission> getPermissions(Group group)
		throws InvalidDataException
	{
		if (group == null) throw new InvalidUseOfNullException("usgroup");
		return Permission.fromInt(getData().getGroups().get(group.getData()));
	}
	
	/**
		Get a query that returns the groups that
		are members of this project. This query excludes groups that the logged 
		in user doesn't have permission to read.

		@see Group#getQuery()
	*/
	public ItemQuery<Group> getGroups()
	{
		ItemQuery<Group> query = Group.getQuery();
		query.joinPermanent(Hql.innerJoin("projects", Item.PROJECT.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.PROJECT.getAlias()), 
				Hql.entity(this)
			)
		);
		return query;
	}

	/**
		@see #getItems(Set, int, int, Permission, Set, Restriction)
		@since 2.3.1
	*/
	public ResultList<Shareable> getItems(Item itemType, int firstItem, int maxItems, 
			Permission permission, Set<Include> include)
	{
		return getItems(itemType == null ? null : Collections.singleton(itemType), 
			firstItem, maxItems, permission, include, null);
	}
	
	/**
		Load the items in a project. If this project is the active project it will load
		all items in the project. If this project isn't the active project it will
		only load items that the logged in user owns.

		@param itemTypes Limit the list to only return items of the specified types,
			use null or an empty set to return items of all types
		@param firstItem The index of the first item to return (0-based)
		@param maxItems The maximum number of items to return, or 0 to return all items
		@param permission The permission the logged in user must have on the item
		@param include Options for which items that should be included/excluded from the
			result, or null to include all non-removed items
		@param restriction An optional restriction that is applied to the
			queries to further reduce the list
		@return A list containing shareable items
		@see SessionControl#setActiveProject(Project)
		@since 2.15
	*/
	public ResultList<Shareable> getItems(Set<Item> itemTypes, int firstItem, int maxItems, 
		Permission permission, Set<Include> include, Restriction restriction)
	{
		if (include == null) include = EnumSet.of(Include.NOT_REMOVED, Include.MINE, Include.OTHERS);
		if (itemTypes == null || itemTypes.size() == 0) itemTypes = Metadata.getShareableItems();
		
		DbControl dc = getDbControl();
		SessionControl sc = dc.getSessionControl();
		org.hibernate.Session session = dc.getHibernateSession();

		// First task is to get all project keys sharing items to this project
		Set<Integer> projectKeys = null;
		if (this.getId() != sc.getActiveProjectId())
		{
			// Load the project keys for the current project
			org.hibernate.query.Query<Integer> query = HibernateUtil.getPredefinedQuery(session, "GET_PROJECTKEYIDS_FOR_PROJECT", Integer.class);
			/*
				SELECT pk.keyId
				FROM ProjectKeys pk
				WHERE pk.projectId = :projectId
			*/
			query.setParameter("projectId", this.getId(), TypeWrapper.H_INTEGER);
			projectKeys = new HashSet<Integer>(HibernateUtil.loadList(query, sc));
			
			include.remove(Include.OTHERS);
		}
		else
		{
			// The project keys for the active project are already loaded
			projectKeys = sc.getProjectKeys();
			
			// If a permission is specified we must filter on that too...
			if (permission != null && permission != Permission.READ)
			{
				Set<Integer> permissionKeys = sc.getProjectKeys(permission);
				if (permissionKeys == null || permissionKeys.size() == 0) 
				{
					permissionKeys = QueryRuntimeFilterFactory.ZERO_SET;
				}
				org.hibernate.Filter permissionFilter = HibernateUtil.enableFilter(session, "ownedByOrInProject");
				permissionFilter.setParameter("owner", sc.getLoggedInUserId());
				permissionFilter.setParameterList("projectKeys", permissionKeys);
			}
		}
		
		// Filter that excludes items not shared to this project
		org.hibernate.Filter keyFilter = HibernateUtil.enableFilter(session, "inProject");
		// The list must contain at least one element
		if (projectKeys == null || projectKeys.size() == 0) 
		{
			projectKeys = QueryRuntimeFilterFactory.ZERO_SET;
		}
		keyFilter.setParameterList("projectKeys", projectKeys);
		keyFilter.setParameter("owner", sc.getLoggedInUserId());
		
		// Add filters for 'removed' option
		if (!(include.contains(Include.REMOVED) && include.contains(Include.NOT_REMOVED)))
		{
			org.hibernate.Filter removedFilter = HibernateUtil.enableFilter(session, "isRemoved");
			removedFilter.setParameter("removed", include.contains(Include.REMOVED));
		}
		// Add filter for 'owned by me' and 'owned by others' options
		if (include.contains(Include.MINE) && include.contains(Include.OTHERS))
		{
			// Do nothing, all items are included
		}
		else if (include.contains(Include.MINE))
		{
			org.hibernate.Filter ownerFilter = HibernateUtil.enableFilter(session, "ownedBy");
			ownerFilter.setParameter("owner", sc.getLoggedInUserId());
		}
		else if (include.contains(Include.OTHERS))
		{
			org.hibernate.Filter otherFilter = HibernateUtil.enableFilter(session, "notOwnedBy");
			otherFilter.setParameter("owner", sc.getLoggedInUserId());			
		}
		else
		{
			org.hibernate.Filter denyAllFilter = HibernateUtil.enableFilter(session, "denyAll");
		}
		
		boolean useLimit = firstItem >= 0 && maxItems > 0;
		List<ShareableData> inProject = new ArrayList<ShareableData>();
		List<ItemProxy> proxies = new ArrayList<ItemProxy>();
		
		for (Item item : itemTypes)
		{
			if (sc.hasPermission(Permission.DENIED, item)) continue;
			
			if (ShareableData.class.isAssignableFrom(item.getDataClass()))
			{
				// Query to load all "shareable" items of a specific item type
				AbstractEntityQuery hql = new AbstractEntityQuery(item, null, false){};
				if (restriction != null) hql.restrict(restriction);
				if (NameableData.class.isAssignableFrom(item.getDataClass()))
				{
					hql.order(Orders.asc(Hql.property("name")));
				}
				hql.order(Orders.asc(Hql.property("id")));
				/*
					SELECT item[.id]
					FROM <class> item
					[WHERE restriction]
				*/
				if (useLimit)
				{
					// Load proxies if we are just going to load a part of the items (it is a lot quicker!)
					for (Integer id : HibernateUtil.loadList(hql.getIdHqlQuery(dc), sc))
					{
						proxies.add(new ItemProxy(id, item));
					}
				}
				else
				{
					// Load all items
					inProject.addAll(HibernateUtil.loadList(hql.getMainHqlQuery(dc, ShareableData.class), sc));
				}
			}
		}
			
		// Disable filters
		HibernateUtil.disableFilter(session, "ownedByOrInProject");
		HibernateUtil.disableFilter(session, "inProject");
		HibernateUtil.disableFilter(session, "ownedBy");
		HibernateUtil.disableFilter(session, "notOwnedBy");
		HibernateUtil.disableFilter(session, "isRemoved");
		HibernateUtil.disableFilter(session, "denyAll");
		
		List<Shareable> items = new ArrayList<Shareable>(useLimit ? maxItems : inProject.size());
		int totalItems = 0;
		if (useLimit)
		{
			totalItems = proxies.size();
			int lastItem = Math.min(firstItem + maxItems, totalItems);
			for (int index = firstItem; index < lastItem; index++)
			{
				try
				{
					BasicItem item = proxies.get(index).getItem(dc);
					items.add((Shareable)item);
				}
				catch (PermissionDeniedException ex)
				{
					// We do not correct for permissions since that would make it impossible to
					// implement paging. We accept that some pages may contain fewer items.
				} 
			}
		}
		else
		{
			totalItems = inProject.size();
			for (ShareableData sd : inProject)
			{
				try
				{
					BasicItem item = dc.getItem(BasicItem.class, (BasicData)sd);
					items.add((Shareable)item);
				}
				catch (PermissionDeniedException ex)
				{
					totalItems--;
				}
			}
		}

		return new UnmodifiableResultList<Shareable>(items, totalItems);
	}

	/**
		Add an item as a default project item. If the item is already
		a default item of the project, nothing is done.
		
		@param defaultItem The item to make default
		@since 3.0
	 */
	public void addDefaultItem(BasicItem defaultItem)
	{
		checkPermission(Permission.WRITE);
		if (defaultItem == null) throw new InvalidUseOfNullException("defaultItem");
		
		ItemParameterValueData defaultItems = getData().getDefaultItems();
		if (defaultItems == null)
		{
			// Create new placeholder object
			defaultItems = new ItemParameterValueData();
			getData().setDefaultItems(defaultItems);
		}
		else if (defaultItems.getValues().indexOf(defaultItem.getData()) >= 0)
		{
			// Check that it is not already added
			return;
		}
		defaultItems.getValues().add(defaultItem.getData());
	}
	
	/**
		Replace an existing default that has the same class and subtype
		as the given item with the given item. 
		@param newItem The new item to make default
		@since 3.2
	*/
	public void replaceDefaultItem(BasicItem newItem)
	{
		checkPermission(Permission.WRITE);
		if (newItem == null) throw new InvalidUseOfNullException("defaultItem");
	
		ItemParameterValueData defaultItems = getData().getDefaultItems();
		if (defaultItems != null)
		{
		
			Class<?> dataClass = newItem.getType().getDataClass();
			ItemSubtypeData subtype = newItem instanceof Subtypable ? 
					((SubtypableData)newItem.getData()).getItemSubtype() : null;
			Iterator<BasicData> it = defaultItems.getValues().iterator();
			while (it.hasNext())
			{
				BasicData data = it.next();
				if (dataClass.isInstance(data))
				{
					// Check if item has been removed
					try
					{
						Hibernate.initialize(data);
					}
					catch (org.hibernate.ObjectNotFoundException ex)
					{
						it.remove();
						continue;
					}
					
					// Found an item of the correct item type... get it's subtype
					ItemSubtypeData otherSubtype = null;
					if (data instanceof SubtypableData)
					{
						otherSubtype = ((SubtypableData)data).getItemSubtype();
					}
					
					// If the subtypes are null or equal we have found a match which should be removed
					if (EqualsHelper.equals(subtype, otherSubtype)) 
					{
						it.remove();
					}
				}
			}
		}
		addDefaultItem(newItem);
	}

	
	/**
		Remove an item from the list of default items for this project.
		If the given item is not a default item nothing is done.
	
		@param defaultItem The item to remove
		@since 3.0
	*/
	public void removeDefaultItem(BasicItem defaultItem)
	{
		checkPermission(Permission.WRITE);
		if (defaultItem == null) return;
		
		ItemParameterValueData defaultItems = getData().getDefaultItems();
		if (defaultItems == null) return;
		
		defaultItems.getValues().remove(defaultItem.getData());
	}
	
	/**
		Find the default items of the given item type. If the strict parameter
		is true, the search will only match items without a subtype. If strict
		is false, the search ignore the subtype.
		
		@param dc A DbControl for database access
		@param itemType The type of item to find
		@return A list with the items that was found (empty if no item was found)
		@since 3.0
	*/
	@SuppressWarnings("unchecked")
	public <T extends BasicItem> List<T> findDefaultItems(DbControl dc, Item itemType, boolean strict)
	{
		if (itemType == null) throw new InvalidUseOfNullException("itemType");
		List<BasicData> defaultData = findAllDefaultData(itemType, null, strict);
		List<T> defaultItems = new ArrayList<>(defaultData.size());
		for (BasicData data : defaultData)
		{
			defaultItems.add((T)dc.getItem(BasicItem.class, data));
		}
		return defaultItems;
	}
	
	/**
		Find the default (subtypable) items that has the given subtype. If the
		strict parameter is true, an exact match must be found, otherwise
		the search is repeated to find an item without a subtype. The search
		will never match an item that has a different subtype.
		
		@param dc A DbControl for database access
		@param subtype The requested subtype
		@param strict TRUE to strictly match the subtype, FALSE to also
			search for items without subtype
		@return A list with the items that was found (empty if no item was found)
		@since 3.0
	*/
	@SuppressWarnings("unchecked")
	public <T extends BasicItem> List<T> findDefaultItems(DbControl dc, ItemSubtype subtype, boolean strict)
	{
		if (subtype == null) throw new InvalidUseOfNullException("subtype");
		List<BasicData> defaultData = findAllDefaultData(subtype.getMainItemType(), subtype.getData(), strict);
		List<T> defaultItems = new ArrayList<T>(defaultData.size());
		for (BasicData data : defaultData)
		{
			defaultItems.add((T)dc.getItem(BasicItem.class, data));
		}
		return defaultItems;
	}
	
	/**
		Utility method for finding default items of the given relatedType.
		If a subtype is given this method first try to find the related
		subtype for the given relatedType. If no subtype is given or if no
		related subtype is found, this method behaves as if {@link #findDefaultItems(DbControl, Item, boolean)}
		was called with strict=true (eg. it will only find items that doesn't have a subtype).
		If a related subtype is found this method behaves as if 
		{@link #findDefaultItems(DbControl, ItemSubtype, boolean)}
		was called with the related subtype and strict=false.
		
		@param dc A DbControl for database access
		@param subtype The main subtype
		@param relatedType The related item type that we are interested in
		@return A list with the items that was found (empty if no item was found)
		@since 3.0
	*/
	public <T extends BasicItem> List<T> findDefaultItemsOfRelatedSubtype(DbControl dc, ItemSubtype subtype, Item relatedType)
	{
		ItemSubtype related = subtype == null ? null : subtype.getRelatedSubtype(relatedType);
		if (related == null)
		{
			return findDefaultItems(dc, relatedType, true);
		}
		else
		{
			return findDefaultItems(dc, related, false);
		}
	}
	
	/**
		Get a list with all default items added to this project.
		@param dc A DbControl for database access
		@return A list with the default items
		@since 3.0
	*/
	public List<BasicItem> getDefaultItems(DbControl dc)
	{
		return getDefaultItems(dc, null);
	}
	
	/**
		Get a list with all default items of a given item type 
		added to this project.
		
		@param dc A DbControl for database access
		@param itemType The type of the default item or null to return all default items
		@return A list with the default items
		@since 3.0
	*/
	@SuppressWarnings("unchecked")
	public <T extends BasicItem> List<T> getDefaultItems(DbControl dc, Item itemType)
	{
		ItemParameterValueData defaultItems = getData().getDefaultItems();
		if (defaultItems == null) return Collections.emptyList();
		
		List<T> items = new ArrayList<>();
		Iterator<BasicData> it = defaultItems.getValues().iterator();
		Class<?> dataClass = itemType == null ? null : itemType.getDataClass();
		while (it.hasNext())
		{
			BasicData data = it.next();
			if (dataClass == null || dataClass.isInstance(data))
			{
				try
				{
					items.add((T)dc.getItem(BasicItem.class, data));
				}
				catch (PermissionDeniedException ex)
				{}
				catch (ItemNotFoundException ex)
				{
					it.remove();
				}
			}
		}
		return items;
	}

	
	/**
		Check if the given item is a default item in this project.
		@param item The item to check
		@return TRUE if the item is a default item, FALSE if not
		@since 3.0
	*/
	public boolean isDefaultItem(BasicItem item)
	{
		if (item == null) return false;
		ItemParameterValueData defaultItems = getData().getDefaultItems();
		if (defaultItems == null) return false;
		return defaultItems.getValues().indexOf(item.getData()) >= 0;
	}

	/**
		Check if the project has a default item of the given
		item type and (optional) subtype.
		
		@param itemType The item type to look for
		@param subtype The subtype to look for, or null to ignore subtypes
		@return TRUE if a matching default item is found, FALSE if not
		@since 3.0
	*/
	public boolean hasDefaultOfType(Item itemType, ItemSubtype subtype)
	{
		return findDefaultData(itemType, subtype == null ? null : subtype.getData(), true) != null;
	}
	
	/**
		Check if the project has a default item that has the same class
		and subtype as the given item.
		@param item The item to check
		@return TRUE if a matching default item is found, FALSE if not
		@since 3.0
	*/
	public boolean hasDefaultForItem(BasicItem item)
	{
		Item itemType = item.getType();
		ItemSubtypeData subtype = item instanceof Subtypable ? 
				((SubtypableData)item.getData()).getItemSubtype() : null;
		return findDefaultData(itemType, subtype, true) != null;
	}
	
	BasicData findDefaultRelatedData(DbControl dc, Subtypable item, Item relatedItemType, boolean strict)
	{
		ItemSubtype relatedSubtype = ItemSubtype.getRelatedSubtype(dc, item, relatedItemType);
		if (relatedSubtype == null) return null;
		
		return findDefaultData(relatedItemType, relatedSubtype.getData(), strict);
	}
	
	BasicData findDefaultData(DbControl dc, String subtypeId, boolean strict)
	{
		int id = SystemItems.getId(subtypeId);
		ItemSubtypeData subtype = HibernateUtil.loadData(dc.getHibernateSession(), ItemSubtypeData.class, id);
		return findDefaultData(Item.fromValue(subtype.getItemType()), subtype, strict);
	}
	
	/**
		Find a default item in this project of the given item type (required)
		and subtype (optional). This method will first try to locate an
		exact match for the given item type and subtype. If no items are found
		and strict is false, a second attempt is made to match items as follows:
		If a subtype is given it will match items WITHOUT a subtype (items with
		a different subtype are not matched). If no subtype is given, the search
		ignore the subtype on the other items.
		
		<p>
		If more than one item matches the critera, the first one found is returned. 
		
	  	@param itemType The main item type of the item we are looking for
	  	@param subtype The subtype of the item we are looking for (optional)
	  	@param strict TRUE if the subtype must match
	  	@since 3.0
	*/
	BasicData findDefaultData(Item itemType, ItemSubtypeData subtype, boolean strict)
	{
		ItemParameterValueData defaultItems = getData().getDefaultItems();
		if (defaultItems == null) return null;
		
		BasicData bestMatch = null;

		Class<?> dataClass = itemType.getDataClass();
		Iterator<BasicData> it = defaultItems.getValues().iterator();
		while (it.hasNext())
		{
			BasicData data = it.next();
			if (dataClass.isInstance(data))
			{
				// Check if item has been removed
				try
				{
					Hibernate.initialize(data);
				}
				catch (org.hibernate.ObjectNotFoundException ex)
				{
					it.remove();
					continue;
				}
				
				// Found an item of the correct item type... get it's subtype
				ItemSubtypeData otherSubtype = null;
				if (data instanceof SubtypableData)
				{
					otherSubtype = ((SubtypableData)data).getItemSubtype();
				}
				
				// If the subtypes are null or equal we have found a match
				if (EqualsHelper.equals(subtype, otherSubtype)) 
				{
					bestMatch = data;
					break;
				}
				
				if (!strict && bestMatch == null)
				{
					// Is this maybe the first non-strict match?
					if (subtype == null || otherSubtype == null)
					{
						bestMatch = data;
					}
				}
			}
		}	
		return bestMatch;
	}

	/**
		Find all default items in this project of the given item type (required)
		and subtype (optional). If a subtype is given, this method first try to
		match exactly. If no item is found, and strict is false, a second attempt
		is made to match against items WITHOUT subtype (items with a different
		subtype are never matched). 
		
	  	@param itemType The main item type of the item we are looking for
	  	@param subtype The subtype of the item we are looking for (optional)
	  	@param strict TRUE if the subtype must match
	  	@since 3.0
	*/
	List<BasicData> findAllDefaultData(Item itemType, ItemSubtypeData subtype, boolean strict)
	{
		ItemParameterValueData defaultItems = getData().getDefaultItems();
		if (defaultItems == null) return Collections.emptyList();
		
		List<BasicData> result = new ArrayList<BasicData>();
		List<BasicData> nonStrictMatches = strict ? null : new ArrayList<BasicData>();
		
		Class<?> dataClass = itemType.getDataClass();
		Iterator<BasicData> it = defaultItems.getValues().iterator();
		while (it.hasNext())
		{
			BasicData data = it.next();
			if (dataClass.isInstance(data))
			{
				// Check if item has been removed
				try
				{
					Hibernate.initialize(data);
				}
				catch (org.hibernate.ObjectNotFoundException ex)
				{
					it.remove();
					continue;
				}

				// Found an item of the correct item type... get it's subtype
				ItemSubtypeData otherSubtype = null;
				if (data instanceof SubtypableData)
				{
					otherSubtype = ((SubtypableData)data).getItemSubtype();
				}
				
				// If the subtypes are null or equal we have found a match
				if (EqualsHelper.equals(subtype, otherSubtype)) 
				{
					result.add(data);
				}
				else if (!strict)
				{
					// Non-strict matches
					if (subtype == null || otherSubtype == null)
					{
						nonStrictMatches.add(data);
					}
				}
			}
		}
		if (result.size() == 0 && !strict) result.addAll(nonStrictMatches);
		return result;
	}

	
	/**
	 	Gets the default raw data type for this project.
	 	If no default value has been set for raw data type this will return Null.
	 	
	  	@return A {@link net.sf.basedb.core.RawDataType} object
	*/
	public RawDataType getDefaultRawDataType()
	{
		RawDataType rdt = null;
		String id = getData().getProjectDefaults().get("raw_data_type");
		rdt = RawDataTypes.getRawDataType(id);		
		return rdt;
	}
	
	/**
	  	Sets a raw data type to be used as default for a project.
	  	The default value will be reset if <code>rawDataType</code> is Null
	  	
	  	@param rawDataType the {@link net.sf.basedb.core.RawDataType} to use as default.
	*/
	public void setDefaultRawDataType(RawDataType rawDataType)
	{
		checkPermission(Permission.WRITE);
		if (rawDataType == null) 
		{
			getData().getProjectDefaults().remove("raw_data_type");
		}
		else
		{		
			getData().getProjectDefaults().put("raw_data_type", rawDataType.getId());
		}
	}
	
	/**
		A runtime filter implementation that limits a query to only
		return projects where the logged in user is a member or owner
		unless the logged in user has generic read permission.
	*/
	private static class QueryRuntimeFilterImpl
		implements QueryRuntimeFilter
	{
		@Override
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			org.hibernate.Filter filter = null;
			SessionControl sc = dc.getSessionControl();
			int owner = sc.getLoggedInUserId();
			Set<Integer> projects = sc.getProjects();
			if (projects == null || projects.size() == 0) projects = Collections.singleton(0);

			if (sc.isDenied(query.getItemPermission(), Item.PROJECT))
			{
				manager.enableFilter("denyAll");
			}
			else if (query.isIncluded(Include.OTHERS) && sc.hasPermission(Permission.READ, query.getItemType()))
			{
				// Logged in user has generic read permission and wants other user's items...
				if (!query.isIncluded(Include.MINE))
				{
					// ... but not the one's with ownership
					filter = manager.enableFilter("notOwnedBy");
					if (filter != null)
					{
						filter.setParameter("owner", owner);
					}
				}
			}
			else if (query.isIncluded(Include.MINE, Include.SHARED))
			{
				// Logged in user wants own and shared projects
				filter = manager.enableFilter("ownedByOrMemberOf");
				if (filter != null)
				{
					filter.setParameter("owner", owner);
					filter.setParameterList("items", projects);
				}
			}
			else if (query.isIncluded(Include.SHARED))
			{
				// Logged in user wants only shared projects
				filter = manager.enableFilter("memberOf");
				if (filter != null)
				{
					filter.setParameter("owner", owner);
					filter.setParameterList("items", projects);
				}
			}
			else if (query.isIncluded(Include.MINE))
			{
				//Logged in user only wants own projects
				filter = manager.enableFilter("ownedBy");
				if (filter != null)
				{
					filter.setParameter("owner", owner);
				}
			}
			else
			{
				manager.enableFilter("denyAll");
			}
		}
	}
	
}



