/**
	$Id: Trashcan.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.Application.Pinger;
import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.data.RemovableData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.ResultList;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.filter.Filter;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import org.hibernate.HibernateException;

/**
	This class contains functionality to manage items that 
 	have been marked for deletion.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (Mon, 22 May 2017) $
*/
public class Trashcan
{

	/**
		Log trashcan events.
	*/
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(Trashcan.class);
	
	private static final boolean isDebug = log.isDebugEnabled();

	/**
		@see #getItems(DbControl, Set, int, int, Restriction)
	*/
	public static ResultList<Removable> getItems(DbControl dc, Item itemType, int firstItem, int maxItems)
		throws BaseException
	{
		return getItems(dc, itemType == null ? null : Collections.singleton(itemType), 
			firstItem, maxItems, null);
	}

	/**
		@see #getItems(DbControl, Set, int, int, Restriction)
	*/
	public static ResultList<Removable> getItems(DbControl dc, Set<Item> itemTypes, 
		int firstItem, int maxItems)
		throws BaseException
	{
		return getItems(dc, itemTypes, firstItem, maxItems, null);
	}

	/**
		Load all items flagged for deletion that are owned by or removed 
		by the logged in user. A restriction may be specified to limit the
		returned items. Note that separate queries are used for each item type
		but the same restriction is applied to all of them. It is recommended
		that the restriction is created with {@link 
		Restrictions#conditionalRestriction(Filter, Restriction, Restriction, boolean)}
		so that a restriction on, for example, the 'name' is only applied
		to {@link Nameable} items.

		@param dc The <code>DbControl</code> to use for database access
		@param itemTypes Limit the list to only return items of the types,
			included in the set, use null or an empty set to return items of all types
		@param firstItem The index of the first item to return (0-based)
		@param maxItems The maximum number of items to return, or 0 to return all items
		@param restriction A restriction to apply when loading items
		@return A list containing removable items
	 	@throws BaseException If there is an error of some kind
	 	@since 3.7
	*/
	public static ResultList<Removable> getItems(DbControl dc, Set<Item> itemTypes, 
		int firstItem, int maxItems, Restriction restriction)
		throws BaseException
	{
		try
		{
			org.hibernate.Session session = dc.getHibernateSession();
			SessionControl sc = dc.getSessionControl();

			if (itemTypes == null || itemTypes.size() == 0) itemTypes = Metadata.getRemovableItems();

			// Only load items owned or removed by the logged in user
			org.hibernate.Filter filter = HibernateUtil.enableFilter(session, "ownedOrRemovedBy");
			filter.setParameter("user", sc.getLoggedInUserId());
			List<RemovableData> removed = new ArrayList<RemovableData>();
			Restriction isRemoved = Restrictions.neq(Hql.property("removedBy"), null);
			
			for (Item item : itemTypes)
			{
				if (RemovableData.class.isAssignableFrom(item.getDataClass()))
				{
					// Query to load all "removed" items of a specific item type
					AbstractEntityQuery hql = new AbstractEntityQuery(item, null, false){};
					hql.restrict(isRemoved);
					if (restriction != null) hql.restrict(restriction);
					removed.addAll(HibernateUtil.loadList(hql.getMainHqlQuery(dc, RemovableData.class), sc));
				}
			}
			HibernateUtil.disableFilter(session, "ownedOrRemovedBy");
			
			List<Removable> items = new ArrayList<Removable>(removed.size());
			int index = 0;
			int totalItems = 0;
			boolean useLimit = firstItem >= 0 && maxItems > 0;
			int lastItem = firstItem + maxItems;
			for (RemovableData rd : removed)
			{
				BasicData data = (BasicData)rd;
				try
				{
					BasicItem item = dc.getItem(BasicItem.class, data);
					if (item.hasPermission(Permission.DELETE)) 
					{
						// The 'ownedOrRemovedBy' filter isn't supported on ChildItem:s
						// so we have to check the owner of parent item
						SharedItem parent = rd.getRemovedBy() != sc.getLoggedInUserId() && item instanceof ChildItem ? 
								dc.getItem(SharedItem.class, ((ChildItem)item).getSharedParent()) : null; 
						
						if (parent == null || parent.isOwner())
						{
							if (!useLimit || (index >= firstItem && index < lastItem))
							{
								items.add((Removable)item);
							}
							index++;
							totalItems++;
						}
					}
				}
				catch (PermissionDeniedException ex)
				{}
			}
			return new UnmodifiableResultList<Removable>(items, totalItems);
		}
		catch (HibernateException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Delete items permanently. This method only delete items that
		<ul>
		<li>have been marked for deletion (unless the ignoreFlag parameter is set)
		<li>the logged in user has permission to delete
		<li>not are used by other items
		</ul>
		Other items are ignored. This method may have to open several transactions
		to be able to delete as many items as possible. This means that this operation
		is not atomic and if an exception is thrown some items may have been deleted
		and some may not.
		
		@param sc The session control to use for connecting to the database. To
			avoid multithread-related problems the session control is cloned 
			before being used by this method. 
		@param items The items to remove
		@param ignoreFlag TRUE to also remove items that hasn't been flagged for removal
		@param progress An optional progress reporter
		@return The number of deleted items
	 	@throws BaseException If an item can't be deleted. 
		@since 2.4
	*/
	public static int delete(SessionControl sc, Collection<? extends Identifiable> items, 
		boolean ignoreFlag, ProgressReporter progress)
		throws BaseException
	{
		ThreadSignalHandler.checkInterrupted();
		if (progress != null) progress.display(0, "Checking which items to remove...");
		
		// Copy items to a set with predictable iteration order
		Set<Identifiable> itemsToRemove = new TreeSet<Identifiable>(Item.DELETION_ORDER);
		for (Identifiable item : items)
		{
			if (item.hasPermission(Permission.DELETE) &&
				(ignoreFlag || (item instanceof Removable && ((Removable)item).isRemoved())))
			{
				itemsToRemove.add(item);
			}
		}
		
		if (isDebug)
		{
			log.debug("Original list: " + items);
			log.debug("Sorted list: " + itemsToRemove);
		}
		
		if (itemsToRemove.size() == 0) return 0;
		int numToRemove = itemsToRemove.size();
		float progressFactor = 95.0f / numToRemove;
		int numRemoved = 0;
		int removedInTransaction = 0;
		boolean hasBrokenCircularReferences = false;
		int projectId = sc.getActiveProjectId();

		DbControl dc = null;
		Pinger pinger = null;
		SessionControl sc2 = null;
		try
		{
			// We create an impersonated session control to 
			// avoid issues where the original session control is closed,
			// for example, due to the user logging out
			sc2 = sc.impersonateLogin(sc.getLoggedInUserId(), "Deleting items from trashcan");
			pinger = Application.newPinger(sc2);
			// We try to delete items from the set as long as there are items
			// left and at least one item was deleted in the last iteration
			do
			{
				ThreadSignalHandler.checkInterrupted();
				dc = sc2.newDbControl();
				log.debug("Start new delete transaction");
				try
				{
					if (projectId != 0 && sc2.getActiveProjectId() != projectId)
					{
						sc2.setActiveProject(Project.getById(dc, projectId));
					}

					removedInTransaction = 0;
					Iterator<Identifiable> iterator = itemsToRemove.iterator();
					deleteTransaction:
					while (iterator.hasNext())
					{
						ThreadSignalHandler.checkInterrupted();
						Identifiable item = iterator.next();
						if (progress != null)
						{
							String name = item instanceof Nameable ? ((Nameable)item).getName() : item.toString();
							progress.display((int)(numRemoved * progressFactor), "Remove " +
									item.getType() + ": " + name + 
									" (" + (numRemoved + 1) + " of " + numToRemove + ")");
						}
						BasicItem basicItem = item.getType().getById(dc, item.getId());
						if (!basicItem.isUsed()) 
						{
							if (isDebug) log.debug("Deleting item " + basicItem);
							dc.deleteItem(basicItem);
							numRemoved++;
							removedInTransaction++;
							iterator.remove();
						}
						else 
						{
							if (isDebug) log.debug("Item is used: " + basicItem);
							if (progress != null) progress.append(" [USED]");
							if (removedInTransaction > 0)
							{
								// Commit, start new transaction and try again
								break deleteTransaction;
							}
						}
					}
					log.debug("Committing transaction; " + removedInTransaction + 
						" items scheduled for deletion; " + numRemoved + " total so far; " +
						itemsToRemove.size() + " remains in list");
					if (progress != null)
					{
						progress.display((int)(numRemoved * progressFactor), 
							numRemoved + " of " + numToRemove + " items removed; flushing to database");
					}
					dc.commit();
				}
				finally
				{
					if (dc != null) dc.close();
				}
				
				if (removedInTransaction == 0 && itemsToRemove.size() > 0 && !hasBrokenCircularReferences)
				{
					// Not all items could be removed, try to break circular references
					// Note that as a (desired) side-effect non-removable items will be
					// removed from 'itemsToRemove'
					hasBrokenCircularReferences = true; // No need to do this twice
					removedInTransaction = breakCircularReferences(sc2, itemsToRemove);
					if (isDebug) 
					{
						log.debug("Removed " + removedInTransaction + " circular references; " +
							itemsToRemove.size() + " items remains in list");
					}
				}

			} while (removedInTransaction > 0 && itemsToRemove.size() > 0);
		}
		finally
		{
			if (pinger != null) pinger.stop();
			if (sc2 != null)
			{
				sc2.logout();
				sc2.close();
			}
		}
		
		if (progress != null)
		{
			String msg = "";
			if (numRemoved < numToRemove)
			{
				msg = numRemoved + " of " + numToRemove + " items removed. " + 
					(numToRemove - numRemoved) + " items are used and can't be removed";
			}
			else
			{
				msg = "All " + numToRemove + " items removed";
			}
			progress.display(100, msg);
		}
		return numRemoved;
	}
	
	/**
		Delete all items in the trashcan of the logged in user. This method
		calls {@link #getItems(DbControl, Item, int, int)} to load all items
		that have been flagged for removal and then 
		{@link #delete(SessionControl, Collection, boolean, ProgressReporter)}
		to remove those items. See the <code>delete</code> for
		more information about transactions.
		
		@param sc The session control to use for connecting to the database
		@param progress An optional progress reporter
		@return The number of deleted items
		@since 2.4
	*/
	public static int deleteAll(SessionControl sc, ProgressReporter progress)
	{
		DbControl dc = sc.newDbControl();
		try
		{
			List<Removable> items = getItems(dc, (Set<Item>)null, 0, 0);
			dc.close();
			return delete(sc, items, false, progress);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	private static int breakCircularReferences(SessionControl sc, Set<Identifiable> itemsToRemove)
	{
		if (isDebug) log.debug("Trying to break circular references: " + itemsToRemove);
		
		DbControl dc = sc.newDbControl();
		int circularRefsBroken = 0;
		try
		{

			// Prepare by loading the using-items for all itemsToRemove
			List<CircularRefInfo> cRef = new LinkedList<CircularRefInfo>();
			// The proxies will speed up the checks
			Set<ItemProxy> itemsToRemoveProxy = new HashSet<ItemProxy>();
			for (Identifiable item : itemsToRemove)
			{
				CircularRefInfo c = new CircularRefInfo(dc, item);
				cRef.add(c);
				itemsToRemoveProxy.add(c.proxy);
			}
			
			// The strategy is to check which items in the itemsToRemove collection
			// that can never be removed since they have references to items that are
			// not in the itemsToRemove collection.
			// Such items are removed from the itemsToRemove collection
			// We need to to iterate until the itemsToRemove collection is stable
			// All remaining items are items that only have references to other items
			// in the itemsToRemove.... We have found the ones with circular references!!
			int numUnremovable;
			do
			{
				numUnremovable = 0;
				Iterator<CircularRefInfo> it = cRef.iterator();
				while (it.hasNext())
				{
					CircularRefInfo c = it.next();
					// For a given item 'c.item', if it has any using-item that is not
					// found among itemsToRemove 'c.item' can't be removed
					if (c.isUsingNonRemovableItem(itemsToRemoveProxy))
					{
						// 3. Remove 'c' from itemsToRemove and other collections
						itemsToRemove.remove(c.item);
						itemsToRemoveProxy.remove(c.proxy);
						it.remove();
						numUnremovable++;
					}
				}
				
			} while(itemsToRemove.size() > 0 && numUnremovable > 0);
			// Repeat until no more items are removed from itemsToRemove
			
			// The remaining items in itemsToRemove are those with possible circular references
			if (isDebug) log.debug(itemsToRemove.size() + " remaining items with circular ref: " + itemsToRemove);
			
			if (itemsToRemove.size() > 0)
			{
				// Query for breaking AnyToAnyLinks between items
				org.hibernate.query.Query<?> query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), 
					"DELETE_ALL_ANYTOANY_TO");
				/*
					DELETE FROM AnyToAnyData ata 
					WHERE ata.toId = :toId AND ata.toType = :toType
				*/
				// Break circular references 
				for (CircularRefInfo c : cRef)
				{
					query.setParameter("toId", c.proxy.getId(), TypeWrapper.H_INTEGER);
					query.setParameter("toType", c.proxy.getType().getValue(), TypeWrapper.H_INTEGER);
					int brokenAnyToAny = query.executeUpdate();
					int brokenRefs = c.proxy.getItem(dc).breakCircularReferences(itemsToRemoveProxy);
					if (isDebug) 
					{
						log.debug(c.item + ": " + brokenRefs + " broken references; " + 
							brokenAnyToAny + " broken any-to-any links");
					}
					circularRefsBroken += brokenAnyToAny + brokenRefs;
				}
			}
			
			dc.commit();
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return circularRefsBroken;
	}

	/**
		Helper class for holding information about a single
		item, other items using it and other information that is 
		relevant for resolving circular references.
		@since 3.8
	*/
	static class CircularRefInfo
	{
		final Identifiable item;
		final ItemProxy proxy;
		final Set<ItemProxy> usingItems;
		
		public CircularRefInfo(DbControl dc, Identifiable item) 
		{
			this.item = item;
			this.proxy = new ItemProxy(item.getId(), item.getType());
			this.usingItems = proxy.getItem(dc).getUsingItems();
		}
		
		/**
			Check if all using items of the main items is found among the
			'itemsToRemove'. If not, this item can't be removed even
			if circular references are broken.
			@return True if this item is using a non-removable item, false otherwise
		*/
		boolean isUsingNonRemovableItem(Set<ItemProxy> itemsToRemove)
		{
			for (ItemProxy p : usingItems)
			{
				if (itemsToRemove.contains(p))
				{
					if (isDebug) log.debug(item+" has possible circular ref to: "+p);
				}
				else
				{
					if (isDebug) log.debug(item+" has ref to non-removable item: "+p);
					return true;
				}
			}
			return false;
		}
		
	}
	
}
