/*
	$Id: PlateData.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.data;

import java.util.Date;
import java.util.Set;
import java.util.Map;
import java.util.HashMap;

/**
	This class holds information about a plate.

	@author Nicklas
	@version 2.0
	@see net.sf.basedb.core.Plate
	@see <a href="../../../../../../html/developer/api/data_api.plates.html">Developer documentation: Array LIMS (plates)</a>
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (Fri, 14 Sep 2012) $
	@hibernate.class table="`Plates`" lazy="false"
*/
public class PlateData
	extends AnnotatedData
	implements RegisteredData
{

	public PlateData()
	{}

	/*
		From the RegisteredData interface
		-------------------------------------------
	*/
	private Date entryDate;
	@Override
	public Date getEntryDate()
	{
		return entryDate;
	}
	public void setEntryDate(Date entryDate)
	{
		this.entryDate = entryDate;
	}
	// -------------------------------------------

	private PlateTypeData plateType;
	/**
		The plate type of this plate.
		@see PlateTypeData#getPlates()
		@hibernate.many-to-one column="`platetype_id`" not-null="true" outer-join="false" update="false"
	*/
	public PlateTypeData getPlateType()
	{
		return plateType;
	}
	public void setPlateType(PlateTypeData plateType)
	{
		this.plateType = plateType;
	}

	/**
		The maximum length of the barcode that can be stored in the database.
		@see #setBarcode(String)
	*/
	public static final int MAX_BARCODE_LENGTH = 255;
	private String barcode;
	/**
		Get the barcode for the plate. It is not required and doesn't have to be unique.
		@hibernate.property column="`barcode`" type="string" length="255" not-null="false" 
	*/
	public String getBarcode()
	{
		return barcode;
	}
	public void setBarcode(String barcode)
	{
		this.barcode = barcode;
	}

	private boolean destroyed;
	/**
		If the plate has been destroyed.
		@hibernate.property column="`destroyed`" type="boolean" not-null="true" 
	*/
	public boolean isDestroyed()
	{
		return destroyed;
	}
	public void setDestroyed(boolean destroyed)
	{
		this.destroyed = destroyed;
	}

	private PlateMappingData plateMapping;
	/**
		If this plate was created by a plate mapping, it is linked here.
		The source plates are kept in the {@link #getParents()} property.
		@hibernate.many-to-one column="`platemapping_id`" not-null="false" outer-join="false" update="false"
	*/
	public PlateMappingData getPlateMapping()
	{
		return plateMapping;
	}
	public void setPlateMapping(PlateMappingData plateMapping)
	{
		this.plateMapping = plateMapping;
	}

	private int destinationIndex;
	/**
		If the plate was created by a plate mapping, this is the index
		of this plate in the list of created plates. First plate have index 0.
		@hibernate.property column="`destination_index`" not-null="true" update="false"
	*/
	public int getDestinationIndex()
	{
		return destinationIndex;
	}
	public void setDestinationIndex(int destinationIndex)
	{
		this.destinationIndex = destinationIndex;
	}

	private Map<PlateData, Integer> parents;
	/**
		If the plate was created by a {@link PlateMappingData} the source plates
		are kept here.
		@see #getPlateMapping()
		@hibernate.map table="`PlateParents`" lazy="true" 
		@hibernate.collection-key column="`plate_id`"
		@hibernate.index-many-to-many column="`parent_id`" class="net.sf.basedb.core.data.PlateData"
		@hibernate.collection-element column="`source_index`" type="int"
	*/
	public Map<PlateData, Integer> getParents()
	{
		if (parents == null) parents = new HashMap<PlateData, Integer>();
		return parents;
	}
	void setParents(Map<PlateData, Integer> parents)
	{
		this.parents = parents;
	}

	private Set<PlateData> parentSet;
	/**
		Duplicate of {@link #getParents()} but with a set instead of a map.
		This is needed if we want to use the parent plates in a query.
		@see #getParents()
		@since 2.9
		@hibernate.set table="`PlateParents`" lazy="true" inverse="true"
		@hibernate.collection-key column="`plate_id`"
		@hibernate.collection-many-to-many column="`parent_id`" 
			class="net.sf.basedb.core.data.PlateData"
	*/
	Set<PlateData> getParentSet()
	{
		return parentSet;
	}
	void setParentSet(Set<PlateData> parentSet)
	{
		this.parentSet = parentSet;
	}

	
	private Set<PlateData> children;
	/**
		The childs plates created by plate mappings. This is the inverse end.
		@see #getParents()
		@hibernate.set table="`PlateParents`" lazy="true" inverse="true"
		@hibernate.collection-key column="`parent_id`"
		@hibernate.collection-many-to-many column="`plate_id`" class="net.sf.basedb.core.data.PlateData"
	*/
	Set<PlateData> getChildren()
	{
		return children;
	}
	void setChildren(Set<PlateData> children)
	{
		this.children = children;
	}

	private Map<PlateEventTypeData, PlateEventData> events;
	/**
		The event for this plate. The event must have an event type that is valid
		for the corresponding plate type. This is the inverse end.
		@see PlateEventData#getPlate()
		@hibernate.map lazy="true" inverse="true" cascade="delete"
		@hibernate.collection-key column="`plate_id`"
		@hibernate.index-many-to-many column="`plateeventtype_id`" class="net.sf.basedb.core.data.PlateEventTypeData"
		@hibernate.collection-one-to-many class="net.sf.basedb.core.data.PlateEventData"
	*/
	public Map<PlateEventTypeData, PlateEventData> getEvents()
	{
		if (events == null) events = new HashMap<PlateEventTypeData, PlateEventData>();
		return events;
	}
	void setEvents(Map<PlateEventTypeData, PlateEventData> events)
	{
		this.events = events;
	}

	private Map<PlateCoordinate, WellData> wells;
	/**
		The wells on this plate. This is the inverse end.
		@see WellData#getPlate()
		@hibernate.map lazy="true" inverse="true" cascade="all"
		@hibernate.collection-key column="`plate_id`"
		@hibernate.collection-composite-index class="net.sf.basedb.core.data.PlateCoordinate"
		@hibernate.collection-one-to-many class="net.sf.basedb.core.data.WellData"
	*/
	public Map<PlateCoordinate, WellData> getWells()
	{
		if (wells == null)
		{
			PlateGeometryData pg = getPlateType().getPlateGeometry();
			int numWells = pg == null ? 0 : pg.getRows() * pg.getColumns();
			wells = new HashMap<PlateCoordinate, WellData>(numWells);
		}
		return wells;
	}
	void setWells(Map<PlateCoordinate, WellData> wells)
	{
		this.wells = wells;
	}

	private Set<ArrayDesignPlateData> arrayDesigns;
	/**
	   Used by Hibernate to link with the arraydesign for this plate.
	   
	   @hibernate.set lazy="true" inverse="true"
	   @hibernate.collection-key column="`plate_id`"
	   @hibernate.collection-one-to-many class="net.sf.basedb.core.data.ArrayDesignPlateData"
	*/
	Set<ArrayDesignPlateData> getArrayDesigns()
	{
	   return arrayDesigns;
	}
	void setArrayDesigns(Set<ArrayDesignPlateData> arrayDesigns)
	{
	   this.arrayDesigns = arrayDesigns;
	}
}
