/**
	$Id: DbLogManagerFactory.java 7805 2020-05-04 11:51:26Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.log.db;

import java.util.HashMap;
import java.util.Map;

import net.sf.basedb.core.AnnotationBatcher.CurrentAnnotationInfo;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.LogControl;
import net.sf.basedb.core.data.AnnotationData;
import net.sf.basedb.core.data.AnnotationSetData;
import net.sf.basedb.core.data.AnyToAnyData;
import net.sf.basedb.core.data.BioMaterialEventData;
import net.sf.basedb.core.data.FileSetMemberData;
import net.sf.basedb.core.data.LoggableData;
import net.sf.basedb.core.data.PlateEventData;
import net.sf.basedb.core.log.EntityLogger;
import net.sf.basedb.core.log.LogManager;
import net.sf.basedb.core.log.LogManagerFactory;
import net.sf.basedb.core.log.ManualLogEntry;
import net.sf.basedb.util.extensions.ActionFactory;
import net.sf.basedb.util.extensions.InvokationContext;

/**
	Log manager factory implementationt that logs changes 
	to the database. Only items that has been tagged with
	the {@link LoggableData} interface are logged.

	@author Nicklas
	@version 2.13
	@base.modified $Date: 2020-05-04 13:51:26 +0200 (Mon, 04 May 2020) $
*/
public class DbLogManagerFactory
	implements LogManagerFactory, ActionFactory<LogManagerFactory>
{
	
	private final Map<Class<?>, EntityLogger> specialLoggers;
	private final DefaultEntityLogger defaultLogger;
	private final boolean detailedProperties;
	private final boolean oldPropertyValues;
	
	public DbLogManagerFactory()
	{
		this(Config.getBoolean("changelog.dblogger.detailed-properties"), Config.getBoolean("changelog.dblogger.old-property-values"));
	}
	
	public DbLogManagerFactory(boolean detailedProperties, boolean oldValues)
	{
		this.specialLoggers = new HashMap<Class<?>, EntityLogger>();
		this.detailedProperties = detailedProperties;
		this.oldPropertyValues = detailedProperties && oldValues;

		this.defaultLogger = new DefaultEntityLogger(this);
		setSpecialLogger(ManualLogEntry.class, new ManualEntryLogger(this));
		setSpecialLogger(BioMaterialEventData.class, new BioMaterialEventLogger(this));
		setSpecialLogger(AnnotationData.class, new AnnotationLogger(this));
		setSpecialLogger(CurrentAnnotationInfo.class, new AnnotationLogger(this));
		setSpecialLogger(FileSetMemberData.class, new FileSetMemberLogger(this));
		setSpecialLogger(PlateEventData.class, new PlateEventLogger(this));
		setSpecialLogger(AnnotationSetData.class, new AnnotationSetLogger(this));
		setSpecialLogger(AnyToAnyData.class, new AnyToAnyLogger(this));
		
	}
	
	/*
		From the ActionFactory interface 
		--------------------------------
	*/
	@Override
	public boolean prepareContext(InvokationContext<? super LogManagerFactory> context) 
	{
		return true;
	}
	
	@Override
	public LogManagerFactory[] getActions(InvokationContext<? super LogManagerFactory> context)
	{
		return new LogManagerFactory[] { this };
	}
	// -----------------------------------

	/*
		From the LogManagerFactory interface
		------------------------------------
	*/
	/**
		@return TRUE if the entity implements the {@link LoggableData}
	*/
	@Override
	public boolean isLoggable(Object entity)
	{
		return entity instanceof LoggableData;
	}

	/**
		If a special logger has been registered for the class of the given
		entity, use that logger, otherwise use the default logger. 
		@return An EntityLogger or null if the entity isn't implementing
			{@link LoggableData}
	*/
	@Override
	public EntityLogger getEntityLogger(LogManager logManager, Object entity)
	{
		if (!isLoggable(entity)) return null;
		Class<?> entityClass = entity.getClass();
		if (org.hibernate.proxy.HibernateProxy.class.isAssignableFrom(entityClass))
		{
			// If the entity is a proxy we must find out the real data class
			entityClass = org.hibernate.proxy.HibernateProxyHelper.getClassWithoutInitializingProxy(entity);
		}
		EntityLogger logger = specialLoggers.get(entityClass);
		if (logger == null) logger = defaultLogger;
		return logger;
	}

	/**
		Creates a new {@link DbLogManager}.
	*/
	@Override
	public LogManager getLogManager(LogControl logControl)
	{
		return new DbLogManager(logControl);
	}
	// ---------------------------------------

	/**
		Register a special logger for entities of the given class.
		@param clazz The class to register a logger for. The class
			must implement the {@link LoggableData} interface
		@param logger The logger to use. It must be a thread-safe
			implementation since it can be used by multiple threads
			at the same time
	*/
	public void setSpecialLogger(Class<? extends LoggableData> clazz, EntityLogger logger)
	{
		if (clazz == null) throw new NullPointerException("clazz");
		if (!LoggableData.class.isAssignableFrom(clazz))
		{
			throw new ClassCastException(clazz + " doesn't implement LoggableData");
		}
		if (logger == null)
		{
			specialLoggers.remove(clazz);
		}
		else
		{
			specialLoggers.put(clazz, logger);
		}
	}

	/**
		Should individual log entries for each changed property be created or not?
		@since 3.3
	*/
	public boolean isLoggingDetailedProperties()
	{
		return detailedProperties;
	}

	/**
		Should log entries with old property values be created or not?
		@since 3.3
	*/
	public boolean isLoggingOldPropertyValues()
	{
		return oldPropertyValues;
	}

}
