/*
	$Id: Response.java 7329 2017-04-07 07:21:57Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.plugin;

import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.signal.Signal;

import java.io.Serializable;
import java.util.List;

/**
	The response interface used by the core.

	@author Nicklas, Enell
	@version 2.0
	@base.modified $Date$
*/
public interface Response
{
	/**
		This method will continue the configuration or execution of the 
		job with a new command to the plugin. If the plug-in is currently
		being configured the configuration will continue with the next
		command. If the plug-in is currently executing the job will
		be returned to job queue. The intention is that plug-ins can use 
		this to survive a temporary shutdown of the system after
		catching the {@link Signal#SHUTDOWN} signal.
			
		@param nextCommand The next command to the plugin
		@throws PermissionDeniedException If another command isn't allowed
	*/
	public void setContinue(String nextCommand)
		throws PermissionDeniedException;
	
	/**
		Can be used by an executing job to also store state information
		that will be restored when the job is starting up again. If used
		in the configuration phase, the state information is ignored.
		
		@param nextCommand The next command to the plugin
		@param state State information that will be restored when resuming
		@since 3.11
		@see #setPause(String, String, Serializable)
	*/
	public void setContinue(String nextCommand, Serializable state);

	/**
		This can be used in the configuration phase of a plug-in
		to clear the stored state and nextCommand information. The 
		intention is that a PAUSED plug-in that is being re-configured
		should be able to force a re-start from the beginning if the
		configuration has changed in a way that restarting from the
		stored state doesn't make any sense.
		@since 3.11
	*/
	public void setClearState();
	
	/**
		This will end the configuration of a job and set the estimated
		execution time for it.
		@param message An optional message
		@param estimatedExecutionTime An estimated execution time for the job,
			null is the same as <code>ExecutionTime.LONG</code>
	*/
	public void setDone(String message, Job.ExecutionTime estimatedExecutionTime);

	/**
		This will end the configuration of a job, and start to execute it
		immediately if the plugin is permitted to do so. If it is not possible
		to execute the job immediately it will be added to the job queue for 
		later execution unless the <code>failIfDenied</code> flag is set to true,
		in which case an exception will be thrown.
		<p>
		If immediate execution is possible the client application must continue to 
		use the same instance of the plugin object. This means that immediate execution
		always happens in the same virtual machine. Plugins should be aware of
		this and avoid excessive use of resources such as memory and CPU time.
		<p>
		The client application may execute the job in another thread.
		@param message An optional message
		@param estimatedExecutionTime An estimated execution time for the job,
			null is the same as <code>ExecutionTime.LONG</code>. This value is only used
			if immediate execution isn't possible and the job has to be added to the
			job queue.
		@param failIfDenied TRUE if a PermissionDeniedException should be thrown if
			immediate exection isn't possible, FALSE to add the job to the job queue 
			for later exection
	 	@throws PermissionDeniedException If logged in user doesn't have write permission on the job. 
		@since 2.2
	*/
	public void setExecuteImmediately(String message, Job.ExecutionTime estimatedExecutionTime, 
		boolean failIfDenied)
		throws PermissionDeniedException;
	
	/**
		This will end the configuration of a <b>export</b> job, and start to execute it
		for immediate download. This method is similar to the {@link 
		#setExecuteImmediately(String, Job.ExecutionTime, boolean)} but:
		<ul>
		<li>The plugin must implement the {@link ImmediateDownloadExporter} interface.
		<li>The core will call {@link ImmediateDownloadExporter#doExport(ExportOutputStream, ProgressReporter)}
			instead of the {@link Plugin#run(Request, Response, ProgressReporter)} method.
		</ul>
		
		@param message An optional message
		@param estimatedExecutionTime An estimated execution time for the job,
			null is the same as <code>ExecutionTime.LONG</code>. This value is only used
			if immediate execution isn't possible and the job has to be added to the
			job queue.
		@param failIfDenied TRUE if a PermissionDeniedException should be thrown if
			immediate exection isn't possible, FALSE to add the job to the job queue 
			for later exection
	 	@throws PermissionDeniedException If the logged in user doesn't have write permission to the job. 
		@since 2.2
	 */
	public void setDownloadImmediately(String message, Job.ExecutionTime estimatedExecutionTime, 
		boolean failIfDenied)
		throws PermissionDeniedException;

	/**
		A plug-in can call this method to notify the core that it has been
		paused. The job should remain in the job queue. A user may manually
		decide to either re-configure the job or resume it from where it 
		was paused.
		
		Note to plug-in developers: The job will be resumed in a different 
		instance. Any state information that is needed when resuming can be
		stored in the 'state' parameter. This information will be provided 
		to the plug-in when re-starting.
		
		@param message An optional message to update the status field
		@param nextCommand The command to send to the plug-in when resuming
		@param state State information that will be restored when resuming
		@since 3.11
	*/
	public void setPause(String message, String nextCommand, Serializable state);

	/**
		This will end the job execution and report it as a success.
		@param message A message containing any relevant information about
			what the plugin did, or null
	*/
	public void setDone(String message);
	
	/**
		A plug-in may call this method in it's configuration phase or
		after it has been executed to indicate that the job should be
		or was executed as a dry-run job. A dry-run job is a job that 
		does all the things a normal job would do except committing the
		changes to the database. The dry-run job may output a 
		log file with detailed information about what should have happened
		if it hadn't been a dry-run.
		<p>
		NOTE! Setting this value to true doesn't change the behaviour
		of the core. It is the responsibility of the plug-in to check
		this value(eg. {@link Job#isDryRun()} and take appropriate action.
		
		@param dryRun TRUE to indicate that the job should be or was
			executed as a dry-run
		@since 2.8
	*/
	public void setDryRun(boolean dryRun);
	
	/**
		A plug-in may call this method on it's job configuration phase
		to suggest a name for the job. Otherwise a default name will be
		generated by the core.
		@since 2.17
	*/
	public void setSuggestedJobName(String jobName);
	
	/**
		This will end the job execution or configuration and report it as a 
		failure. If in a configuration sequence the client application may try 
		the same command again with different parameters.
		
		@param message Information about the reason for the error
		@param errorList Detailed information about the error 
	*/
	public void setError(String message, List<? extends Throwable> errorList);
	
	public enum Status
	{
		DONE(true), 
		CONTINUE(false), 
		ERROR(true), 
		EXECUTE_IMMEDIATELY(true), 
		DOWNLOAD_IMMEDIATELY(true),
		PAUSE(false);
		
		private final boolean allowSaveParameters;
		
		private Status(boolean allowSaveParameters)
		{
			this.allowSaveParameters = allowSaveParameters;
		}
		
		public boolean allowSaveParameters()
		{
			return allowSaveParameters;
		}
		
	}
}
