/*
	$Id$

	Copyright (C) 2020 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.util.Values;

/**
	Restriction that filters on a list of ID values:
	
	id IN (a, b, c, ....)
	
	@author nicklas
	@since 3.16
*/
public class IdListRestriction
	implements Restriction
{

	private final Set<Integer> idList;
	private final Expression idExpression;
	
	public IdListRestriction()
	{
		this(new HashSet<>());
	}
	
	public IdListRestriction(Set<Integer> idList)
	{
		this.idList = new HashSet<>(idList);
		this.idExpression = Hql.property("id");
	}
	
	@Override
	public String toQl(Query query, DbControl dc) 
	{
		String idString = idList.size() == 0 ? "0" : Values.getString(idList, ", ", true);
		return "(" + idExpression.toQl(query, dc) + " IN (" + idString + "))";
	}

	@Override
	public Collection<? extends QueryElement> getChildren()
	{
		return Collections.singleton(idExpression);
	}
	@Override
	public String toString() 
	{
		String idString = idList.size() == 0 ? "0" : Values.getString(idList, ", ", true);
		return idExpression.toString() + " IN (" + idString + ")";
	}

	/**
		Merge this id list with another list. This list will now 
		contain the combined set of ids from both lists.
	*/
	public void union(IdListRestriction other)
	{
		this.idList.addAll(other.idList);
	}
	
	/**
		Merge this id list with another list. This list will
		now contain only the ids that are found in both lists.
	*/
	public void intersect(IdListRestriction other)
	{
		this.idList.retainAll(other.idList);
	}
	
	
}
