/**
	$Id: SnapshotManager.java 7767 2020-02-06 12:33:37Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.snapshot;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.Annotation;
import net.sf.basedb.core.AnnotationSet;
import net.sf.basedb.core.Application;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.Project;
import net.sf.basedb.util.StaticCache;
import net.sf.basedb.util.filter.Filter;
import net.sf.basedb.util.filter.StaticFilter;

/**
	Manager for annotation snapshots. The manager keeps an internal cache in memory
	of snapshots that has been loaded. If an unload snapshot is needed the manager
	will first look in the static cache. If it is not found there the last resort
	is to generate a new snapshot.
	
	@author Nicklas
	@version 2.14
	@base.modified $Date: 2020-02-06 13:33:37 +0100 (Thu, 06 Feb 2020) $
*/
public class SnapshotManager
{

	private final Map<Integer, AnnotationSetSnapshot> snapshots;
	private DisabledInheritanceFilter disabledInheritanceFilter;
	
	/**
		The static cache key under which the snapshot for the
		annotation set with the given id is cached.
		@param annotationSetId The ID of an annotation set
	*/
	public static String getCacheKey(int annotationSetId)
	{
		int subdir = annotationSetId % 1000;
		return "/snapshots-v" + AnnotationSetSnapshot.FILE_VERSION +
			"/" + subdir + "/annotation-set-" + annotationSetId + ".ser";
	}
	
	/**
		Removes a snapshot from the static cache.
		@param annotationSetId The ID of the annotation set  
		@return TRUE if the snapshot was removed or didn't
			exists to begin with, FALSE if it existed and could
			not be removed
	*/
	public static boolean removeSnapshot(int annotationSetId)
	{
		StaticCache cache = Application.getStaticCache();
		String cacheKey = getCacheKey(annotationSetId);
		return cache.delete(cacheKey, 1000);
	}

	/**
		@see #removeSnapshots(Collection)
		@since 3.5
	*/
	public static int removeSnapshots(List<Integer> annotationSetIds)
	{
		return removeSnapshots(annotationSetIds);
	}

	/**
		Removes a list of snapshots from the static cache.
		@param annotationSetIds A list with IDs of annotation set  s
		@return The number of successfully removed snapshots
		@since 3.10
	*/
	public static int removeSnapshots(Collection<Integer> annotationSetIds)
	{
		StaticCache cache = Application.getStaticCache();
		int removed = 0;
		for (Integer i : annotationSetIds)
		{
			String cacheKey = getCacheKey(i);
			if (cache.delete(cacheKey, 1000)) removed++;
		}
		return removed;
	}
	
	/**
		Create a new snapshot manager.
	*/
	public SnapshotManager()
	{
		this.snapshots = new HashMap<Integer, AnnotationSetSnapshot>();
	}
	
	/**
		Clear the manager from all memory-cached information.
		@since 3.16
	*/
	public void clear()
	{
		snapshots.clear();
	}
	
	private Integer projectId;
	
	/**
		When a project is set, the snapshot manager will only 
		search for project-specific annotations explicitely belonging
		to that project (or default values if project == null).
		
		Note that this option typically always need to be off 
		({@link #setNoProject()}) when searching for annotations to an 
		annotation type that doesn't have enabled project-specific annotations.
		
		Calling this method with a null parameter IS NOT the same
		as calling {@link #setNoProject()}.
		
		@param project The project or null to only search for default
			values. Use {@link #setNoProject()} to revert to the default
			behaviour (which load default values if a project-specific values
			doesn't exists)
		@since 3.11
	*/
	public void setProject(Project project)
	{
		this.projectId = project == null ? 0 : project.getId();
	}
	
	/**
		Revert to the default bahavior for loading project-specific annotations.
		
		@since 3.11
	*/
	public void setNoProject()
	{
		this.projectId = null;
	}
	
	/**
		Get the annotation snapshot for the annotation set with the given id.
		@param dc A DbControl to use if the snapshot has to be created from
			the database cache
		@param annotationSetId The id of the annotation set
		@return A snapshot, or null if the annotation set with the given
		 id doesn't exists
	*/
	public AnnotationSetSnapshot getSnapshot(DbControl dc, int annotationSetId)
	{
		// First check the internal cache
		AnnotationSetSnapshot snapshot = snapshots.get(annotationSetId);
		if (snapshot != null) return snapshot;
		
		// Then check the static cache
		StaticCache cache = Application.getStaticCache();
		String cacheKey = getCacheKey(annotationSetId);
		snapshot = cache.load(cacheKey, 1000);
		if (snapshot == null) 
		{
			// As a last resort, create a new snapshot and cache it
			snapshot = AnnotationSet.createSnapshot(dc, annotationSetId);
			if (snapshot != null) cache.store(cacheKey, snapshot, 1000);
		}
		snapshots.put(annotationSetId, snapshot);
		return snapshot;
	}

	/**
		Find annotations. Alwasys search primary and optionally inherited.
		@see #findAnnotations(DbControl, AnnotationSetSnapshot, Filter, boolean, boolean)
	*/
	public List<AnnotationSnapshot> findAnnotations(DbControl dc, AnnotationSetSnapshot snapshot, 
			Filter<? super AnnotationSnapshot> filter, boolean searchInherited)
	{
		return findAnnotations(dc, snapshot, filter, true, searchInherited);
	}

	/**
		Search in a snapshot for all annotations matching a given filter.
		If the snapshot inherits annotations and/or annotation sets from other
		annotations sets, those snapshots are search recursively one level down.
		
		@param dc A DbControl to use for database access if a snapshot doesn't
			exists for an inherited annotation set
		@param snapshot The snapshot to search
		@param filter A filter that should match the wanted annotation snapshots
			or null to match all annotations
		@param searchPrimary TRUE if primary annotations should be searched
		@param searchInherited TRUE if inherited annotations should be searched
		@return A list with the found annotations
		@since 3.5
	*/
	public List<AnnotationSnapshot> findAnnotations(DbControl dc, AnnotationSetSnapshot snapshot, 
			Filter<? super AnnotationSnapshot> filter, boolean searchPrimary, boolean searchInherited)
	{
		List<AnnotationSnapshot> result = new ArrayList<AnnotationSnapshot>();
		if (!searchPrimary && !searchInherited || snapshot == null) return result;
		
		if (filter == null) filter = new StaticFilter<AnnotationSnapshot>(true);
		int activeProjectId = projectId == null ? 
			dc.getSessionControl().getActiveProjectId() : projectId;

		// Needed for recursive searching of inherited annotations
		AnnotationFilter annotationFilter = null;
		DisabledInheritanceFilter recursiveFilter = null;
		if (searchInherited)
		{
			if (disabledInheritanceFilter == null) disabledInheritanceFilter = new DisabledInheritanceFilter(dc);
			recursiveFilter = new DisabledInheritanceFilter(disabledInheritanceFilter);
			annotationFilter = new AnnotationFilter();
		}
		
		// Keep track of ID:s of annotations that has been overridden 
		// by project-specific annotations
		Set<Integer> overridden = new HashSet<>();
		
		for (AnnotationSnapshot shot : snapshot.getAnnotations())
		{
			int shotProjectId = shot.getProjectId();
			// Ignore all annotations not belonging to the specified project
			if (projectId != null && shotProjectId != projectId) continue;
			// Ignore project-specfic annotations not belonging to the current project
			if (shotProjectId != 0 && shotProjectId != activeProjectId) continue;
			
			// Ignore default annotations that already have a project-specific value
			// NOTE! Depends on correct sort order in snapshots
			if (overridden.contains(shot.getThisAnnotationId())) continue;
			
			if (shot.getSource() == Annotation.Source.PRIMARY)
			{
				// Primary annotation
				if (!searchPrimary) continue; // with the next snapshot in the list
				
				if (filter.evaluate(shot) && shot.hasPermission(dc, Permission.READ)) 
				{
					shot.setItem(snapshot.getItemId(), snapshot.getItemType());
					result.add(shot);
					// This is a project-specific annotation overriding a default value
					if (shotProjectId != 0) overridden.add(shot.getOverrideId());
				}
			}
			else
			{
				if (!searchInherited) continue; // with the next snapshot in the list
				
				// If the filter matches we must recursively load the inherited annotations
				if (filter.evaluate(shot))
				{
					shot.setItem(snapshot.getItemId(), snapshot.getItemType());
					
					AnnotationSetSnapshot recursive = getSnapshot(dc, shot.getInheritedAnnotationSetId());
					AnnotationSnapshot inheritedFrom = null;
					if (recursive != null)
					{
						// The main filter has already matched this annotation
						// Use the recursive filter to load the specific annotation only
						annotationFilter.setAnnotationId(shot.getInheritedAnnotationId());
						recursiveFilter.setParentFilter(annotationFilter);
						
						List<AnnotationSnapshot> primary = findAnnotations(dc, recursive, recursiveFilter, false);
						if (primary.size() > 0)
						{
							inheritedFrom = primary.get(0);
						}
					}
					shot.setInheritedFrom(inheritedFrom);
					if (inheritedFrom != null || shot.getSource() == Annotation.Source.CLONED)
					{
						result.add(shot);
						if (shotProjectId != 0) overridden.add(shot.getOverrideId());
					}
				}
			}			
		}
		return result;
	}
	
	/**
		@see #findAnnotations(DbControl, Annotatable, Filter, boolean, boolean)
	*/
	public List<AnnotationSnapshot> findAnnotations(DbControl dc, Annotatable item, 
			Filter<? super AnnotationSnapshot> filter, boolean searchInherited)
	{
		return findAnnotations(dc, item, filter, true, searchInherited);
	}
	
	/**
		Utility method for calling <code>getSnapshot</code> and <code>findAnnotations</code>
		in one go.
		@return A list with the annotation snapshots (an empty list is returned if the
			item doesn't have any annotations)
		@see #getSnapshot(DbControl, int)
		@see #findAnnotations(DbControl, AnnotationSetSnapshot, Filter, boolean, boolean)
		@since 3.5
	*/
	public List<AnnotationSnapshot> findAnnotations(DbControl dc, Annotatable item, 
		Filter<? super AnnotationSnapshot> filter, boolean searchPrimary, boolean searchInherited)
	{
		if (!item.isAnnotated()) return Collections.emptyList();
		AnnotationSetSnapshot snapshot = getSnapshot(dc, item.getAnnotationSet().getId());
		return findAnnotations(dc, snapshot, filter, searchPrimary, searchInherited);
	}

}
