/**
	$Id: StreamCacher.java 7551 2018-12-12 14:41:10Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import net.sf.basedb.core.File;

/**
	Caches the contents of a stream to a local temporary file. The 
	{@link #getNewStream()} method will create a stream that appears
	to be reading from the start of the master stream, except that
	it will read cached data to begin with and continue with the master
	stream only if needed. 
	<p>
	
	This class is useful when you know (or suspect) that you need
	to read the same stream multiple times, but the stream may be
	"expensive" to retrieve, either in terms of network capacity or
	cpu processing.

	@author Nicklas
	@since 2.16
	@base.modified $Date: 2018-12-12 15:41:10 +0100 (Wed, 12 Dec 2018) $
	@see File#getCachedDownloadStream()
*/
public class StreamCacher
	implements AutoCloseable
{

	// The master stream we are reading from
	private final InputStream master;
	// The splitter stream which will write master-data to the temp-file 
	private final InputStream splitter;
	// The temporary file for storing master-data
	private final java.io.File cacheFile;
	// Stream for writing to the tmp-file
	private final OutputStream cacheOut;
	
	/**
		Create a new stream cacher for reading from the given master stream.
		
		@param master The master input stream
	*/
	public StreamCacher(InputStream master)
		throws IOException
	{
		this(master, null);
	}
	
	/**
		Create a new stream cacher for reading from the given master stream.
		
		@param master The master input stream
		@param cacheFile Use this file for the cached data, if the file
			already exist it will be overwritten, if this parameter is null
			a random file will be created in the default temporary directory
	*/
	public StreamCacher(InputStream master, java.io.File cacheFile)
		throws IOException
	{
		if (master == null) throw new NullPointerException("master");
		this.master = master;
		if (cacheFile == null)
		{
			cacheFile = java.io.File.createTempFile("cache", ".tmp");
			cacheFile.deleteOnExit();
		}
		this.cacheFile = cacheFile;
		this.cacheOut = FileUtil.getOutputStream(cacheFile);
		this.splitter = new InputStreamSplitter(master, true, false, cacheOut) 
		{
			// Ignore close since we want to keep the master stream open.
			@Override
			public void close()
			{}
		};
	}

	
	/**
		Close the master stream and remove the cached file data.
	*/
	@Override
	public void close()
	{
		FileUtil.close(master);
		FileUtil.close(cacheOut);
		cacheFile.delete();
	}
	
	/**
		Get a new stream that appears to be reading from the start of the
		master stream. The new stream will read cached data (if any), and
		only when the end of that data is reached it will start reading from
		the master stream. The data from the master stream is automatically
		copied to the cache.
		
		@return An Input stream
	*/
	public InputStream getNewStream()
		throws IOException
	{
		cacheOut.flush(); // Make sure anything from previous reads is flushed to the temp file
		return new InputStreamCombiner(FileUtil.getInputStream(cacheFile), splitter);
	}
	
	
	
	
}
