/*
	$Id: Values.java 7828 2020-06-18 06:20:57Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.util.Date;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.util.Collection;

import net.sf.basedb.util.formatter.Formatter;

/**
	This class contains a set of static methods that may be useful 
	for handling strings, ints and other values.
	
	@author Nicklas
	@version 2.0
	@base.modified $Date: 2020-06-18 08:20:57 +0200 (Thu, 18 Jun 2020) $
*/
public class Values
{

	/**
		Convert a string to an integer.

		@param value The string to convert
		@return The converted value, or 0 if the string could not be converted
	*/
	public static final int getInt(String value)
	{
		return getInt(value, 0);
	}

	/**
		Convert a string to an integer.

		@param value The string to convert
		@param defaultValue The value to return if the string cannot be converted
		@return The converted value or the default value
	*/
	public static final int getInt(String value, int defaultValue)
	{
		if (value != null)
		{
			try { return Double.valueOf(value).intValue(); }
			catch (Throwable t) {}
		}
		return defaultValue;
	}
	
	public static final Integer getInteger(String value, Integer defaultValue)
	{
		if (value != null)
		{
			try { return Double.valueOf(value).intValue(); }
			catch (Throwable t) {}
		}
		return defaultValue;
	}
	
	/**
		Convert an array of strings to an array of Integer:s. Typically used
		when you have an array of item id:s and need to use them in a query.
		Positions with unconvertable values are skipped.
		@param values The array of strings
		@return The converted values
	*/
	public static final Integer[] getInt(String[] values)
	{
		Integer[] result;
		if (values != null)
		{
			Integer[] temp = new Integer[values.length];
			int j = 0;
			for (String v : values)
			{
				Integer i = getInteger(v, null);
				if (i != null)
				{
					temp[j++] = i;
				}
			}
			if (j < values.length)
			{
				result = new Integer[j];
				System.arraycopy(temp, 0, result, 0, j);
			}
			else
			{
				result = temp;
			}
		}
		else
		{
			result = new Integer[0];
		}
		return result;
	}

	/**
		Convert a string to a long.

		@param value The string to convert
		@return The converted value, or 0 if the string could not be converted
	*/
	public static final long getLong(String value)
	{
		return getLong(value, 0);
	}
	
	/**
		Convert a string to an long.

		@param value The string to convert
		@param defaultValue The value to return if the string cannot be converted
		@return The converted value or the default value
	*/
	public static final long getLong(String value, long defaultValue)
	{
		if (value != null)
		{
			try { return Long.parseLong(value); }
			catch (Exception ex) {}
		}
		return defaultValue;
	}

	/**
		Convert a string to an long.

		@param value The string to convert
		@param defaultValue The value to return if the string cannot be converted
		@return The converted value or the default value
	*/
	public static final Long getLong(String value, Long defaultValue)
	{
		if (value != null)
		{
			try { return Long.parseLong(value); }
			catch (Exception ex) {}
		}
		return defaultValue;
	}
	
	/**
		Convert an array of strings to an array of Long:s.
		Positions with unconvertable values are skipped.
		@param values The array of strings
		@return The converted values
	*/
	public static final Long[] getLong(String[] values)
	{
		Long[] result;
		if (values != null)
		{
			Long[] temp = new Long[values.length];
			int j = 0;
			for (String v : values)
			{
				Long i = getLong(v, null);
				if (i != null)
				{
					temp[j++] = i;
				}
			}
			if (j < values.length)
			{
				result = new Long[j];
				System.arraycopy(temp, 0, result, 0, j);
			}
			else
			{
				result = temp;
			}
		}
		else
		{
			result = new Long[0];
		}
		return result;
	}

	
	/**
		Convert a string to an float.

		@param value The string to convert
		@return The converted value, or 0 if the string could not be converted
	*/
	public static final float getFloat(String value)
	{
		return getFloat(value, 0);
	}

	/**
		Convert a string to an integer.

		@param value The string to convert
		@param defaultValue The value to return if the string cannot be converted
		@return The converted value or the default value
	*/
	public static final float getFloat(String value, float defaultValue)
	{
		if (value != null)
		{
			try { return Float.parseFloat(value); }
			catch (Exception ex) {}
		}
		return defaultValue;
	}

	public static final Float getFloat(String value, Float defaultValue)
	{
		if (value != null)
		{
			try { return Float.parseFloat(value); }
			catch (Exception ex) {}
		}
		return defaultValue;
	}

	/**
		Convert an array of strings to an array of Float:s. 
		@param values The array of strings
		@return The converted values
	*/
	public static final Float[] getFloat(String[] values)
	{
		Float[] result;
		if (values != null)
		{
			Float[] temp = new Float[values.length];
			int j = 0;
			for (String v : values)
			{
				Float i = getFloat(v, null);
				if (i != null)
				{
					temp[j++] = i;
				}
			}
			if (j < values.length)
			{
				result = new Float[j];
				System.arraycopy(temp, 0, result, 0, j);
			}
			else
			{
				result = temp;
			}
		}
		else
		{
			result = new Float[0];
		}
		return result;
	}

	/**
		Convert a string to a double.

		@param value The string to convert
		@return The converted value, or 0 if the string could not be converted
	*/
	public static final double getDouble(String value)
	{
		return getDouble(value, 0);
	}

	/**
		Convert a string to a double.

		@param value The string to convert
		@param defaultValue The value to return if the string cannot be converted
		@return The converted value or the default value
	*/
	public static final double getDouble(String value, double defaultValue)
	{
		if (value != null)
		{
			try { return Double.parseDouble(value); }
			catch (Exception ex) {}
		}
		return defaultValue;
	}

	public static final Double getDouble(String value, Double defaultValue)
	{
		if (value != null)
		{
			try { return Double.parseDouble(value); }
			catch (Exception ex) {}
		}
		return defaultValue;
	}

	/**
		Convert an array of strings to an array of Double:s. 
		@param values The array of strings
		@return The converted values
	*/
	public static final Double[] getDouble(String[] values)
	{
		Double[] result;
		if (values != null)
		{
			Double[] temp = new Double[values.length];
			int j = 0;
			for (String v : values)
			{
				Double i = getDouble(v, null);
				if (i != null)
				{
					temp[j++] = i;
				}
			}
			if (j < values.length)
			{
				result = new Double[j];
				System.arraycopy(temp, 0, result, 0, j);
			}
			else
			{
				result = temp;
			}
		}
		else
		{
			result = new Double[0];
		}
		return result;
	}
	/**
		Convert a string to a string, and map NULL to the empty string.
		@param value The string to convert
		@return The original string, or the empty string ("") if the original is null
	*/
	public static final String getString(String value)
	{
		return getString(value, "");
	}

	/**
		Convert a string to a string.

		@param value The string to convert
		@param defaultValue The value to return if the string is null
		@return The original string, or the default value if the original is null
	*/
	public static final String getString(String value, String defaultValue)
	{
		return value == null ? defaultValue : value;
	}
	
	/**
		Convert the empty string ("") to null after removing leading and trailing
		whitespace.
		
		@param value The string to convert
		@return The original string with leading and trailing whitespace has been
			removed or null if the value is the empty string ("")
	*/
	public static final String getStringOrNull(String value)
	{
		if (value != null) value = value.trim();
		return "".equals(value) ? null : value;
	}
	
	public static final String getString(Collection<?> values, String deliminator, boolean skipNull)
	{
		return getString(values, deliminator, skipNull, null);
	}
	
	public static final <T> String getString(Collection<T> values, String deliminator, boolean skipNull, Formatter<? super T> formatter)
	{
		StringBuilder sb = new StringBuilder();
		boolean firstElement = true;
		if (values != null)
		{
			for (T value : values)
			{
				if (value == null)
				{
					if (!skipNull)
					{
						if (!firstElement) sb.append(deliminator);
						firstElement = false;
					}
				}
				else
				{
					String s;
					if (!firstElement) sb.append(deliminator);
					if (formatter != null)
					{
						s = formatter.format(value);
					}
					else 
					{
						s = value.toString();
					}
					sb.append(s);
					firstElement = false;
				}
			}
		}
		return sb.toString();		
	}
	
	/**
		Convert a collection of objects to an array of strings. If a formatter
		is specified, the {@link Formatter#format(Object)} method is used, otherwise
		{@link Object#toString()}.
	 	@param values Collection to be converted. 
	 	@return A String array
	*/
	public static final <T> String[] toStrings(Collection<T> values, Formatter<T> formatter)
	{
		String[] strings = null;
		if (values != null)
		{
			strings = new String[values.size()];
			int index = 0;
			for (T value : values)
			{
				if (value == null)
				{
					strings[index] = null;
				}
				else
				{
					strings[index] = formatter == null ? value.toString() : formatter.format(value);
				}
				++index;
			}
		}
		return strings;		
	}	
	
	/**
		Trims long strings to a maximum of <code>maxLength</code>
		characters.
		@param value The string to trim
	 	@return A string or null if value is null.
	*/
	public static final String trimString(String value, int maxLength)
	{
		if (value == null) return null;
		if (value.length() > maxLength) 
		{
			StringBuilder sb = new StringBuilder(maxLength+3);
			sb.append(value.substring(0, maxLength-1));
			sb.append("...");
			value = sb.toString();
		}
		return value;
	}

	/**
		Convert a string to a boolean. '0', 'no', 'false', '' and the null string are
		converted to FALSE, all other values are converted to TRUE. Case is ignored.
		@param value The value to convert
		@return TRUE or FALSE
	*/
	public static final boolean getBoolean(String value)
	{
		return (value == null
			|| value.equals("")
			|| value.equals("0")
			|| value.equalsIgnoreCase("no")
			|| value.equalsIgnoreCase("false")
			) ? false : true;
	}
	
	/**
		Convert a string to a boolean. If the string is null, the default
		value is returned. '0', 'no', 'false' and '' are
		converted to FALSE, all other values are converted to TRUE.
		@param value The value to convert
		@param defaultValue The value to return if the string is null
		@return TRUE or FALSE
	*/
	public static final boolean getBoolean(String value, boolean defaultValue)
	{
		return value == null ? defaultValue : getBoolean(value);
	}

	public static final Date[] getDate(String[] values, Formatter<Date> dateFormatter)
	{
		Date[] result;
		if (values != null)
		{
			Date[] temp = new Date[values.length];
			int j = 0;
			for (String v : values)
			{
				Date i = dateFormatter.parseString(v);
				if (i != null)
				{
					temp[j++] = i;
				}
			}
			if (j < values.length)
			{
				result = new Date[j];
				System.arraycopy(temp, 0, result, 0, j);
			}
			else
			{
				result = temp;
			}
		}
		else
		{
			result = new Date[0];
		}
		return result;
	}
	
	/**
		Formats a decimal number with the specified number of decimals.

		@param number The number to be formatted
		@param decimals The number of decimals to display, use a negative value to
			display all decimals
		@return A string with the formatted number, or an empty string
			if the number is null
	*/
	public static final String formatNumber(Float number, int decimals)
	{
		return formatNumber(number, decimals, null);
	}
	
	/**
		Formats a decimal number with the specified number of decimals
		and optionally adding a unit.

		@param number The number to be formatted
		@param decimals The number of decimals to display, use a negative value to
			display all decimals
		@param unit A string that is added to the end of the formatted number, or
			null to not add anything
		@return A string with the formatted number, or an empty string
			if the number is null
	*/
	public static final String formatNumber(Float number, int decimals, String unit)
	{
		if (number == null || number.isNaN()) return "";
		double fNumber = Math.abs(number.doubleValue());
		double exp = Math.pow(10, decimals);
		StringBuilder result = new StringBuilder();
		if (number < 0) result.append("-");
		if (decimals == 0)
		{
			result.append(String.valueOf(Math.round(fNumber)));
		}
		else if (decimals < 0)
		{
			result.append(fNumber);
		}
		else
		{
			long lNumber = (long)fNumber;
			long remain = Math.round(fNumber*exp-lNumber*exp);
			if (remain >= exp) 
			{
				remain -= exp;
				lNumber++;
			}
			String theDecimals = Long.toString(remain); 
			result.append(lNumber).append(".");
			for (int i = theDecimals.length(); i < decimals; ++i)
			{
				result.append('0');
			}
			result.append(theDecimals);
		}
		if (unit != null) result.append(unit);
		return result.toString();
	}

	public static final long GB = 1073741824;
	public static final long MB = 1048576;
	public static final long kB = 1024;
	
	/**
		Formats a value using units of bytes, kilobytes, megabytes or gigabytes.

		@param bytes The value to format
		@return A string with a decimal value follwed by bytes, Kb, Mb or Gb
			or an empty string if the bytes value is null
	*/
	public static final String formatBytes(Long bytes)
	{
		return formatBytes(bytes, 1);
	}
	
	/**
		Formats a value using units of bytes, kilobytes, megabytes or gigabytes.
		@param bytes The value to format, -1 = unknown
		@param decimals The number of decimals to display in KB, MB, and GB values
		@since 2.6.2
	*/
	public static final String formatBytes(Long bytes, int decimals)
	{
		if (bytes == null) return "";
		if (bytes >= GB)
		{
			float gb = (float)bytes / GB;
			return formatNumber(gb, decimals)+" GB";
		}
		else if (bytes >= MB)
		{
			float mb = (float)bytes / MB;
			return formatNumber(mb, decimals)+" MB";
		}
		else if (bytes >= kB)
		{
			float kb = (float)bytes / kB;
			return formatNumber(kb, decimals)+" kB";
		}
		else if (bytes >= 0)
		{
			return bytes + " bytes";
		}
		else
		{
			return "unknown";
		}
	}
	
	private static final Pattern BYTES_REGEXP = Pattern.compile("(\\d+\\.?\\d*)\\s*(G|M|K)?(b|bytes)?", Pattern.CASE_INSENSITIVE);

	/**
		Parses a string containing the number of bytes formatted as
		the {@link #formatBytes(Long)} method.
		@param fBytes The string containing the value to parse
		@param defaultValue This value is returned if the string cannot be parsed
	 	@return Bytes as a Long object. 
	*/
	public static final Long parseBytes(String fBytes, Long defaultValue)
	{
		if (fBytes == null) return defaultValue;
		Long bytes = defaultValue;
		Matcher m = BYTES_REGEXP.matcher(fBytes);
		if (m.matches())
		{
			double f = Double.parseDouble(m.group(1));
			String unit = m.group(2);
			if ("G".equalsIgnoreCase(unit))
			{
				bytes = (long)(f * GB);
			}
			else if ("M".equalsIgnoreCase(unit))
			{
				bytes = (long)(f * MB);
			}
			else if ("K".equalsIgnoreCase(unit))
			{
				bytes = (long)(f * kB);
			}
			else
			{
				bytes = (long)(f);
			}
		}
		return bytes;
	}

	/**
		Formats a number of seconds to a more proper time format. Example:
		7925 -&gt; 2 hours 12 minutes 5 seconds

		@param seconds The number of seconds
		@return The formatted string
	*/
	public static final String formatTime(long seconds)
	{
		StringBuilder sb = new StringBuilder();
		if (seconds >= 86400L)
		{
			long days = seconds / 86400;
			seconds = seconds - 86400L*days;
			sb.append(days);
			sb.append(days == 1 ? " day " : " days ");
		}
		if (seconds >= 3600L)
		{
			long hours = seconds / 3600;
			seconds = seconds - 3600L*hours;
			sb.append(hours);
			sb.append(hours == 1 ? " hour " : " hours ");
		}
		if (seconds >= 60L)
		{
			long minutes = seconds / 60;
			seconds = seconds - 60L*minutes;
			sb.append(minutes);
			sb.append(minutes == 1 ? " minute " : " minutes ");
		}
		sb.append(seconds);
		sb.append(seconds == 1 ? " second" : " seconds");
		return sb.toString();
	}

	/**
		Insert the values of parameters into a template message.
		@param message The message template, with placeholders of the
			form {1}, {2}, etc. that will be replaced with the values
			of the parameters
		@param parameters The parameters values that will replace the placeholders
			in the template message
		@return The message with the placeholders replaced by the parameters
	*/
	public static final String formatMessage(String message, Object... parameters)
	{
		if ((parameters != null) && (parameters.length > 0))
		{
			for (int i = parameters.length; i > 0; i--)
			{
				message = message.replaceAll("\\{"+i+"\\}", "<span class=\"parameter\">"+parameters[i-1].toString()+"</span>");
			}
		}
		return message;
	}
	
	/**
		Trim a string to a maximum length. If the input string is longer than
		maxLength the string is cut of at the end and ... is added to it.
	 	@param s String to be trimmed. 
	 	@param maxLength The maximum length the string is allowed to be.
	 	@return A string or null if s is null.
	*/
	public static String trim(String s, int maxLength)
	{
		if (s != null && s.length() > maxLength)
		{
			s = s.substring(0, maxLength - 3) + "...";
		}
		return s;
	}
	
}

