/**
	$Id:ExtensionsInvoker.java 4187 2008-03-20 11:15:25Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions;

import java.util.Collection;
import java.util.Iterator;

import net.sf.basedb.util.extensions.events.EventType;

/**
	Object of this class handles a single invokation of the extensions
	for one or several extension points. Call 
	{@link Registry#useExtensions(ClientContext, ExtensionsFilter, String...)}
	to create an invoker object. With this object you can
	{@link #iterate()} over all actions created by the extensions, or
	use one of the render methods to render all actions automatically.

	<p>
	The {@link #renderDefault()} method uses the default render objects 
	registered with the extension points and/or extensions (see 
	{@link ExtensionPoint#getRendererFactory()} and {@link Extension#getRendererFactory()}).
	Since the registration of rendering factories are optional, the
	<code>renderDefault()</code> method only works with extensions that have
	done so. If an action can't be associated with a renderer, an exception
	is thrown.

	<p>
	If there is no default rendering factory or if you want to override the default
	renderer, you can use the {@link #render(Renderer)} method. This will force
	the use of the specified renderer for all extensions, ignoring all renderers
	set by extension points or extensions.

	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public class ExtensionsInvoker<A extends Action>
	implements Iterable<A>
{

	private final Collection<ExtensionContext<A>> contexts;
	private boolean clearErrors;
	
	/**
		Creates a new invoker object.
	*/
	ExtensionsInvoker(Collection<ExtensionContext<A>> contexts)
	{
		this.contexts = contexts;
	}
	
	/**
		Set a flag to indicate if existing errors should be cleared 
		before an action is rendered. If this setting is not enabled (default)
		old errors are never cleared.
		@param clearErrors TRUE to clear old errors, FALSE to keep them
		@since 2.9
	*/
	public void setClearErrors(boolean clearErrors)
	{
		this.clearErrors = clearErrors;
	}

	/**
		Get the number of enabled extensions in this invokation.
		Note that this number doesn't have to correspond to the actual
		number of actions since an extension may choose to create 0, 1 or
		more actions.
		@return The number of enabled extensions
		@since 2.12
	*/
	public int getNumExtensions()
	{
		return contexts.size();
	}
	
	/**
		Create an iterator that iterates over all {@link Action}:s created
		by the extensions. The iterator may return zero actions if none
		has been created. 
		@return An iterator
	*/
	public ActionIterator<A> iterate()
	{
		return new ActionIterator<A>(contexts.iterator());
	}
	
	/**
		Render all actions using the renderers created by the
		default renderer factories as specified by the registered 
		extension points and/or extensions.
		@see ExtensionPoint#getRendererFactory()
		@see Extension#getRendererFactory()
	*/
	public void renderDefault()
	{
		renderDefault(null);
	}
	
	/**
		Render all actions by wrapping the default renderers by the
		given wrapping renderer. 
		@param wrapper A wrapping renderer, or null to not wrap
			the default renderers
		@since 2.12
	*/
	public void renderDefault(WrappingRenderer<A> wrapper)
	{
		ActionIterator<A> it = iterate();
		while (it.hasNext())
		{
			A action = null;
			Renderer<? super A> renderer = null;
			try
			{
				if (clearErrors) it.clearError();
				action = it.next();
				renderer = it.getRenderer();
				if (renderer == null) 
				{
					throw new NullPointerException("No renderer for extension: " + it.getExtension());
				}
				if (wrapper != null)
				{
					wrapper.setParent(renderer);
					renderer = wrapper;
				}
				renderer.render(action);
			}
			catch (Throwable t)
			{
				it.setError("Could not render action '" + action + "' with renderer '" + 
						renderer + "'", t);
			}
		}
	}

	
	/**
		Render all actions using a specific renderer. The given 
		renderer will be used even if an extension point has provided
		a renderer factory and disallowed overriding it.
		
		@see ExtensionPoint#getRendererFactory()
		@see ExtensionPoint#allowRendererOverride()
	*/
	public void render(Renderer<A> renderer)
	{
		renderWithEvent(renderer, null);
	}
	
	/**
		Render all actions using a specific renderer and send the specified event
		to event handlers.
		@param event The event to send or null to not send any event
		@since 3.10
	*/
	public void renderWithEvent(Renderer<A> renderer, EventType event)
	{
		ActionIterator<A> it = iterate();
		while (it.hasNext())
		{
			A action = null;
			try
			{
				if (clearErrors) it.clearError();
				action = it.next();
				renderer.render(action);
				if (event != null) it.handleEvent(event);
			}
			catch (Throwable t)
			{
				it.setError("Could not render action '" + action + "' with renderer '" + 
						renderer + "'", t);
			}
		}
		
	}

	@Override
	public Iterator<A> iterator()
	{
		return iterate();
	}
}
