/**
	$Id: FirstPassSectionSpotsParser.java 7665 2019-03-20 13:31:03Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.importer.spotdata;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.basefile.BaseFileParser;
import net.sf.basedb.util.basefile.BaseFileSectionParser;
import net.sf.basedb.util.importer.spotdata.BaseFileInfo.ChildBioAssay;
import net.sf.basedb.util.importer.spotdata.BaseFileInfo.SpotSectionInfo;
import net.sf.basedb.util.parser.FlatFileParser;

/**
	Parses a 'spots' section of a BASEfile with bioassay set spot
	data. This section is optional but can also appear more than
	once (eg. serial BASEfile). The section must have an 'assays'
	header that contains the id:s of the bioassays that have spot
	data in this section. The id values should either reference an
	id from the 'assays' section or the id of an existing bioassay
	in the database.
	<p>
	
	The section must also contain a 'columns'	header that includes at 
	least 'position', 'reporter' and 'assayData' columns. The 'assayData' 
	is a meta column that expands to the
	columns defined by the required 'assayFields' header. This
	header must define as many 'intensityN' columns as there are
	channels in the experiment. For two-channel data, it may instead
	optionally use M (l2ratio1_2) and A (l10intgmean1_2). Columns with extra
	values can be defined by including a 'setExtraFloats' header.
	This header just enumerates the extra columns. A macthing column
	must also be presen in the 'assayFields' header.
	<p>
	NOTE! Column names can be redefined by calling
	{@link BaseFileParser#setRedefinedColumnName(String, String, String)}.
	<p>
	Here is an example of the headers and the first data line of a matrix 
	BASEfile which has two two-channel assays and one extra value column.
	
	<pre class="code">
section	spots
assays	101	102
columns	position	reporter	assayData
assayFields intensity1 intensity2 extra1
setExtraFloats extra1
%
1	1	1.0	2.0	0.5	2.0	3.0	0.2
</pre>

	The first pass will check the headers and extract position/reporter
	mappings. This is important because we need to know if the child
	bioassay set has the same or a different position/reporter mapping
	than the parent bioassay set.

	@author Nicklas
	@version 2.14
	@base.modified $Date: 2019-03-20 14:31:03 +0100 (Wed, 20 Mar 2019) $
*/
public class FirstPassSectionSpotsParser
	implements BaseFileSectionParser
{

	/**
		The section of the BASEfile this parser can parse, eg. 'spots'.
	*/
	public static final String SECTION = "spots";
	
	private final DbControl dc;
	private final BaseFileInfo info;
	private final BioAssaySet parent;
	private final String totalBytes;
	private final long progressReportInterval;
	private long nextProgressReport;
	
	private int sectionCount;
	private boolean mapZeroToNull;
	
	private List<SpotIntensityParser> spotIntensityParsers;
	
	/**
		Creates a new 'section spot' parser for the first pass. This
		parser will extract header information and position/reporter
		mapping information.
		
		@param dc A DbControl to use for database access
		@param info Information about the file we are parsing
		@param parent The parent bioassay set
	*/
	public FirstPassSectionSpotsParser(DbControl dc, BaseFileInfo info, BioAssaySet parent)
	{
		this.dc = dc;
		this.info = info;
		this.parent = parent;
		this.progressReportInterval = info.getSize() / 100;
		this.nextProgressReport = progressReportInterval;
		this.totalBytes = Values.formatBytes(info.getSize());
		int channels = parent.getRawDataType().getChannels();
		if (channels == 2) addSpotIntensityParser(new MAParser());
		addSpotIntensityParser(new GenericIntensityParser(channels));
	}
	
	/*
		From the BaseFileSectionParser interface
		----------------------------------------
	*/
	@Override
	public void parseSection(BaseFileParser parser, FlatFileParser ffp)
		throws IOException
	{
		++sectionCount;
		
		// If this is the first 'spots' section we must load some
		// info from the parent bioassay set so we can check if
		// the position/reporter mapping has changed. If it has
		// we are going to need a new datacube. 
		if (sectionCount == 1)
		{
			info.loadParentReporterPositions(dc, parent);
			mapZeroToNull = info.getParentHasNullReporter() && !info.getParentHasZeroReporter();
		}
	
		String filename = info.getName();
		ffp.parseHeaders();
		int line = ffp.getParsedLines();
		SpotSectionInfo ssInfo = new SpotSectionInfo();

		// Get the "position", "reporter" and "assayData" columns from the "columns" header
		List<String> columns = parser.getRequiredHeader(ffp, "columns", "\\t", SECTION, filename);
		int positionIndex = parser.getRequiredIndex(columns, 
				parser.getRedefinedColumnName(SECTION, "position"), "columns", SECTION, line, filename);
		int reporterIndex = parser.getRequiredIndex(columns, 
				parser.getRedefinedColumnName(SECTION, "reporter"), "columns", SECTION, line, filename);
		int assayDataIndex = parser.getRequiredIndex(columns, 
				parser.getRedefinedColumnName(SECTION, "assayData"), "columns", SECTION, line, filename);

		ssInfo.setPositionIndex(positionIndex);
	
		// Check the "assays" header
		List<String> assays = parser.getRequiredHeader(ffp, "assays", "\\t", SECTION, filename);
		List<Integer> assayIds = new ArrayList<Integer>(assays.size());
		for (String assay : assays)
		{
			Integer assayId = Values.getInt(assay);
			assayIds.add(assayId);
			if (info.addChildAssay(new ChildBioAssay(assayId)))
			{
				info.checkBioAssay(dc, assayId, line);
			}
		}
		ssInfo.setAssays(assayIds);
		
		// Check the "assayFields" header
		List<String> assayFields = parser.getRequiredHeader(ffp, "assayFields", "\\t", SECTION, filename);
		if (spotIntensityParsers != null)
		{
			for (SpotIntensityParser spiParser : spotIntensityParsers)
			{
				if (spiParser.hasRequiredAssayFields(parser, ffp, assayFields))
				{
					ssInfo.setSpotIntensityParser(spiParser);
					break;
				}
			}
		}
		if (ssInfo.getSpotIntensityParser() == null)
		{
			throw new BaseException("Can't find any intensity columns in 'assayFields' header " +
					"in section 'spots' at line " + line + " in file '" + filename + "'");
		}
		ssInfo.setAssayFields(assayFields);
		ssInfo.setFirstAssayFieldIndex(assayDataIndex);
		
		// Check the "setExtraFloats" header
		String extraFloats = ffp.getHeader("setExtraFloats");
		if (extraFloats != null)
		{
			List<String> extraFloatColumns = Arrays.asList(extraFloats.split("\\t"));
			for (String ef : extraFloatColumns)
			{
				int index = assayFields.indexOf(ef);
				if (index < 0)
				{
					throw new BaseException(
						"Can't find extra float column '" + ef + "' in 'assayFields' " +
						"header in section 'spots' at line " + line + " in file '" + filename + "'");
				}
				info.addExtraFloat(dc, ef);
				ssInfo.addExtraFloatParser(new ExtraFloatParser(ef, index));
			}
		}
		info.addSpotSectionInfo(sectionCount, ssInfo);
		
		// Parse data and check if each position has same reporter as before
		FlatFileParser.Data data;
		while ((data = ffp.nextData()) != null)
		{
			ThreadSignalHandler.checkInterrupted();
			
			long parsedBytes = ffp.getParsedBytes();
			line = ffp.getParsedLines();
			if (parsedBytes >= nextProgressReport)
			{
				nextProgressReport = parsedBytes + progressReportInterval;
				parser.setProgress(parsedBytes, "Importing spot data (first pass): " + 
						Values.formatBytes(parsedBytes) + " of " + totalBytes);
			}
			
			// Get the position and reporter
			Integer position = data.getInt(positionIndex);
			Integer newReporterId = data.getInt(reporterIndex);
			// Error if position is missing
			if (position == null)
			{
				throw new BaseException("Missing or invalid value for 'position' " +
						"in section 'spots' at line " + line + " in file '" + filename + "'");
			}
			// Convert 0 -> null since some BASE1 plug-ins convert null -> 0
			if (mapZeroToNull && newReporterId != null && newReporterId.intValue() == 0)
			{
				newReporterId = null;
			}

			// Store this position->reporter mapping. If the same position has
			// been mapped to a different reporter, throw an exception
			Integer registeredReporterId = info.addChildReporter(position, newReporterId);
			if (newReporterId != registeredReporterId)
			{
				throw new BaseException("Invalid value for 'reporter' (" + newReporterId + 
						") for position '" + position + "' in section 'spots' at line " +	
						line + " in file '" + filename + 
						"'. Expected '" + registeredReporterId + "'");
			}
		}
	}
	// ----------------------------------------------------
	
	/**
		Adds a spot intensity parser to the list of used parsers. 
		The first found parser that signals that all required assay fields
		are present will be used to parse the spot intensity data.
		@see GenericIntensityParser
		@see MAParser
	*/
	public void addSpotIntensityParser(SpotIntensityParser parser)
	{
		if (spotIntensityParsers == null)
		{
			spotIntensityParsers = new LinkedList<SpotIntensityParser>();
		}
		spotIntensityParsers.add(parser);
	}
	
}
