/**
	$Id: AnnotationValidator.java 6960 2015-10-01 10:59:58Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.Annotation;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.snapshot.AnnotationFilter;
import net.sf.basedb.core.snapshot.AnnotationSnapshot;
import net.sf.basedb.core.snapshot.SnapshotManager;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.OverviewUtil;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.filter.ItemTypeFilter;

/**
	Validator implementation for annotations. The node containing
	the annotations is expected to live in a folder-type node that
	has an {@link Annotatable} item as it's parent. 
	Validation rules:
	<ul>
	<li>Annotation is a protocol parameter: {@link Validator#ANNOTATION_IS_PARAMETER}
	<li>Annotation is not valid for item type: {@link Validator#ANNOTATION_INVALID_ITEM}
	<li>Annotation has invalid value: {@link Validator#ANNOTATION_INVALID_VALUE}
	<li>Annotation has too many values: {@link Validator#ANNOTATION_TOO_MANY_VALUES}
	<li>Access denied to annotation type: {@link Validator#DENIED_ANNOTATIONTYPE}
	<li>Annotation is inherited from non-parent item: {@link Validator#ANNOTATION_INHERIT_FROM_NONPARENT}
	<li>Multiple annotations of same annotation type: {@link Validator#ANNOTATION_INHERIT_MULTIPLE}
	<li>Missing a 'Required by MIAME' annotation: {@link Validator#MISSING_MIAME_ANNOTATION}
	<li>Missing experimental factor value: {@link Validator#MISSING_FACTORVALUE}
	</ul>
	
	@author Nicklas
	@version 2.10
	@base.modified $Date: 2015-10-01 12:59:58 +0200 (Thu, 01 Oct 2015) $
*/
public class AnnotationValidator
	extends BasicNodeValidator<Annotation>
{
	
	public AnnotationValidator()
	{
		super(null, null);
	}
	
	/*
		From BasicNodeValidator class
		-----------------------------
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		super.postValidate(dc, context, node, parentNode);
		if (parentNode.getNodeType() == Node.Type.FOLDER) parentNode = parentNode.getParent();
		Annotatable parent = (Annotatable)parentNode.getItem(dc);
		Item parentItemType = parentNode.getItemType();
		
		Annotation a = (Annotation)node.getItem(dc);
		SnapshotManager manager = context.getSnapshotManager();
		// The primary use of the snapshot is to load the values if the annotation is a primary annotation
		// Thus, we do not search inherited annotations
		List<AnnotationSnapshot> snapshots = manager.findAnnotations(dc, parent, new AnnotationFilter(a), false);
		AnnotationSnapshot snapshot = snapshots.size() > 0 ? snapshots.get(0) : null;
		boolean isPrimary = snapshot != null && snapshot.getSource() == Annotation.Source.PRIMARY;

		AnnotationType at = null;
		try
		{
			at = a.getAnnotationType();
			
			// Check that the (primary) annotation is not a protocol parameter
			if (isPrimary && at.isProtocolParameter())
			{
				context.createFailure(Validator.ANNOTATION_IS_PARAMETER, node, 
					"Annotation '" + at.getName() + "' is a protocol parameter",
					new Fix("Remove annotation value", (BasicItem)parent, a),
					new Fix("Change 'parameter' flag of annotation type", at)
				);
			}

			// Check if the (primary) annotation can be used on this item type
			if (isPrimary && !at.isEnabledForItem(parentItemType))
			{
				context.createFailure(Validator.ANNOTATION_INVALID_ITEM, node, 
					"Annotation '" + at.getName() + "' is invalid for item type: " + parentItemType, 
					new Fix("Remove annotation value", (BasicItem)parent, a),
					new Fix("Add '" + parentItemType.name() + "' to annotation type", at)
				);
			}

			// Check that the (primary) annotation values are valid
			if (isPrimary)
			{
				try
				{
					List<?> values = snapshot.getActualValues();
					if (values != null)
					{
						if (values.size() > at.getMultiplicity())
						{
							context.createFailure(Validator.ANNOTATION_TOO_MANY_VALUES, node, 
								"Annotation '" + at.getName() +  "' have " + values.size() + 
								" values but only " + at.getMultiplicity() + " are allowed",
								new Fix("Remove annotation values", (BasicItem)parent, a),
								new Fix("Increase the 'multiplity' option for the annotation type", at)
							);
						}
						
						for (Object value : values)
						{
							at.validateAnnotationValue(value);
						}
					}
				}
				catch (InvalidDataException ex)
				{
					context.createFailure(Validator.ANNOTATION_INVALID_VALUE, node, ex.getMessage(),
						new Fix("Change annotation value", (BasicItem)parent, a),
						new Fix("Change annotation type restrictions", at)
					);
				}
			}
		}
		catch (PermissionDeniedException ex)
		{
			// When the user doesn't have access to the annotation type 
			// of the annotation we can't validate or check other things
			context.createFailure(Validator.DENIED_ANNOTATIONTYPE, node, null);
		}

		// If the annotation is inherited from a 'parent item', verify that 
		// it is a reachable/existing parent
		if (!isPrimary)
		{
			Set<Annotatable> reachableParents = OverviewUtil.getAnnotatableParents(dc, context, parent);
			try
			{
				Annotation inherited = a.getInheritedFrom();
				if (inherited == null)
				{
					context.createFailure(Validator.ANNOTATION_CLONED_FROM_UNKNOWNPARENT, node, 
						"Cloned annotation '" + (at != null ? 
						at.getName() : a.getValueType().toString()) + "' from unknown parent",
						new Fix("Clone from another item", (BasicItem)parent, a)
					);
				}
				else
				{
					Annotatable inheritedFrom = inherited.getAnnotationSet().getItem();
					if (!reachableParents.contains(inheritedFrom))
					{
						context.createFailure(Validator.ANNOTATION_INHERIT_FROM_NONPARENT, node, 
							"Inheriting annotation '" + (at != null ? 
							at.getName() : a.getValueType().toString()) + "' from non-parent",
							new Fix("Inherit from another item", (BasicItem)parent, a)
						);
					}
					
					// Check if a cloned annotation is out-of-sync
					if (a.getSource() == Annotation.Source.CLONED)
					{
						if (a.getLastUpdate().before(inherited.getLastUpdate()))
						{
							context.createFailure(Validator.ANNOTATION_CLONE_OUTOFSYNC, node, 
								"Cloned annotation '" + (at != null ? 
								at.getName() : a.getValueType().toString()) + "' is out-of-sync with source annotation",
								new Fix("Sync cloned annotation", (BasicItem)parent, a)
							);
						}
					}
				}
			}
			catch (PermissionDeniedException ex)
			{
				context.createFailure(Validator.ANNOTATION_INHERIT_FROM_DENIEDPARENT, node, 
						"Inheriting annotation '" + (at != null ? 
							at.getName() : a.getValueType().toString()) + "' from denied parent",
						new Fix("Inherit from another item", (BasicItem)parent, a)
					);
			}
		}
	}
	
	@Override
	public void postValidateFolder(DbControl dc, OverviewContext context, Node folderNode, Node parentNode) 
	{
		super.postValidateFolder(dc, context, folderNode, parentNode);
		Annotatable parent = (Annotatable)parentNode.getItem(dc);
		
		// Check for 'Required by MIAME' annotations and duplicates
		Set<AnnotationType> miame = OverviewUtil.getMiameAnnotationTypes(dc, context, parentNode.getItemType());
		Set<AnnotationType> all = new HashSet<AnnotationType>();
		Map<AnnotationType, Annotation> duplicates = new HashMap<AnnotationType, Annotation>();
		if (folderNode != null && folderNode.numChildren() > 0)
		{
			for (Node child : folderNode.getChildren())
			{
				Annotation a = (Annotation)child.getItem(dc);
				try
				{
					AnnotationType at = a.getAnnotationType();
					if (!all.add(at)) duplicates.put(at, a);
				}
				catch (PermissionDeniedException ex)
				{}
			}
		}
		
		// Report duplicate annotation types
		if (duplicates.size() > 0)
		{
			for (Map.Entry<AnnotationType, Annotation> dup : duplicates.entrySet())
			{
				context.createFailure(Validator.ANNOTATION_INHERIT_MULTIPLE, parentNode, 
					"Duplicate annotations of type '" + dup.getKey().getName() + "'",
					new Fix("Modify annotations", parentNode.getItem(), dup.getValue())
				);
			}
		}

		// Report missing 'Required by MIAME' annotations
		if (miame != null && miame.size() > 0)
		{
			for (AnnotationType at : miame)
			{
				if (!all.contains(at))
				{
					context.createFailure(Validator.MISSING_MIAME_ANNOTATION, parentNode, 
							"No value for annotation '" + at.getName() + "' (required fo MIAME)", 
							new Fix("Add value for annotation '" + at.getName() + "'", (BasicItem)parent, at, false),
							new Fix("Remove 'Required for MIAME' flag on '" + at.getName() + "'", at)
						);
				}
			}
		}
		
		// Check for missing experimental factor values (if we have an experiment in the parent path)
		if (parentNode.getItemType() == Item.ROOTRAWBIOASSAY)
		{
			Node experimentNode = parentNode.getFirstParent(new ItemTypeFilter(Item.EXPERIMENT));
			if (experimentNode != null)
			{
				Experiment exp = (Experiment)experimentNode.getItem(dc);
				ItemQuery<AnnotationType> query = context.initQuery(exp.getExperimentalFactors());
				ItemResultIterator<AnnotationType> it = query.iterate(dc);
				while (it.hasNext())
				{
					AnnotationType factor = it.next();
					if (!all.contains(factor))
					{
						Fix fix = factor.isEnabledForItem(Item.ROOTRAWBIOASSAY) ?
							new Fix("Add value to experimental factor", (BasicItem)parent, factor, false) :
							new Fix("Inherit annotation from a parent item", (BasicItem)parent, factor, true);
						context.createFailure(Validator.MISSING_FACTORVALUE, parentNode, 
							"Missing value for experimental factor '" + factor.getName() + "'", fix);
					}
				}
			}
		}
	}

	/**
		@return Always "null"
	*/
	@Override
	protected Fix getMissingItemFix(DbControl dc, Node parentNode)
	{
		return null;
	}
	// ----------------------------
	
}
