/**
	$Id$

	Copyright (C) 2012 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import net.sf.basedb.core.BioPlate;
import net.sf.basedb.core.BioPlateType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.MeasuredBioMaterial;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;

/**
	Validator implementation for bioplates. Validation rules:
	<ul>
	<li>Missing bioplate: {@link Validator#MISSING_BIOPLATE}
	<li>Access denied: {@link Validator#DENIED_BIOPLATE}
	</ul>

	@author Nicklas
	@since 3.2
	@base.modified $Date$
*/
public class BioPlateValidator
	extends NameableNodeValidator<BioPlate>
{
	
	public BioPlateValidator()
	{
		super(Validator.MISSING_BIOPLATE, Validator.DENIED_BIOPLATE);
	}

	/* 
		From BasicValidator class
		-------------------------
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode) 
	{
		super.postValidate(dc, context, node, parentNode);

		if (parentNode != null)
		{
			Item parentType = parentNode.getItemType();

			if (parentType == Item.SAMPLE || parentType == Item.EXTRACT)
			{
				checkBioMaterialInWell(dc, context, node, parentNode);
			}
		}
	}
	
	/**
		@return Suggested fix is to put the biomaterial on a bioplate.
	*/
	@Override
	protected Fix getMissingItemFix(DbControl dc, Node bioMaterialNode)
	{
		return new Fix("Put biomaterial on a bioplate", bioMaterialNode.getItem());
	}
	// ----------------------------
	
	/**
		Check that the main type and subtype of the biomaterial is matching the
		requirements of the bioplate type. 
	*/
	public static void checkBioMaterialInWell(DbControl dc, OverviewContext context, Node plateNode, Node bioMaterialNode)
	{
		MeasuredBioMaterial bioMaterial = (MeasuredBioMaterial)bioMaterialNode.getItem(dc);
		BioPlate plate = (BioPlate)plateNode.getItem(dc);
		
		try
		{
			BioPlateType plateType = plate.getBioPlateType();
			
			// Check main biomaterial type
			Item expectedBioMaterialType = plateType.getBioMaterialType();
			Item bioMaterialType = bioMaterial.getType();
			if (expectedBioMaterialType != null && expectedBioMaterialType != bioMaterialType)
			{
				context.createFailure(Validator.INCORRECT_BIOMATERIAL_TYPE, bioMaterialNode, 
					"Bioplate '" + plate.getName() + "' should not be used for " + expectedBioMaterialType.name() + " items",
					new Fix("Move '" + bioMaterial.getName() + "' to another plate", bioMaterial));
			}
			
			// Check biomaterial subtype
			ItemSubtype expectedSubtype = plateType.getItemSubtype();
			ItemSubtype subtype = bioMaterial.getItemSubtype();
			
			if (expectedSubtype != null && !expectedSubtype.equals(subtype))
			{
				Fix subtypeFix = new Fix("Change subtype of '" + bioMaterial.getName() + "' to '" + expectedSubtype.getName() + "'", bioMaterial);
				Fix plateFix = new Fix("Change biomaterial subtype of '" + plateType.getName() + "' to '" + subtype.getName() + "'", plateType);
				if (bioMaterialType == Item.SAMPLE)
				{
					context.createFailure(Validator.INCORRECT_SAMPLE_TYPE, bioMaterialNode, "Expected a sample with subtype: " + expectedSubtype.getName(),
						subtypeFix, plateFix);
				}
				else
				{
					context.createFailure(Validator.INCORRECT_EXTRACT_TYPE, bioMaterialNode, "Expected an extract with subtype: " + expectedSubtype.getName(),
						subtypeFix, plateFix);
				}
			}
		}
		catch (PermissionDeniedException ex)
		{}
	}

}
