/*
	$Id: OneTimeFix.java 7521 2018-11-06 08:00:49Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.install;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.DriverManager;
import java.sql.SQLException;

import net.sf.basedb.core.Application;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.Update;
import net.sf.basedb.core.authentication.LoginRequest;
import net.sf.basedb.util.ConsoleProgressReporter;

/**
	This class is used for one-time fixes that are needed to be
	executed manually before or after an updgrade.

	@author Nicklas
	@version 2.7.2
	@base.modified $Date: 2018-11-06 09:00:49 +0100 (Tue, 06 Nov 2018) $
*/
public class OneTimeFix
{
	public static void main(String[] args)
	{
		
		if (args.length == 0) 
		{
			showUsage();
			return;
		}
		try
		{
			String cmd = args[0];
			if ("remaining_quantity".equals(cmd))
			{
				String login = getOption(args, "-u", "root");
				String pwd = getOption(args, "-p", "root");
				
				String msg = 
					"-------------------------------------------------------------------\n" +
					"This script re-calculates the remaining quantity of all biomaterial\n" +
					"items since it may have become incorrect due to a bug.\n" +
					"For more information see: http://base.thep.lu.se/ticket/1873\n\n" +
					"Press ENTER to continue or CTRL+C to abort\n" +
					"-------------------------------------------------------------------";
				System.out.println(msg);
				waitForEnter();
				InitDB.showDbInfo(null);
				System.out.println("Starting BASE. Please wait...");
				Application.start(false);
				SessionControl sc = Application.newSessionControl(null, null, null);
				sc.login(new LoginRequest(login, pwd));
				int numModified = Update.recalculateRemainingQuantity(new ConsoleProgressReporter(false));
				sc.logout();
				sc.close();
				Application.stop();
			}
			else if ("free_wells".equals(cmd))
			{
				String login = getOption(args, "-u", "root");
				String pwd = getOption(args, "-p", "root");
				
				String msg = 
					"-------------------------------------------------------------------\n" +
					"This script re-calculates the free wells on all bioplate items\n" +
					"since it may have become incorrect due to a bug.\n" +
					"For more information see: http://base.thep.lu.se/ticket/2088\n\n" +
					"Press ENTER to continue or CTRL+C to abort\n" +
					"-------------------------------------------------------------------";
				System.out.println(msg);
				waitForEnter();
				InitDB.showDbInfo(null);
				System.out.println("Starting BASE. Please wait...");
				Application.start(false);
				SessionControl sc = Application.newSessionControl(null, null, null);
				sc.login(new LoginRequest(login, pwd));
				int numModified = Update.recalculateFreeWells(new ConsoleProgressReporter(false));
				sc.logout();
				sc.close();
				Application.stop();
			}
			else
			{
				showUsage();
			}
		}
		catch (Throwable t)
		{
			System.out.println("\n");
			t.printStackTrace();
		}
	}

	private static void showUsage()
	{
		System.out.println("Usage: onetimefix.sh cmd <options>");
		System.out.println("generic options: ");
		System.out.println("    -u <login>    : Login name for the root user account; (root)");
		System.out.println("    -p <password> : Password for the root user account; (root)");
		System.out.println("cmd:");
		System.out.println("    remaining_quantity: Re-calculate remaining quantity of biomaterials");
		System.out.println("    free_wells: Re-calculate free wells on all bioplates");
	}

	public static String getOption(String[] args, String option, String defaultValue)
	{
		String value = defaultValue;
		for (int i = 0; i < args.length; ++i)
		{
			if (option.equals(args[i]))
			{
				if (i+1 < args.length && !args[i+1].startsWith("-"))
				{
					value = args[i+1];
				}
				else
				{
					value = option;
				}
				return value;
			}
		}
		return value;
	}
	
	
	/**
		Get a JDBC {@link Connection} to the current BASE database.
		@since 2.9
	*/
	private static Connection getConnection()
		throws SQLException
	{
		String dbUrl = Config.getString("db.url");
		String dbUser = Config.getString("db.username");
		String dbPassword = Config.getString("db.password");
		return DriverManager.getConnection(dbUrl, dbUser, dbPassword);
	}
	
	/**
		Checks if the database is MySQL.
		@since 2.9
	*/
	private static boolean isMySQL()
		throws Exception
	{
		Connection c = null;
		try
		{
			c = getConnection();
			DatabaseMetaData metaData = c.getMetaData();
			String dbName = metaData.getDatabaseProductName();
			return dbName.contains("MySQL");
		}
		finally
		{
			if (c != null) c.close();
		}
	}
	

	public static void waitForEnter()
	{
		try
		{
			System.in.read();
			System.in.skip(System.in.available());
		}
		catch (Exception ex)
		{}
	}

}

