/*
	$Id: AbstractFlatFileImporter.java 7668 2019-03-21 11:50:52Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.core.NumberOutOfRangeException;
import net.sf.basedb.core.ParameterType;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.FileParameterType;
import net.sf.basedb.core.IntegerParameterType;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.File;
import net.sf.basedb.core.IntegerUtil;
import net.sf.basedb.core.StringTooLongException;
import net.sf.basedb.core.StringUtil;
import net.sf.basedb.core.UnsupportedFileFormatException;
import net.sf.basedb.core.plugin.AbstractPlugin;
import net.sf.basedb.core.plugin.AutoDetectingImporter;
import net.sf.basedb.core.plugin.ImportInputStream;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.signal.SignalException;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.plugins.util.Parameters;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.NumberFormatUtil;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.error.ClassMapErrorHandler;
import net.sf.basedb.util.error.ErrorHandler;
import net.sf.basedb.util.error.SimpleErrorHandler;
import net.sf.basedb.util.formatter.DateFormatter;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.util.jep.JepFunction;
import net.sf.basedb.util.parser.CropStringMapper;
import net.sf.basedb.util.parser.DefaultItemFunction;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.JepMapper;
import net.sf.basedb.util.parser.Mapper;

import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.regex.Pattern;
import java.io.IOException;
import java.io.InputStream;


/**
	An abstract base class for all importers that imports data from one or more
	flat files. The implementation in this class uses a {@link FlatFileParser}
	for parsing the files and uses a callback method {@link #handleData(FlatFileParser.Data)}
	that lets the subclass do whatever it needs to insert a single line
	of data into the database.
	<p>
	The subclass must also generate the {@link net.sf.basedb.core.RequestInformation} object
	for both plugin and job configuration. However, this implementation expects
	to find the regular expression needed for the {@link FlatFileParser} to
	be found in the plugin configuration and the {@link File} to import to
	be found in the job configuration.
	<p>
	All of the needed {@link PluginParameter} objects needed to ask for these 
	parameters are declared as protected variables in this class.

	<pre class="code">
// <b>ReporterFlatFileImporter.java</b>
// RequestInformation object for command = CONFIGURE_PLUGIN
List&lt;PluginParameter&lt;?&gt;&gt; parameters = 
   new ArrayList&lt;PluginParameter&lt;?&gt;&gt;();

// Parser regular expressions
parameters.add({@link #headerRegexpParameter});
parameters.add({@link #dataHeaderRegexpParameter});
parameters.add({@link #dataSplitterRegexpParameter});
parameters.add({@link #ignoreRegexpParameter});
parameters.add({@link #dataFooterRegexpParameter});

// Column mappings
parameters.addAll(allColumnMappings);

// Reporter type
parameters.add(reporterTypeParameter);

configurePlugin = new RequestInformation
(
   Request.COMMAND_CONFIGURE_PLUGIN,
   "Parser settings",
   "Please enter all settings needed by the flat file parser",
   parameters
);
</pre>

	This class implements the <code>invoke</code> method but only for
	the {@link Request#COMMAND_EXECUTE} command. The subclass must override
	the <code>invoke</code> method and provide implementations for the
	other commands ({@link Request#COMMAND_CONFIGURE_PLUGIN} and 
	{@link Request#COMMAND_CONFIGURE_JOB}). The normal implementation of
	would be to store a request parameter.
	<pre class="code">
// <b>ReporterFlatFileImporter.java</b>
if (command.equals(Request.COMMAND_CONFIGURE_JOB))
{
   List&lt;Throwable&gt; errors = validateRequestParameters(jobParameters, request);
   if (errors != null)
   {
      response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
      return;
   }
   storeValue(job, request, {@link #fileParameter});
   response.setDone(null);
}
</pre>

	<p>
	For multi-file support (added in BASE 2.9) the subclass needs to
	override the {@link #getFileIterator()} and {@link #getTotalFileSize()}
	methods.
	
	@base.modified $Date: 2019-03-21 12:50:52 +0100 (Thu, 21 Mar 2019) $
	@author Enell, Nicklas
	@version 2.0
*/
public abstract class AbstractFlatFileImporter 
	extends AbstractPlugin
	implements AutoDetectingImporter, SignalTarget
{

	protected static final ParameterType<String> requiredRegexpType = new StringParameterType(65000, null, true);
	protected static final ParameterType<String> optionalRegexpType = new StringParameterType(255, null, false);
	protected static final ParameterType<Integer> numDataColumnsType = new IntegerParameterType(0, null, null, false);
	protected static final ParameterType<File> fileType = new FileParameterType(null, true, 1);

	/**
		Parameter definition that asks for the regular expression that 
		matches a section line and extracts the section name. This is an 
		optional parameter.
		@see FlatFileParser#setHeaderRegexp(Pattern)
	*/
	protected static final PluginParameter<String> sectionRegexpParameter = new PluginParameter<String>(
		"sectionRegexp",
		"Section",
		"A regular expression that matches a section line and extracts the name part. "+
		"For example: section\\t(.*)",
		optionalRegexpType
	);
	
	/**
		Parameter definition that asks for the regular expression that 
		splits a header line in key/value pair. This is an optional parameter.
		@see FlatFileParser#setHeaderRegexp(Pattern)
	*/
	protected static final PluginParameter<String> headerRegexpParameter = new PluginParameter<String>(
		"headerRegexp",
		"Header",
		"A regular expression that matches a header line and extracts the name and a value parts. " + 
		"For example, split on equal symbol: (.+)=(.*)",
		optionalRegexpType
	);

	/**
		Parameter definition that asks for the regular expression that 
		matches the data header. This is a required parameter for the plugin configuration.
		@see FlatFileParser#setDataHeaderRegexp(Pattern)
	*/
	protected static final PluginParameter<String> dataHeaderRegexpParameter = new PluginParameter<String>(
		"dataHeaderRegexp",
		"Data header",
		"A regular expression that matches the header line just before the data begins. " +
		"For example: Block\\tRow\\tColumn.*",
		requiredRegexpType
	);
		
	/**
		Parameter definition that asks for the regular expression that 
		splits a data line into individual columns. This is a required parameter 
		for the plugin configuration.
		@see FlatFileParser#setDataSplitterRegexp(Pattern)
	*/
	protected static final PluginParameter<String> dataSplitterRegexpParameter = new PluginParameter<String>(
		"dataSplitterRegexp",
		"Data splitter",
		"A regular expression that splits each data line into individual columns. "+
		"For example, split on tabs: \\t",
		new StringParameterType(255, "\\t", true)
	);
	
	/**
		Parameter definition that asks if quotes around quoted values should be removed
		or not.
		@see FlatFileParser#setTrimQuotes(boolean)
	*/
	protected static final PluginParameter<Boolean> trimQuotesParameter = new PluginParameter<Boolean>(
		"trimQuotes",
		"Remove quotes",
		"If true quotes (\" or ') around data value will be removed.",
		new BooleanParameterType(true, false)
	);
	
	/**
		Parameter definition that asks for the regular expression that 
		matches a line that should be ignored. This is an optional parameter 
		for the plugin configuration.
		@see FlatFileParser#setIgnoreRegexp(Pattern)
	*/
	protected static final PluginParameter<String> ignoreRegexpParameter = new PluginParameter<String>(
		"ignoreRegexp",
		"Ignore",
		"A regular expression that matches any line that should be ignored. " +
		"For example, ignore lines starting with #: ^#.*",
		optionalRegexpType
	);

	/**
		Parameter definition that asks for the regular expression that 
		matches the data footer. This is an optional parameter for the plugin configuration.
		@see FlatFileParser#setDataFooterRegexp(Pattern)
	*/
	protected static final PluginParameter<String> dataFooterRegexpParameter = new PluginParameter<String>(
		"dataFooterRegexp",
		"Data footer",
		"A regular expression that matches the first line of non-data after the data lines. " +
		"For example: __END_OF_DATA__",
		optionalRegexpType
	);

	/**
		Parameter definition that asks for the minimum number of columns produced by
		the data splitter regexp for a line to be a data line. This is an optional parameter 
		for the plugin configuration.
		@see FlatFileParser#setMinDataColumns(int)
	*/
	protected static final PluginParameter<Integer> minDataColumnsParameter = new PluginParameter<Integer>(
		"minDataColumns",
		"Min data columns",
		"The minimum number of columns for a line to be counted as a data line.",
		numDataColumnsType
	);
	
	/**
		Parameter definition that asks for the maximum number of columns produced by
		the data splitter regexp for a line to be a data line. This is an optional parameter 
		for the plugin configuration.
		@see FlatFileParser#setMaxDataColumns(int)
	*/
	protected static final PluginParameter<Integer> maxDataColumnsParameter = new PluginParameter<Integer>(
		"maxDataColumns",
		"Max data columns",
		"The maximum number of columns for a line to be counted as a data line, or 0 to allow any number of columns.",
		numDataColumnsType
	);

	/**
		Parameter definition that asks for the file that should be imported.
		This is a required parameter for the job configuration.
	*/
	protected static final PluginParameter<File> fileParameter = new PluginParameter<File>(
		"file",
		"File",
		"The file to import the data from",
		fileType
	);
	
	/**
		Parameter definition for specifying the name of the sheet to use if the selected file
		is an Excel file.
		@since 3.15
	*/
	protected static final PluginParameter<String> excelSheetParameter = new PluginParameter<>(
		"excelSheet", "Excel sheet", 
		"If the selected file is an Excel workbook, this parameter should hold the name of the " +
		"sheet to import from. If not specified, the first sheet is used.", 
		optionalRegexpType);
	
	/**
		Section definition for grouping all file parser settings (ie. regular expressions)
	*/
	protected static final PluginParameter<String> parserSection = new PluginParameter<String>(
		"parserSection",
		"File parser regular expressions",
		"Enter the regular expressions used to find the data in the file.", 
		null
	);
	
	/**
		Section definition for grouping all column mapping expressions
	*/
	protected static final PluginParameter<String> mappingSection = new PluginParameter<String>(
		"mappingSection",
		"Column mapping expressions",
		"Enter expressions for mapping columns in the file to the properties in the database. " +
		"You may use <b>\\0\\</b> or <b>\\ID\\</b> to get the values from the first column and the column " +
		"with the name ID. You can also combine several columns into a single value: " +
		"<b>\\Species\\.\\ClusterID\\</b> " +
		"You can also use an expression for numerical values. For example to get the " +
		"diameter when the file only contains the radius: <b>=2 * col('Radius')</b>", 
		null
	);

	/**
 		Parameter that asks if complex column mappings should be enabled or not.
 		Use {@link #checkColumnMapping(String, boolean, String)} to check a mapping.
		@since 2.4
	*/
	protected static final PluginParameter<String> complexMappings = new PluginParameter<String>(
		"complexExpressions",
		"Complex column mappings",
		"disallow = Only allow simple mappings that are constant value or pick the value " +
		"from one column only, for example, '1.6' or '\\Row\\'\n" +
		"allow = Allow expression and complex mappings, for example, " +
		"'\\Row\\, \\Column\\' or '=2*col('radius')'",
		new StringParameterType(255, "disallow", true, 1, 0, 0, 
			Arrays.asList(new String[] { "disallow", "allow" })
		)
	);
	
	/**
		Section definition for grouping error handling options.
		@see Parameters#errorSection(String, String)
	*/
	protected static final PluginParameter<String> errorSection = 
		Parameters.errorSection(null, null);
	
	/**
		@see Parameters#defaultError(String, String, String, String...)
	*/
	protected static final PluginParameter<String> defaultErrorParameter = 
		Parameters.defaultError(null, null, null);
	
	/**
		@see Parameters#stringTooLongError(String, String, String, String...)
	*/
	protected static final PluginParameter<String> stringTooLongErrorParameter = 
		Parameters.stringTooLongError(null, null, null);
	
	/**
		@see Parameters#invalidUseOfNullError(String, String, String, String...)
	*/
	protected static final PluginParameter<String> invalidUseOfNullErrorParameter =
		Parameters.invalidUseOfNullError(null, null, null);

	/**
		@see Parameters#numberOutOfRangeError(String, String, String, String...)
	 */
	protected static final PluginParameter<String> numberOutOfRangeErrorParameter = 
		Parameters.numberOutOfRangeError(null, null, null);
	
	/**
		@see Parameters#numberFormatError(String, String, String, String...)
	*/
	protected static final PluginParameter<String> numberFormatErrorParameter = 
		Parameters.numberFormatError(null, null, null);
	
	private File currentFile;
	private long totalFileSize;
	private long finishedFileSize;
	private int skippedLines;
	private ClassMapErrorHandler errorHandler;
	private ThreadSignalHandler signalHandler;

	public AbstractFlatFileImporter()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	/**
		Return {@link net.sf.basedb.core.plugin.Plugin.MainType#IMPORT}.
	*/
	@Override
	public Plugin.MainType getMainType()
	{
		return Plugin.MainType.IMPORT;
	}

	/**
		Return TRUE, since the implementation requires it for finding the
		regular expressions used by the {@link FlatFileParser}. If this
		method is overridden and returns FALSE, the subclass must also override
		the {@link #getInitializedFlatFileParser()} method and provide a
		parser with all regular expressions and other options set.
	*/
	@Override
	public boolean requiresConfiguration()
	{
		return true;
	}
	/**
		Implements the {@link Request#COMMAND_EXECUTE} command. Subclasses
		must override this to implement other commands. Subclasses should not
		call this method for other commands than {@link Request#COMMAND_EXECUTE}
		since this method will set an error response status.
	*/
	@Override
	public void run(Request request, Response response, ProgressReporter progress)
	{
		if (signalHandler != null) signalHandler.setWorkerThread(null);
		String command = request.getCommand();
		if (command.equals(Request.COMMAND_EXECUTE))
		{
			createLogFile((String)job.getValue(Parameters.LOGFILE_PARAMETER));
			boolean hasCalledFinish = false;
			int numSuccess = 0;
			int numTotal = 0;
			String finalMessage = null;
			DbControl dc = null;
			try
			{
				start();
				String message = null;
				finishedFileSize = 0;
				Iterator<File> files = getFileIterator();
				totalFileSize = getTotalFileSize();
				dc = sc.newDbControl();
				
				while (files.hasNext())
				{
					numTotal++;
					InputStream in = null;
					try
					{
						currentFile = File.getById(dc, files.next().getId());
						log("Parsing file: " + currentFile.getName());
						in = currentFile.getDownloadStream(0);
						boolean importable = isImportable(in);
						FileUtil.close(in);
						if (!importable)
						{
							throw new UnsupportedFileFormatException("File '"+currentFile.getName()+
								"' doesn't match the expected file format.");
						}
						in = currentFile.getDownloadStream(0);
						doImport(in, progress);
						FileUtil.close(in);
						message = getSuccessMessage(skippedLines);
						if (message == null) 
						{
							message = "File '" + currentFile.getName() + 
								"' imported successfully. " + 
								skippedLines + " lines skipped.";
						}
						log(message);
						numSuccess++;
					}
					catch (SignalException s)
					{
						// The user aborted the plug-in -> always exit
						throw s;
					}
					catch (Throwable t)
					{
						if (!continueWithNextFileAfterError(t)) throw t;
						log("Failed", t);
					}
					finally
					{
						if (currentFile != null && currentFile.getSize() != -1)
						{
							finishedFileSize += currentFile.getSize();
						}
						FileUtil.close(in);
						log("--------------------------------------");
					}
				}
				hasCalledFinish = true;
				finalMessage = finish(null);
				if (numSuccess == 0)
				{
					if (finalMessage == null)
					{
						finalMessage = "No files could be imported";
					}
					response.setError(finalMessage + ": " + message, null);
				}
				else
				{
					if (finalMessage == null && numTotal > 1)
					{
						finalMessage = numSuccess + " of " + numTotal + " files could be imported";
					}
					response.setDone(finalMessage == null ? message : finalMessage);
				}
			}
			catch (Throwable t)
			{
				String message = hasCalledFinish ? null : finish(t);
				if (message == null) message = t.getMessage();
				log("Failed: " + message, t);
				response.setError(message, Arrays.asList(t));
			}
			finally
			{
				if (dc != null) dc.close();
				if (finalMessage != null) log(finalMessage);
				closeLogFile();
			}
		}
		else
		{
			response.setError("Unknown command: "+command, null);
		}
	}
	// -------------------------------------------

	/*
		From the AutoDetctingImporter interface
		-------------------------------------------
	*/
	@Override
	public final boolean isImportable(InputStream in)
		throws BaseException
	{
		FlatFileParser ffp = getInitializedFlatFileParser();
		try
		{
			String charset = null;
			if (in instanceof ImportInputStream)
			{
				charset = ((ImportInputStream)in).getCharacterSet();
			}
			if (charset == null) charset = getCharset();
			in = wrapInputStream(in);
			ffp.setInputStream(in, charset);
			ffp.nextSection();
			FlatFileParser.LineType result = ffp.parseHeaders();
			if (result == FlatFileParser.LineType.UNKNOWN)
			{
				return false;
			}
			else
			{
				return isImportable(ffp);
			}
		}
		catch (IOException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	@Override
	public void doImport(InputStream in, ProgressReporter progress)
		throws BaseException
	{
		setUpErrorHandling();
		FlatFileParser ffp = getInitializedFlatFileParser();
		FlatFileParser.Data dataline = null;
		boolean useNullIfException = "null".equals(job != null ? job.getValue("numberFormatError") : "");
		ffp.setUseNullIfException(useNullIfException);
		boolean isSuccess = false; // false until we have finished parsing
		try
		{
			String charset = null;
			if (in instanceof ImportInputStream)
			{
				charset = ((ImportInputStream)in).getCharacterSet();
			}
			if (charset == null) charset = getCharset();
			in = wrapInputStream(in);
			ffp.setInputStream(in, charset);
			skippedLines = 0;
			begin(ffp);
			do
			{

				FlatFileParser.Line section = ffp.nextSection();
				if (section != null && section.type() == FlatFileParser.LineType.SECTION)
				{
					if (progress != null) 
					{
						progress.display(getProgress(ffp), "Parsing section '"+section.name()+
								"' in file '" + currentFile.getName() + "'...");
					}
					try
					{
						handleSection(section);
					}
					catch (Throwable t)
					{
						try
						{
							if (errorHandler == null)
							{
								throw t;
							}
							else
							{
								errorHandler.handleError(t);
							}
							log("Skipped: " + t.getMessage(), section);
							skippedLines++;
						}
						catch (Throwable t2)
						{
							throw new BaseException(t2.getMessage() + 
								" on line " + section.lineNo() + 
								" in file '" + currentFile.getName() + "': " + 
								StringUtil.trimString(section.line(), 20), t2);	
						}
					}
				}
				else
				{
					if (progress != null) 
					{
						progress.display(getProgress(ffp), 
							"Parsing headers in file '" + currentFile.getName() + "'...");
					}
				}
				ThreadSignalHandler.checkInterrupted();
				FlatFileParser.LineType result = ffp.parseHeaders();
				FlatFileParser.Line line = null;
				try
				{
					for (int i = 0; i < ffp.getLineCount(); i++)
					{
						line = ffp.getLine(i);
						if (line.type() == FlatFileParser.LineType.HEADER)
						{
							try
							{
								handleHeader(line);
							}
							catch (Throwable t)
							{
								if (errorHandler == null)
								{
									throw t;
								}
								else
								{
									errorHandler.handleError(t);
								}
								log("Skipped: " + t.getMessage(), line);
								skippedLines++;
							}
						}
					}
					ThreadSignalHandler.checkInterrupted();
				}
				catch (Throwable t)
				{
					throw new BaseException(t.getMessage() + 
						" on line " + (line == null ? "-1" : line.lineNo() + 
						" in file '" + currentFile.getName() + "': " + 
						 StringUtil.trimString(line.line(), 20)), t);			
				}
				if (result == FlatFileParser.LineType.DATA || result == FlatFileParser.LineType.DATA_HEADER)
				{
					if (progress != null) 
					{
						progress.display(getProgress(ffp), 
							"Parsing data in file '" + currentFile.getName() + "'...");
					}
					int progressLine = 0;
					
					beginData();
					try
					{
						while (ffp.hasMoreData())
						{
							dataline = ffp.nextData();
							try
							{
								handleData(dataline);
							}
							catch (Throwable t)
							{
								if (errorHandler == null)
								{
									throw t;
								}
								else
								{
									errorHandler.handleError(t);
								}
								log("Skipped: " + t.getMessage(), dataline);
								skippedLines++;
							}
							progressLine++;
							if (progressLine == 100)
							{
								progressLine = 0;
								if (progress != null)
								{
									progress.display(getProgress(ffp), 
										"Parsing data in file '" + currentFile.getName() + 
										"'... (" + ffp.getParsedLines() + " lines so far)");
								}
							}
							// In case the server is shutting down... throw exception, rollback and quit
							ThreadSignalHandler.checkInterrupted();
						}
					}
					catch (Throwable t)
					{
						throw new BaseException(t.getMessage() + 
							" on line " + (dataline == null ? "-1" : dataline.lineNo() + 
							" in file '" + currentFile.getName() + "': " + 
							StringUtil.trimString(dataline.line(), 20)), t);
					}
				}
			}
			while (ffp.hasMoreSections());
			isSuccess = true;
			if (progress != null) 
			{
				progress.display(getProgress(ffp), 
					"Done parsing file '" + currentFile.getName() + 
					"' ("+ffp.getParsedLines()+" lines)\n");
			}
		}
		catch (IOException ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			end(isSuccess);
		}
	}
	// -------------------------------------------
	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	/**
		Create a new {@link ThreadSignalHandler} that supports
		the {@link net.sf.basedb.core.signal.Signal#ABORT} signal.
		Subclasses may override this to provide another signal 
		handler, or return null if they don't support signals.
		@since 2.6
	*/
	@Override
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------

	/**
		Called once before starting the import. 
		The default implementation does nothing.
		@since 2.9
	*/
	protected void start()
	{}
	
	/**
		Get an iterator that returns the files to be imported. The default
		implementation returns the single file found in the job's "file"
		parameter. Subclasses that needs multi-file/item import should override
		this method to provide their own iterator. They should also override
		the {@link #getTotalFileSize()} method to return sum of all
		file sizes. Eg. {@link File#getSize()}.
		@since 2.9
	*/
	protected Iterator<File> getFileIterator()
	{
		return Collections.singleton((File)job.getValue("file")).iterator();
	}	
	
	/**
		Get the total file size of all files that are going to be imported.
		A subclass that is going to import from multiple files needs to override
		this method. The default implementation return the size of the file
		in the job's "file" parameter.
		@return The sum of the file sizes, or -1 if not known
		@since 2.9
	*/
	protected long getTotalFileSize()
	{
		return ((File)job.getValue("file")).getSize();
	}
	
	/**
		Get the progress of import as a percentage value. The default implementation
		calls sums the file size of the completed files and 
		{@link #getNumBytes(FlatFileParser)}. This values is divided by 
		{@link #getTotalFileSize()}.
		@param ffp The file parser that is used to parsed the file
		@return A value between 0 and 100
		@since 2.6
	*/
	protected int getProgress(FlatFileParser ffp)
	{
		return totalFileSize == -1 ? 
			-1 : 
			(int) (100 * (getNumBytes(ffp) + finishedFileSize) / totalFileSize);
	}
	
	/**
		Get the number of bytes read from the file. The value should indicate the
		how far into the file parsing has proceeded. The default
		implementation calls {@link FlatFileParser#getParsedBytes()}. If a subclass has 
		wrapped the input stream the number of parsed byts may not correspond to the 
		number of bytes read from the file.
		For example, if the file is a compressed file the number of parsed bytes will
		be higher than the number of bytes read from the file.
		@param ffp The file parser that is used to parsed the file
		@return The number of bytes read from the original file
		@see #wrapInputStream(InputStream)
		@see #getProgress(FlatFileParser)
		@since 2.6
	*/
	protected long getNumBytes(FlatFileParser ffp)
	{
		return ffp.getParsedBytes();
	}
	
	/**
		This method is called by the {@link #isImportable(InputStream)} method after
		{@link FlatFileParser#nextSection()} and {@link FlatFileParser#parseHeaders()}
		has been called and if data has been found. Thus, the default implementation of this
		method always returns TRUE. Subclasses may override this method to do more checks, for
		example to make sure certain headers are present or parse more data from the file.
		
		@param ffp The FlatFileParser object used to parse the file
		@return Always TRUE
		@since 2.4
	 */
	protected boolean isImportable(FlatFileParser ffp)
		throws IOException
	{
		return true;
	}
	
	/**
		This method is called before the parser starts reading from the
		input stream. A subclass may override this method to wrap the inputstream
		with a filtering stream, for example, a gzip input stream. The default
		implementation of this method returns the original stream unmodified. If a subclass
		overrides this methods it may also need to override one of the 
		{@link #getNumBytes(FlatFileParser)} or {@link #getProgress(FlatFileParser)} 
		methods.
		@param in The input stream to wrap
		@return The same or a different input stream
		@since 2.6
		@see #getNumBytes(FlatFileParser)
	*/
	protected InputStream wrapInputStream(InputStream in)
		throws IOException
	{
		return in;
	}
	
	/**
		Called just before parsing of the file begins. A subclass
		may override this method if it needs to initialise some
		resources before the parsing starts. Note that this method is called
		once for each file returned by {@link #getFileIterator()}.
		@see #end(boolean)
	*/
	protected void begin(FlatFileParser ffp)
		throws BaseException
	{}
	
	/**
		Called by the parser for every line in the file that is
		a header line.
	*/
	protected void handleHeader(FlatFileParser.Line line)
		throws BaseException
	{}
	
	/**
		Called by the parser for every line in the file that is
		a section line.
	*/
	protected void handleSection(FlatFileParser.Line line)
		throws BaseException
	{}
	
	/**
		Called by the parser after the headers have been parsed but before
		the first data line is parsed.
	 */
	protected void beginData()
		throws BaseException
	{}
	
	/**
		Called by the parser for every line in the file that is
		a data line.
	*/
	protected abstract void handleData(FlatFileParser.Data data)
		throws BaseException;

	/**
		Called just after the last line of the file has been parsed,
		or immediately after an error has ocurred. A subclass should clean
		up any resources aquired in the {@link #begin(FlatFileParser)} method here.
		Note that this metod is called once for every file returned by the
		{@link #getFileIterator()} iterator.
		@param success TRUE if the file was parsed successfully, FALSE otherwise
		@see #begin(FlatFileParser)
	*/
	protected void end(boolean success)
		throws BaseException
	{}
	
	/**
		Called if the parsing was successful to let the subclass generate a simple
		message that is sent back to the core and user interface. An example
		message might by: <code>178 reporters imported successfully</code>.
		The default implementation always return null. Note that this method is called
		once for every file returned by {@link #getFileIterator()}.
		@param skippedLines The number of data lines that were skipped due to errors
	*/
	protected String getSuccessMessage(int skippedLines)
	{
		return null;
	}
	
	/**
		Called once when all files has been imported or when exiting due to an error.
		This method may return a message that is sent back to the core and user interface. 
		If this method returns null the last message returned by {@link #getSuccessMessage(int)}
		is used instead.
		@param t Null if no error has happened
		@return A message or null
		@since 2.9
	*/
	protected String finish(Throwable t)
	{
		return null;
	}
	
	/**
		Create and initialise a flat file parser by setting all regular expressions
		and other options. This implementation gets all parameters from the 
		{@link #job} or {@link #configuration} settings. If a subclass doesn't store the
		parameters there it must override this method and initialise the parser. Note that
		this method is called once for each file returned by the 
		{@link #getFileIterator()} and that a new parser is needed for each file.
		@return An intialised flat file parser
	*/
	protected FlatFileParser getInitializedFlatFileParser()
		throws BaseException
	{
		FlatFileParser ffp = new FlatFileParser();

		ffp.setSectionRegexp(getPattern(sectionRegexpParameter.getName()));
		ffp.setHeaderRegexp(getPattern(headerRegexpParameter.getName()));
		ffp.setDataHeaderRegexp(getPattern(dataHeaderRegexpParameter.getName()));
		ffp.setDataSplitterRegexp(getPattern(dataSplitterRegexpParameter.getName()));
		Boolean trimQuotes = (Boolean)getJobOrConfigurationValue(trimQuotesParameter.getName());
		if (trimQuotes != null) ffp.setTrimQuotes(trimQuotes);
		ffp.setIgnoreRegexp(getPattern(ignoreRegexpParameter.getName()));
		ffp.setDataFooterRegexp(getPattern(dataFooterRegexpParameter.getName()));
		ffp.setMinDataColumns(IntegerUtil.getInt((Integer)getJobOrConfigurationValue(minDataColumnsParameter.getName())));
		ffp.setMaxDataColumns(IntegerUtil.getInt((Integer)getJobOrConfigurationValue(maxDataColumnsParameter.getName())));
		
		ffp.setDefaultNumberFormat(getNumberFormat());
		ffp.setDefaultDateFormat(getDateFormatter("yyyy-MM-dd"));
		ffp.setDefaultTimestampFormat(getTimestampFormatter("yyyy-MM-dd HH:mm:ss"));

		ffp.setExcelSheet((String)getJobOrConfigurationValue(excelSheetParameter.getName()));
		return ffp;
	}
	
	/**
		Get the character set the file uses. This method looks for a character set
		in this order:
		<ol>
		<li>Job parameter with name {@link Parameters#CHARSET_PARAMETER}.
		<li>Configuration parameter with name {@link Parameters#CHARSET_PARAMETER}.
		<li>Character set specified by file to import from {@link File#getCharacterSet()}.
		<li>System default characeter set {@link Config#getCharset()}.
		</ol>
		
		first checks the job
		parameters for a value, then the configuration parameters. If not 
		found the {@link Config#getCharset()} is returned.
		@return The name of the charset to use
		@see #getCharset(Request)
	*/
	protected String getCharset()
	{
		String charset = (String)getJobOrConfigurationValue(Parameters.CHARSET_PARAMETER);
		if (charset == null && job != null)
		{
			File file = (File)job.getValue("file");
			if (file != null) charset = file.getCharacterSet();
		}
		if (charset == null) charset = Config.getCharset();
		return charset;
	}
	
	/**
		Get the character set the file uses. This method first checks 
		the request object, and then calls {@link #getCharset()}.
		@return The name of the charset to use
		@since 2.8
	*/
	protected String getCharset(Request request)
	{
		String charset = null;
		if (request != null) 
		{
			charset = (String)request.getParameterValue(Parameters.CHARSET_PARAMETER);
		}
		if (charset == null) charset = getCharset();
		return charset;
	}
	
	/**
		Get the decimal separator used by numbers in the file. This method first 
		checks the job parameters for a value, then the configuration parameters. 
		If not found null is returned.
		@return The decimal separator or null to use Float.valueOf() or Double.valueOf()
		@since 2.2
	*/
	protected String getDecimalSeparator()
	{
		return (String)getJobOrConfigurationValue(Parameters.DECIMAL_SEPARATOR_PARAMETER);
	}
	
	/**
		Get a number formatter that is able to parse numbers with the
		specified decimal separator. Returns null if no decimal separator
		has been specified which causes numbers to be parsed with Float.valueOf()
		or Double.valueOf().
		@return The number format or null to use Float.valueOf() or Double.valueOf()
		@since 2.2
	*/
	protected NumberFormat getNumberFormat()
	{
		String decimalSeparator = getDecimalSeparator();
		char ds = 0;
		if ("dot".equals(decimalSeparator))
		{
			ds = '.';
		}
		else if ("comma".equals(decimalSeparator))
		{
			ds = ',';
		}
		return ds == 0 ? null : NumberFormatUtil.getNumberFormat(ds, ';');
	}
	
	protected Formatter<Date> getDateFormatter(String defaultFormat)
	{
		String dateFormat = Values.getString((String)getJobOrConfigurationValue(Parameters.DATE_FORMAT_PARAMETER), defaultFormat);
		if (dateFormat == null) return null;		
		DateFormatter dateFormatter = new DateFormatter(dateFormat);
		dateFormatter.getDateFormat().setLenient(false);
		return dateFormatter;
	}

	protected Formatter<Date> getTimestampFormatter(String defaultFormat)
	{
		String dateFormat = Values.getString((String)getJobOrConfigurationValue(Parameters.TIMESTAMP_FORMAT_PARAMETER), defaultFormat);
		if (dateFormat == null) return null;		
		DateFormatter dateFormatter = new DateFormatter(dateFormat);
		dateFormatter.getDateFormat().setLenient(false);
		return dateFormatter;
	}
	
	protected Pattern getPattern(String name)
		throws BaseException
	{
		String regexp = (String)getJobOrConfigurationValue(name);
		return regexp != null ? Pattern.compile(regexp) : null;
	}
	
	/**
		Check if a column mapping is a valid mapping expression and 
		optionally if it is complex mapping.

		@param mapExpression The mapping expression
		@param allowComplex If complex column mappings should be allowed
		@param name The name of the column (used if an error message needs to be generated)
		@return The mapping string
		@throws InvalidDataException If the mapping isn't a valid column 
			mapping or if the allowComplex parameter is false and the mapping is complex
		@since 2.4
	*/
	protected String checkColumnMapping(String mapExpression, boolean allowComplex, String name)
		throws InvalidDataException
	{
		return checkColumnMapping(null, mapExpression, allowComplex, name);
	}
	
	/**
		Checks the syntax column mapping and verifies that the given file parser
		has found the columns that are used in the file.
		
		@param ffp An optional flat file parser that should have parsed the file headers
			with {@link FlatFileParser#parseHeaders()}
		@param mapExpression The mapping expression
		@param allowComplex If complex column mappings should be allowed
		@param name The name of the column (used if an error message needs to be generated)
		@return The mapping string
		@throws InvalidDataException If the mapping isn't a valid column 
			mapping or if the allowComplex parameter is false and the mapping is complex
		@since 2.9
	*/
	protected String checkColumnMapping(FlatFileParser ffp, String mapExpression, 
		boolean allowComplex, String name)
		throws InvalidDataException
	{
		if (mapExpression == null || mapExpression.length() == 0) 
		{
			return mapExpression;
		}
		
		// Check if it is a JEP mapping
		boolean isJep = mapExpression.startsWith("=");
		
		// Check if the mapping is simple
		if (!allowComplex)
		{
			if (isJep)
			{
				throw new InvalidDataException("Not allowed to use expression mapping for column '" +
						name + "': " +mapExpression);
			}
			if (mapExpression.contains("\\"))
			{
				// The regexp matches strings that start and end with \
				if (!mapExpression.matches("\\\\[^\\\\]+\\\\"))
				{
					throw new InvalidDataException("Not allowed to use complex mapping for column '" +
							name + "': " + mapExpression);
				}
			}
		}
		
		if (ffp != null)
		{
			try
			{
				ffp.getMapper(mapExpression);
			}
			catch (RuntimeException t)
			{
				throw new InvalidDataException("Invalid expression for column '" +
						name + "': " + mapExpression + " (" + t.getMessage() + ")", t);
			}
		}
		else if (isJep)
		{
			try
			{
				new JepMapper(mapExpression.substring(1), new ArrayList<String>(), null, null, false, 
					new DefaultItemFunction<Nameable>(null, null, null));
			}
			catch (Throwable t)
			{
				throw new InvalidDataException("Invalid expression for column '" +
						name + "': " + mapExpression + " (" + t.getMessage() + ")", t);
			}
		}
		return mapExpression;
	}
	
	/**
		@see #getMapper(FlatFileParser, String, Integer, Mapper, JepFunction...)
	*/
	protected Mapper getMapper(FlatFileParser ffp, String mapExpression, Integer maxStringLength, Mapper defaultMapper)
	{
		return getMapper(ffp, mapExpression, maxStringLength, defaultMapper, (JepFunction[])null);
	}

	/**
		Get a mapper for the specified flat file parser. This method
		calls the {@link FlatFileParser#getMapper(String)} method to create 
		a mapper, if the map expression isn't null. If a max string length has been
		specified the created mapper is wrapped by a {@link CropStringMapper}
		that crops strings returned by the {@link Mapper#getValue(FlatFileParser.Data)} method 
		to the specified length. Use this method mainly for creating mappers for string values.
		
		@param ffp The flat file parser
		@param mapExpression The map expression, a null value is allows but no mapper is created
		@param maxStringLength The maximum allowed string length or null to allow any length
		@param defaultMapper The mapper to return if the map expression is null
		@param functions Additional JEP functions that should be included in the parser
			in case the map expression is a JEP expression
		@return A mapper, or null if the map expression and default mapper is null
		@since 3.1
	*/
	protected Mapper getMapper(FlatFileParser ffp, String mapExpression, Integer maxStringLength, 
		Mapper defaultMapper, JepFunction... functions)
	{
		Mapper mapper = defaultMapper;
		if (mapExpression != null)
		{
			mapper = ffp.getMapper(mapExpression, functions);
			if (maxStringLength != null && maxStringLength > 0) mapper = new CropStringMapper(mapper, maxStringLength);
		}
		return mapper;
	}
	
	/**
		Get the value for an error handling parameter. If no value
		is found for the specified parameter the value for the 'defaultError' parameter
		is returned.
		@param parameterName The name of the error parameter
		@return The error option or <code>fail</code> if no option has been set
		@since 2.4
	*/
	protected String getErrorOption(String parameterName)
	{
		Object option = null;
		option = job.getValue(parameterName);
		if (option == null) option = job.getValue("defaultError");
		return option == null ? "fail" : option.toString();
	}
	
	protected ErrorHandler getErrorHandler(String method, ErrorHandler defaultErrorHandler)
	{
		ErrorHandler handler = defaultErrorHandler;
		if ("skip".equals(method))
		{
			handler = new SimpleErrorHandler(true);
		}
		else if ("fail".equals(method))
		{
			handler = new SimpleErrorHandler(false);
		}
		return handler;
	}
	
	/**
		If the importer should continue with the next file after an error.
		The default implementation of this method return FALSE. A subclass
		may override this method to let the import continue after an error.
		This method isn't called until an error has happened, and it is
		possible to control what should happen on a file-by-file bases. 
		The importer will of course exit if this method returns FALSE.
		@param t The error that happened
		@return TRUE to contine, FALSE to abort
		@since 2.9
	*/
	protected boolean continueWithNextFileAfterError(Throwable t)
	{
		return false;
	}
	
	/**
		Initialise the error handling system. This method is called just before the 
		import of a file is starting. A subclass may override this method to add specific 
		error handlers. If <code>super.setUpErrorHandling()</code> isn't called
		error handling in AbstractFlatFileImporter is disabled and the subclass
		must do all it's error handling in it's own code. The subclass may also add 
		error handlers in the {@link #begin(FlatFileParser)} method. Note that
		the error handling system is re-initialised for every file returned by
		{@link #getFileIterator()}.
	*/
	protected void setUpErrorHandling()
	{
		ErrorHandler defaultErrorHandler = new SimpleErrorHandler("skip".equals(job.getValue("defaultError")));
		errorHandler = new ClassMapErrorHandler(defaultErrorHandler);
		addErrorHandler(StringTooLongException.class, 
				getErrorHandler((String)job.getValue("stringTooLongError"), defaultErrorHandler));
		addErrorHandler(InvalidUseOfNullException.class, 
				getErrorHandler((String)job.getValue("invalidUseOfNullError"), defaultErrorHandler));
		addErrorHandler(NumberOutOfRangeException.class, 
				getErrorHandler((String)job.getValue("numberOutOfRangeError"), defaultErrorHandler));
		addErrorHandler(NumberFormatException.class,
				getErrorHandler((String)job.getValue("numberFormatError"), defaultErrorHandler));
	}
	
	/**
		Add an error handler for the specified class of error. The error handler
		also handles error that are subclasses of the specified class.
		@see ClassMapErrorHandler#addErrorHandler(Class, ErrorHandler)
	*/
	protected void addErrorHandler(Class<? extends Throwable> t, ErrorHandler handler)
	{
		errorHandler.addErrorHandler(t, handler);
	}
	
	/**
		Log a message about a data line to the log file created by 
		{@link #createLogFile(String)}. If no log file has been created, 
		this method does nothing.
		
		@param message The message to log
		@param data The data line the log message is related to
		@since 2.8
	*/
	protected void log(String message, FlatFileParser.Data data)
	{
		log("Line " + data.lineNo() + ": " + message);
	}

	/**
		Log an error message about a data line to the log file created 
		by {@link #createLogFile(String)}. If no log file has been created, 
		this method does nothing.
		
		@param message The message to log
		@param data The data line the log message is related to
		@param t The error
		@since 2.8
	*/
	protected void log(String message, FlatFileParser.Data data, Throwable t)
	{
		log("Line " + data.lineNo() + ": " + message, t);
	}
	
	/**
		Log a message about a header line to the log file created by 
		{@link #createLogFile(String)}. If no log file has been created, 
		this method does nothing.
		
		@param message The message to log
		@param line The header line line the log message is related to
		@since 2.8
	*/
	protected void log(String message, FlatFileParser.Line line)
	{
		log("Line " + line.lineNo() + ": " + message);
	}

	/**
		Log an error message about a header line to the log file created by 
		{@link #createLogFile(String)}. If no log file has been created, 
		this method does nothing.
		
		@param message The message to log
		@param line The header line line the log message is related to
		@param t The error
		@since 2.8
	*/
	protected void log(String message, FlatFileParser.Line line, Throwable t)
	{
		log("Line " + line.lineNo() + ": " + message, t);
	}

	
}
