/*
	$Id: TestExperiment.java 5060 2009-08-19 07:02:11Z nicklas $

	Copyright (C) 2010 Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicExtraValueQuery;
import net.sf.basedb.core.DynamicResultIterator;
import net.sf.basedb.core.DynamicSpotQuery;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.ExtraValue;
import net.sf.basedb.core.ExtraValueType;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileSet;
import net.sf.basedb.core.FileSetMember;
import net.sf.basedb.core.FileStoreUtil;
import net.sf.basedb.core.FilterBatcher;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.IntensityTransform;
import net.sf.basedb.core.ItemResultList;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.PositionExtraValueBatcher;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.SpotBatcher;
import net.sf.basedb.core.SpotExtraValueBatcher;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.Transformation;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.core.query.Selects;

import net.sf.basedb.util.IntensityCalculatorUtil;
import net.sf.basedb.util.jep.Jep;
import net.sf.basedb.util.jep.RawFunction;
import net.sf.basedb.util.ConsoleProgressReporter;

import java.io.PrintWriter;
import java.util.Collections;
import java.util.List;
import java.util.ArrayList;


public class TestBasicAnalysis
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		boolean wasSetup = TestAnalyzeUtil.setupExperiment();
		write("++Testing basic analysis");
		try
		{
			int experimentId = TestAnalyzeUtil.getItemId("experiment.genepix");
			int rootBasId = TestAnalyzeUtil.getItemId("bioassayset.root");
			if (experimentId == 0 || rootBasId == 0)
			{
				ok = false;
				write("Test experiment was not setup correctly.");
				return ok;
			}
			
			// Create some root bioassay sets
			int emptyRootId = test_create_empty_root_bioassayset(experimentId, "Empty root");
			int queryRootId = test_create_root_bioassayset_using_query(experimentId, "mean", "Query mean");
			TestAnalyzeUtil.setItemId("bioassayset.root.empty", emptyRootId);
			TestAnalyzeUtil.setItemId("bioassayset.root.query", queryRootId);
	
			// Create a filtered bioassay set using core api
			int filterBasId = test_filter_bioassayset(queryRootId, "Filtered; ch1 > 2*ch2", 2.0f);
			TestAnalyzeUtil.setItemId("bioassayset.filtered", filterBasId);
			
			// Create a transformed bioassay set using core api
			int transformedBasId = test_transform_bioassayset(filterBasId, "Transformed; ch1*2, ch2/2", 2.0f);
			TestAnalyzeUtil.setItemId("bioassayset.transformed", transformedBasId);
	
			// Create a file-only bioassay set using core api
			int fileBasId = test_create_file_bioassayset(rootBasId, "/spotdata-" + rootBasId + ".txt", 
				"File bioassayset");
			TestAnalyzeUtil.setItemId("bioassayset.file-only", fileBasId);
			
			// Create some extra values using core api
			int intExtraId = TestAnalyzeUtil.getItemId("extravaluetype.int");
			int floatExtraId = TestAnalyzeUtil.getItemId("extravaluetype.float");
			int stringExtraId = TestAnalyzeUtil.getItemId("extravaluetype.string");
			
			int extraSpotId = test_calculate_spotextravalues(filterBasId, "Spot extra values",
					intExtraId, floatExtraId, stringExtraId);
			int extraPosId = test_calculate_positionextravalues(filterBasId, "Position extra values",
					intExtraId, floatExtraId, stringExtraId);
			
			test_list_bioassaysets(experimentId, 8);
			test_list_bioassays(transformedBasId, 2);
			
			write_data_header();
			test_list_bioassayset_data(extraSpotId, 10, 2, intExtraId, floatExtraId, stringExtraId);
			
			write_extra_values_header();
			test_list_extra_values(extraSpotId, intExtraId, 10, 2);
			
			if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
			TestAnalyzeUtil.test_delete_bioassayset(extraPosId);
			test_delete_data_files(fileBasId);
		}
		finally
		{
			if (wasSetup) TestAnalyzeUtil.cleanupExperiment();
			write("++Testing basic analysis "+(ok ? "OK" : "Failed")+"\n");
		}
		return ok;
	}


	static void write_item(int i, RawBioAssay rba)
		throws BaseException
	{
		if (!TestUtil.getSilent()) 
		{
			System.out.println(i+":\t"+rba.getId()+"\t"+rba.getName()+"\t"+rba.getDescription()+
				"\t"+rba.getRawDataType().getName()+"\t"+rba.getNumDbSpots()+"\t"+rba.getBytes()+"\t"+rba.getArrayDesign()+
				"\t"+rba.getSoftware()+"\t"+rba.getProtocol());
		}
	}
	static void write_item(int i, BioAssay ba)
		throws BaseException
	{
		if (!TestUtil.getSilent()) 
		{
			System.out.println(i+":\t"+ba.getId()+"\t"+ba.getName()+"\t"+ba.getDescription()+
				"\t"+ba.getRawDataType().getName()+"\t"+ba.getNumSpots());
		}
	}
	static void write_item(int i, BioAssaySet bas)
		throws BaseException
	{
		if (!TestUtil.getSilent()) 
		{
			System.out.println(i+":\t"+bas.getId()+"\t"+bas.getName()+"\t"+bas.getDescription()+
				"\t"+bas.getTransformation()+"\t"+bas.getNumReporters()+"\t"+bas.getNumSpots() +
				"\t" + bas.getNumFileReporters() + "\t" + bas.getNumFileSpots());
		}
	}
	static void write_item(int i, ExtraValue ev)
		throws BaseException
	{
		if (!TestUtil.getSilent()) 
		{
			System.out.println("\tExtra value: "+i+":\t"+ev.getId()+"\t"+ev.getValueType()+"\t"+
				ev.getCoordinateType()+"\t"+ev.getExtraValueType().getExternalId()+"\t"+ev.getNumValues());
		}
	}
	static void write_data_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tColumn\tPosition\tChannel 1\tChannel 2\tReporter\tX  \tY  \tExtra");
			write("-- \t------\t--------\t---------\t---------\t--------\t---\t---\t-----");
		}
	}
	static void write_extra_values_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tColumn\tPosition\tExtra");
			write("-- \t------\t--------\t-----");
		}
	}
	static void write_data(int i, short column, int position, float ch1, float ch2, 
		String reporter, float x, float y, List<Object> extra)
	{
		if (!TestUtil.getSilent())
		{
			System.out.println(i+":\t"+column+"\t"+position+"\t"+ch1+"\t"+ch2+"\t"+reporter+
				"\t"+x+"\t"+y+"\t"+extra);
		}
	}
	static void write_extra_value(int i, short column, int position, Object extra)
	{
		if (!TestUtil.getSilent())
		{
			System.out.println(i+":\t"+column+"\t"+position+"\t"+extra);
		}
	}
	
	static void write(String message)
	{
		System.out.println(message);
	}
	


	
	static int test_create_root_bioassayset_using_query(int experimentId, String intensityFormula, String name)
	{
		if (experimentId == 0) return 0;
		int id = 0;
		DbControl dc = null;
		long time;
		try
		{
			dc = TestUtil.getDbControl();
			Experiment e = Experiment.getById(dc, experimentId);
			
			List<RawBioAssay> rbaCollection = e.getRawBioAssays().list(dc);
			
			RawDataType rdt = e.getRawDataType();
			String[] formulas = rdt.getIntensityFormula(intensityFormula).getExpressions();
			RawFunction raw = new RawFunction(dc, e.getRawDataType(), true);
			Expression ch1 = Jep.formulaToExpression(formulas[0], raw);
			Expression ch2 = Jep.formulaToExpression(formulas[1], raw);

			time = System.currentTimeMillis();
			BioAssaySet root = IntensityCalculatorUtil.createRootBioAssaySet(
				dc, e, rbaCollection, null, new Expression[] { ch1, ch2 }, 
				TestUtil.getSilent() ? null : new ConsoleProgressReporter() );
			root.setName(name);
			root.getTransformation().setName(name);
			dc.commit();

			time = System.currentTimeMillis()-time;
			int numSpots = root.getNumSpots();
			id = root.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(root, false);
			write_item(0, root);
			write("--Create root bioassayset using query OK ("+numSpots+" spots inserted; "+time+" ms)");
		}
		catch (Throwable ex)
		{
			write("--Create root bioassayset using query FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}
	
	static int test_create_empty_root_bioassayset(int experimentId, String name)
	{
		if (experimentId == 0) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Experiment e = Experiment.getById(dc, experimentId);
			
			List<RawBioAssay> rbaCollection = e.getRawBioAssays().list(dc);
			
			Transformation t = e.newTransformation(null, rbaCollection);
			t.setName(name);
			BioAssaySet root = t.newProduct("new", "new", false);
			root.setName(name);
			dc.saveItem(t);
			dc.saveItem(root);
			
			for (RawBioAssay rba : rbaCollection)
			{
				// Create a new BioAssay for each raw bioassay
				BioAssay ba = root.newRootBioAssay(Collections.singleton(rba));
				ba.setName(rba.getName());
				dc.saveItem(ba);
			}
			dc.commit();

			int numSpots = root.getNumSpots();
			id = root.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(root, false);
			write_item(0, root);
			write("--Create empty root bioassayset OK ("+numSpots+" spots inserted)");
		}
		catch (Throwable ex)
		{
			write("--Create empty root bioassayset FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}
	
	
	static int test_create_file_bioassayset(int bioAssaySetId, String path, String name)
	{
		if (bioAssaySetId == 0) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			
			BioAssaySet source = BioAssaySet.getById(dc, bioAssaySetId);
			Transformation t = source.newTransformation(null);
			t.setName(name);
			dc.saveItem(t);

			BioAssaySet bas = t.newProduct(null, "new", false);
			bas.setName(name);
			dc.saveItem(bas);

			for (BioAssay parentBa : source.getBioAssays().list(dc))
			{
				BioAssay childBa = bas.newBioAssay(parentBa);
				childBa.setNumFileSpots(parentBa.getNumSpots());
				dc.saveItem(childBa);
			}
			
			File f = File.getByPath(dc, new Path(path, Path.Type.FILE), true);
			f.setMimeType("text/plain");
			f.setItemSubtype(ItemSubtype.getById(dc, SystemItems.getId(File.SPOT_DATA)));
			if (!f.isInDatabase()) dc.saveItem(f);
			FileStoreUtil.setDataFile(dc, bas, DataFileType.GENERIC_SPOT_DATA, f);
			bas.setNumFileSpots(source.getNumSpots());
			bas.setNumFileReporters(source.getNumReporters());
			
			PrintWriter out = new PrintWriter(f.getUploadStream(false));
			int numChannels = bas.getRawDataType().getChannels();
			out.print("REPORTER\tASSAY\tPOSITION");
			for (int ch = 1; ch <= numChannels; ++ch)
			{
				out.print("\tCH");
				out.print(ch);
			}
			out.print('\n');
			
			DynamicSpotQuery query = source.getSpotData();
			query.select(Dynamic.selectReporter("externalId"));
			query.select(Dynamic.select(VirtualColumn.COLUMN));
			query.select(Dynamic.select(VirtualColumn.POSITION));
			for (int ch = 1; ch <= numChannels; ++ch)
			{
				query.select(Dynamic.select(VirtualColumn.channelRaw(ch)));
			}
			DynamicResultIterator it = query.iterate(dc);

			while (it.hasNext())
			{
				SqlResult result = it.next();
				out.print(result.getString(1)); // Reporter
				out.print('\t');
				out.print(result.getInt(2)); // Assay
				out.print('\t');
				out.print(result.getInt(3)); // Position
				
				for (int ch = 1; ch <= numChannels; ++ch)
				{
					out.print('\t');
					out.print(result.getFloat(3+ch));
				}
				out.print('\n');
			}
			out.flush();
			out.close();

			dc.commit();
			id = bas.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(bas, false);
			write_item(0, bas);
			write("--Create file bioassayset OK ("+bas.getNumFileSpots()+" spots inserted)");
		}
		catch (Throwable ex)
		{
			write("--Create file bioassayset FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}
	
	
	static int test_transform_bioassayset(int bioAssaySetId, String name, float factor)
	{
		if (bioAssaySetId == 0) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioAssaySet source = BioAssaySet.getById(dc, bioAssaySetId);
			
			Transformation t = source.newTransformation(null);
			t.setName(name);
			dc.saveItem(t);

			BioAssaySet bas = t.newProduct(null, "new", true);
			bas.setName(name);
			bas.setIntensityTransform(source.getIntensityTransform());
			dc.saveItem(bas);

			int numSpots = 0;
			SpotBatcher spotBatcher = bas.getSpotBatcher();
			
			DynamicSpotQuery query = source.getSpotData();
			query.select(Dynamic.select(VirtualColumn.COLUMN));
			query.select(Dynamic.select(VirtualColumn.POSITION));
			query.select(
				Selects.expression(
					Expressions.multiply(
						Dynamic.column(VirtualColumn.channelRaw(1)),
						Expressions.aFloat(factor)
					),
					"ch1"
				)
			);
			query.select(
				Selects.expression(
					Expressions.divide(
						Dynamic.column(VirtualColumn.channelRaw(2)),
						Expressions.aFloat(factor)
					),
					"ch2"
				)
			);
			spotBatcher.insert(query);
			spotBatcher.close();

			dc.commit();
			id = bas.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(bas, false);
			write_item(0, bas);
			write("--Create bioassayset OK ("+numSpots+" spots inserted)");
		}
		catch (Throwable ex)
		{
			write("--Create bioassayset FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static int test_filter_bioassayset(int bioAssaySetId, String name, float factor)
	{
		if (bioAssaySetId == 0) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioAssaySet source = BioAssaySet.getById(dc, bioAssaySetId);
			
			Transformation t = source.newTransformation(null);
			t.setName(name);
			dc.saveItem(t);

			BioAssaySet bas = t.newProduct(null, null, true);
			bas.setName(name);
			bas.setIntensityTransform(source.getIntensityTransform());
			dc.saveItem(bas);

			DynamicSpotQuery query = source.getSpotData();
			query.select(Dynamic.select(VirtualColumn.COLUMN));
			query.select(Dynamic.select(VirtualColumn.POSITION));
			query.restrict(
				Restrictions.gt(
					Dynamic.column(VirtualColumn.channelRaw(1)),
					Expressions.multiply(
						Expressions.parameter("factor"),
						Dynamic.column(VirtualColumn.channelRaw(2))
					)
				)
			);
			query.setParameter("factor", factor, Type.FLOAT);
			
			FilterBatcher filterBatcher = bas.getFilterBatcher();
			DynamicResultIterator spots = query.iterate(dc);
			int column = spots.getIndex(VirtualColumn.COLUMN.getName());
			int position = spots.getIndex(VirtualColumn.POSITION.getName());
			int numSpots = 0;
			while (spots.hasNext())
			{
				SqlResult r = spots.next();
				filterBatcher.insert(r.getShort(column), r.getInt(position));
				numSpots++;
			}
			filterBatcher.close();

			dc.commit();
			id = bas.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(bas, false);
			write_item(0, bas);
			write("--Create filtered bioassayset OK ("+numSpots+" spots inserted)");
		}
		catch (Throwable ex)
		{
			write("--Create filtered bioassayset FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static int test_calculate_spotextravalues(int bioAssaySetId, String childName, 
			int intExtraId, int floatExtraId, int stringExtraId)
	{
		if (bioAssaySetId == 0) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioAssaySet source = BioAssaySet.getById(dc, bioAssaySetId);
			
			Transformation t = source.newTransformation(null);
			dc.saveItem(t);

			BioAssaySet addTo = source;
			if (childName != null)
			{
				t.setName(childName);
				BioAssaySet child = t.newProduct(null, null, true);
				child.setName(childName);
				dc.saveItem(child);
				addTo = child;
			}
			
			SpotExtraValueBatcher<Integer> intBatcher = null;
			if (intExtraId != 0)
			{
				ExtraValueType intExtraValue = ExtraValueType.getById(dc, intExtraId);
				intBatcher = addTo.getSpotExtraValueBatcher(Integer.class, intExtraValue, null);
			}
			SpotExtraValueBatcher<Float> floatBatcher = null;
			if (floatExtraId != 0)
			{
				ExtraValueType floatExtraValue = ExtraValueType.getById(dc, floatExtraId);
				floatBatcher = addTo.getSpotExtraValueBatcher(Float.class, floatExtraValue, null);
			}
			SpotExtraValueBatcher<String> stringBatcher = null;
			if (stringExtraId != 0)
			{
				ExtraValueType stringExtraValue = ExtraValueType.getById(dc, stringExtraId);
				stringBatcher = addTo.getSpotExtraValueBatcher(String.class, stringExtraValue, null);
			}
			
			DynamicSpotQuery query = source.getSpotData();
			DynamicResultIterator spots = query.iterate(dc);
			int column = spots.getIndex(VirtualColumn.COLUMN.getName());
			int position = spots.getIndex(VirtualColumn.POSITION.getName());
			int ch1 = spots.getIndex(VirtualColumn.channelIntensity(1).getName());
			int ch2 = spots.getIndex(VirtualColumn.channelIntensity(2).getName());
			int numSpots = 0;
			while (spots.hasNext())
			{
				numSpots++;
				SqlResult r = spots.next();
				
				if (numSpots == 1)
				{
					// Test NULL and Float.NaN
					if (intBatcher != null) intBatcher.insert(r.getShort(column), r.getInt(position), null);
					if (floatBatcher != null) floatBatcher.insert(r.getShort(column), r.getInt(position), Float.NaN);
					if (stringBatcher != null) stringBatcher.insert(r.getShort(column), r.getInt(position), null);
				}
				else
				{
					if (intBatcher != null) intBatcher.insert(r.getShort(column), r.getInt(position), numSpots);
					if (floatBatcher != null) floatBatcher.insert(r.getShort(column), r.getInt(position), r.getFloat(ch1) * r.getFloat(ch2));
					if (stringBatcher != null) stringBatcher.insert(r.getShort(column), r.getInt(position), "Spot number " + numSpots);
				}
			}

			if (intBatcher != null) intBatcher.close();
			if (floatBatcher != null) floatBatcher.close();
			if (stringBatcher != null) stringBatcher.close();

			dc.commit();
			id = addTo.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(addTo, false);
			write_item(0, addTo);
			write("--Calculate spot extra values OK ("+numSpots+" values inserted)");

		}
		catch (Throwable ex)
		{
			write("--Calculate spot extra values FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static int test_calculate_positionextravalues(int bioAssaySetId, String childName, int intExtraId, int floatExtraId, int stringExtraId)
	{
		if (bioAssaySetId == 0) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioAssaySet source = BioAssaySet.getById(dc, bioAssaySetId);
			
			Transformation t = source.newTransformation(null);
			t.setName(childName);
			dc.saveItem(t);

			BioAssaySet bas = t.newProduct(null, null, true);
			bas.setName(childName);
			IntensityTransform transform = source.getIntensityTransform();
			bas.setIntensityTransform(transform);
			dc.saveItem(bas);
			
			PositionExtraValueBatcher<Integer> intBatcher = null;
			if (intExtraId != 0)
			{
				ExtraValueType intExtraValue = ExtraValueType.getById(dc, intExtraId);
				intBatcher = bas.getPositionExtraValueBatcher(Integer.class, intExtraValue, null);
			}
			PositionExtraValueBatcher<Float> floatBatcher = null;
			if (floatExtraId != 0)
			{
				ExtraValueType floatExtraValue = ExtraValueType.getById(dc, floatExtraId);
				floatBatcher = bas.getPositionExtraValueBatcher(Float.class, floatExtraValue, null);
			}
			PositionExtraValueBatcher<String> stringBatcher = null;
			if (stringExtraId != 0)
			{
				ExtraValueType stringExtraValue = ExtraValueType.getById(dc, stringExtraId);
				stringBatcher = bas.getPositionExtraValueBatcher(String.class, stringExtraValue, null);
			}

			Formula.AverageMethod avgMethod = transform.getAverageMethod();
			Expression avgCh1 = avgMethod.getAverageExpression(Dynamic.column(VirtualColumn.channelRaw(1)));
			Expression avgCh2 = avgMethod.getAverageExpression(Dynamic.column(VirtualColumn.channelRaw(2)));
			Expression product = Expressions.multiply(avgCh1, avgCh2);
			DynamicSpotQuery query = source.getSpotData();
			query.select(Dynamic.select(VirtualColumn.POSITION));
			query.select(Selects.expression(product, "product"));
			query.select(Selects.expression(Expressions.string("static"), "static"));
			query.group(Dynamic.column(VirtualColumn.POSITION));

			DynamicResultIterator spots = query.iterate(dc);
			int position = spots.getIndex(VirtualColumn.POSITION.getName());
			int prod = spots.getIndex("product");
			int stat = spots.getIndex("static");
			int numSpots = 0;
			while (spots.hasNext())
			{
				numSpots++;
				SqlResult r = spots.next();
				if (numSpots == 1)
				{
					if (intBatcher != null) intBatcher.insert(r.getInt(position), null);
					if (floatBatcher != null) floatBatcher.insert(r.getInt(position), Float.NaN);
					if (stringBatcher != null) stringBatcher.insert(r.getInt(position), null);
				}
				else
				{
					if (intBatcher != null) intBatcher.insert(r.getInt(position), numSpots);
					if (floatBatcher != null) floatBatcher.insert(r.getInt(position), r.getFloat(prod));
					if (stringBatcher != null) stringBatcher.insert(r.getInt(position), r.getString(stat));
				}
			}

			if (intBatcher != null) intBatcher.close();
			if (floatBatcher != null) floatBatcher.close();
			if (stringBatcher != null) stringBatcher.close();

			dc.commit();
			id = bas.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(bas, false);
			write_item(0, bas);
			write("--Calculate position extra values OK ("+numSpots+" values inserted)");

		}
		catch (Throwable ex)
		{
			write("--Calculate position extra values FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	

	
	static void test_list_bioassays(int bioassaysetId, int expectedResults)
	{
		if (bioassaysetId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioAssaySet bas = BioAssaySet.getById(dc, bioassaysetId);
			
			ItemResultList<BioAssay> l = bas.getBioAssays().list(dc);
			for (int i = 0; i < l.size(); i++)
			{
				BioAssay ba = l.get(i);
				write_item(i, ba);
				for (RawBioAssay rba : ba.getRawBioAssays().list(dc))
				{
					write_item(i, rba);
				}
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List bioassays OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List bioassays FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_list_bioassaysets(int experimentId, int expectedResults)
	{
		if (experimentId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Experiment e = Experiment.getById(dc, experimentId);

			ItemResultList<BioAssaySet> l = e.getBioAssaySets().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				BioAssaySet bas = l.get(i);
				write_item(i, bas);
				
				ItemResultList<ExtraValue> ev = bas.getExtraValues().list(dc);
				
				for (int j = 0; j < ev.size(); j++)
				{
					write_item(j, ev.get(j));	
				}
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List bioassaysets OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List bioassaysets FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_list_bioassayset_data(int bioAssaySetId, int limit, int offset, int... extraValueTypes)
	{
		if (bioAssaySetId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioAssaySet bas = BioAssaySet.getById(dc, bioAssaySetId);
			
			DynamicSpotQuery query = bas.getSpotData();
			query.select(Dynamic.select(VirtualColumn.COLUMN));
			query.select(Dynamic.select(VirtualColumn.POSITION));
			query.select(Dynamic.select(VirtualColumn.channelRaw(1)));
			query.select(Dynamic.select(VirtualColumn.channelRaw(2)));
			query.select(Dynamic.selectReporter("externalId"));
			query.select(Dynamic.selectRawData("x"));
			query.select(Dynamic.selectRawData("y"));
			query.joinReporters(JoinType.LEFT);
			query.joinRawData(JoinType.LEFT);
			
			for (int extraValueTypeId : extraValueTypes)
			{
				ExtraValueType evt = ExtraValueType.getById(dc, extraValueTypeId);
				ExtraValue ev = bas.getExtraValue(evt);
				query.joinExtraValue(ev, JoinType.INNER);
				query.select(Dynamic.select(ev, "ev"+extraValueTypeId));
			}
			
			query.setMaxResults(limit);
			query.setFirstResult(offset);
			query.setReturnTotalCount(true);
			
			DynamicResultIterator spots = query.iterate(dc);
			int column = spots.getIndex(VirtualColumn.COLUMN.getName());
			int position = spots.getIndex(VirtualColumn.POSITION.getName());
			int ch1 = spots.getIndex(VirtualColumn.channelRaw(1).getName());
			int ch2 = spots.getIndex(VirtualColumn.channelRaw(2).getName());
			int reporter = spots.getIndex("externalId");
			int x = spots.getIndex("x");
			int y = spots.getIndex("y");
			int[] extraIndex = new int[extraValueTypes.length];
			for (int i = 0; i < extraIndex.length; ++i)
			{
				extraIndex[i] = spots.getIndex("ev"+extraValueTypes[i]);	
			}
			List<Object> extraValues = new ArrayList<Object>(extraIndex.length);
			int listed = 0;
			while (spots.hasNext())
			{
				SqlResult r = spots.next();
				for (int i = 0; i < extraIndex.length; ++i)
				{
					extraValues.add(r.getObject(extraIndex[i]));	
				}
				write_data(listed, r.getShort(column), r.getInt(position), 
					r.getFloat(ch1), r.getFloat(ch2), r.getString(reporter), 
					r.getFloat(x), r.getFloat(y), extraValues);
				extraValues.clear();
				listed++;
			}
			spots.close();
			
			dc.close();
			write("--List bioassayset data (listed = " + listed + "; total = " + spots.getTotalCount() + ") OK");
		}
		catch (Throwable ex)
		{
			write("--List bioassayset data FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list_extra_values(int bioAssaySetId, int extraValueTypeId, int limit, int offset)
	{
		if (bioAssaySetId == 0 || extraValueTypeId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioAssaySet bas = BioAssaySet.getById(dc, bioAssaySetId);
			ExtraValueType evt = ExtraValueType.getById(dc, extraValueTypeId);
			ExtraValue ev = bas.getExtraValue(evt);
	
			DynamicExtraValueQuery query = ev.getValues();
			query.setMaxResults(limit);
			query.setFirstResult(offset);
			query.setReturnTotalCount(true);
			
			DynamicResultIterator values = query.iterate(dc);
			int column = values.getIndex(VirtualColumn.COLUMN.getName());
			int position = values.getIndex(VirtualColumn.POSITION.getName());
			int extra = values.getIndex(ev.getVirtualColumn().getName());
			int listed = 0;
			while (values.hasNext())
			{
				SqlResult r = values.next();
				write_extra_value(listed, r.getShort(column), r.getInt(position), 
					r.getObject(extra));
				listed++;
			}
			values.close();
			
			dc.close();
			write("--List extra values (listed = " + listed + "; total = " + values.getTotalCount() + ") OK");
		}
		catch (Throwable ex)
		{
			write("--List extra values FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}

	}

	static void test_delete_data_files(int bioAssaySetId)
	{
		if (bioAssaySetId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioAssaySet bas = BioAssaySet.getById(dc, bioAssaySetId);
			List<File> files = new ArrayList<File>();
			if (bas.hasFileSet())
			{
				FileSet fs = bas.getFileSet();
				List<FileSetMember> members = fs.getMembers().list(dc);
				for (FileSetMember fm : members)
				{
					files.add(fm.getFile());
					fs.removeMember(fm);
				}
			}
			bas.setNumFileReporters(0);
			bas.setNumFileSpots(0);
			dc.commit();
			
			dc = TestUtil.getDbControl();
			for (File f : files)
			{
				dc.deleteItem(f);
			}
			dc.commit();
			write("--Delete data files OK (" + files.size() + ")");
		}
		catch (Throwable ex)
		{
			write("--Delete data files FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
}
