/**
	$Id: TestUtil.java 7522 2018-11-06 14:50:29Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.test;

import java.io.File;
import java.io.FileFilter;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.Writer;
import java.net.InetAddress;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.Driver;
import java.sql.DriverManager;
import java.util.Properties;

import net.sf.basedb.core.Application;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.authentication.LoginRequest;

/**

	@author nicklas
	@version 2.0
	@base.modified $Date: 2018-11-06 15:50:29 +0100 (Tue, 06 Nov 2018) $
*/
public class TestUtil
{
	
	// Require that asserts are enabled for test programs
	static 
	{
		boolean assertsEnabled = false;
		assert (assertsEnabled = true) == true;
		if (!assertsEnabled) throw new AssertionError("Asserts must be enabled.");
	}

	
	private static String client = "net.sf.basedb.clients.test";
	private static SessionControl sc = null;
	private static Writer memoryLog = null;
	private static long startTime = 0;
	private static Runtime runtime = Runtime.getRuntime();

	/**
		Start the testing. Starts BASE and creates a SessionControl.
	*/
	public static void begin(boolean logMemoryUsage, boolean checkTestdata)
	{
		if (Application.isRunning()) return;
		showDbInfo();
		try
		{
			if (checkTestdata)
			{
				File testdata = new File("../..", "testdata");
				if (!testdata.exists() && !testdata.isDirectory())
				{
					throw new FileNotFoundException("Could not find 'testdata' directory. " +
						"It must be located in the project root.");
				}
			}
			write("++Starting BASE...");
			if (logMemoryUsage)
			{
				memoryLog = new FileWriter("memory.log");
				memoryLog.write("Time (ms)\tTotal\tFree\tUsed\n");
				startTime = System.currentTimeMillis();
			}
			Application.start(false);
			write("done.\n");
			write("--Cleaning static cache from old items\n");
			Application.getStaticCache().cleanUp(new FileFilter()
			{
				@Override
				public boolean accept(File pathname)
				{
					return true;
				}
			});
			sc = Application.newSessionControl(client, getLocalIp(), null);
		}
		catch (Throwable ex)
		{
			write("FAILURE INITIALIZING TEST\n");
			ex.printStackTrace();
			System.exit(0);
		}
	}
	
	/**
		Stop testing. Logs out and stops BASE.
	*/
	public static void stop()
	{
		if (!Application.isRunning()) return;
		write("++Stopping BASE...");
		try
		{
			if (sc != null && sc.isLoggedIn()) sc.logout();
			Application.stop();
			if (memoryLog != null) memoryLog.close();
			write("done.\n");
		}
		catch (Throwable ex)
		{
			ex.printStackTrace();
		}
	}

	/**
		Get the SessionControl used for this test session.
	*/
	public static SessionControl getSessionControl()
	{
		if (sc == null) begin(false, true);
		return sc;
	}

	/**
		Create a new DbControl.
	*/
	public static DbControl getDbControl()
	{
		if (memoryLog != null)
		{
			try
			{
				long time = System.currentTimeMillis() - startTime;
				long total = runtime.totalMemory();
				long free = runtime.freeMemory();
				long used = total - free;
				memoryLog.write(time + "\t" + total +"\t"+free+"\t"+used+"\n");
				memoryLog.flush();
			}
			catch (Throwable t)
			{}
		}
		return getSessionControl().newDbControl();
	}

	/**
	 	Login to BASE.
	*/
	public static void login(String login, String password)
	{
		write("--Logging in as: " + login + "\n");
		LoginRequest loginRequest = new LoginRequest(login, password);
		loginRequest.setComment("Running test program");
		getSessionControl().login(loginRequest);
	}

	/**
		Log out.
	*/
	public static void logout()
	{
		if (sc != null) 
		{
			write("--Logging out\n");
			sc.logout();
		}
	}

	
	/**
		Get the local IP address as a string.
	*/
	public static String getLocalIp()
	{
		try
		{
			return InetAddress.getLocalHost().toString();
		}
		catch (Throwable ex)
		{
			return "unknown";
		}
	}
	
	/**
		Prints <code>Press ENTER to contine...</code> on standard output and
		waits for the user to press enter.
	*/
	public static void waitForEnter()
	{
		try
		{
			System.out.print("Press ENTER to continue...");
			System.in.read();
			System.in.skip(System.in.available());
		}
		catch (Exception ex)
		{}
	}

	/**
		Print a message to standard output.
	*/
	public static void write(String msg)
	{
		System.out.print(msg);
	}
	private static final long MB = 1024 * 1024;
	public static void gc()
	{
		Runtime rt = Runtime.getRuntime();
		long freeBefore = rt.freeMemory() / MB;
		System.gc();
		long freeAfter = rt.freeMemory() / MB;
		write("Free memory: before=" + freeBefore + "MB; after=" + freeAfter + "MB\n");
	}

	private static void showDbInfo()
	{
		final Properties properties = System.getProperties();
		final String dbUrl = Config.getString("db.url");
		String driverClass = "unknown class";
		String driverVersion = "unknown version";
		String database = "Unknown database";
		try
		{
			final Driver driver = DriverManager.getDriver(dbUrl);
			driverClass = driver.getClass().getName();
			driverVersion = "version " + driver.getMajorVersion() + "." + driver.getMinorVersion();
			Properties p = new Properties();
			p.setProperty("user", Config.getString("db.username"));
			p.setProperty("password", Config.getString("db.password"));
			Connection c = driver.connect(dbUrl, p);
			DatabaseMetaData data = c.getMetaData();
			database = data.getDatabaseProductName() + " " + data.getDatabaseProductVersion();
			c.close();
		}
		catch (Throwable t)
		{}
		
		String message = 
			"--System information-----------------------------\n" + 
			"BASE      : " + Version.getVersion() + "\n" + 
			"Database  : " + database + "\n" + 
			"Dialect   : " + Config.getString("db.dialect") + "\n" + 
			"Hibernate : " + org.hibernate.Version.getVersionString() + "\n"+
			"JDBC      : " + driverClass + "; " + driverVersion + "\n" +
			"URL       : " + dbUrl + "\n" +
			"Java      : " + properties.getProperty("java.runtime.name") + "; " +
				properties.getProperty("java.runtime.version") + "; " +
				properties.getProperty("java.vendor") + "\n" + 
			"OS        : " + properties.getProperty("os.name") +"; " +
				properties.getProperty("os.arch") + "; " +
				properties.getProperty("os.version") +"\n" +
			"-------------------------------------------------";
		System.out.println(message);
	}
	
	
}
